/* last edit: Ilja Schmelzer -------------- 10-JUN-1994 15:08:01.93	*/
/************************************************************************/
/*                                                                      */
/*  <<< I B G >>> - Intersection - Based Grid generation package 	*/
/*                                                                      */
/*  Version 1.1 by Ilja Schmelzer   schmelzer@iaas-berlin.d400.de       */
/*                                                                      */
/*  to be distributed under IBG license conditions (see "readme.ibg")	*/
/*                                                                      */
/************************************************************************/
#ifndef IBGD_H
#define IBGD_H

#ifndef IBG_H
#include "ibg.h"
#endif

/* standard operations: */
extern void		ibgdInit();
extern int		ibgdQuit();
extern ibGeometry	ibgdCopy(ibGeometry old);
extern int		ibgdFree(ibGeometry old);

/* trivial geometry --- only one region: */
extern ibGeometry ibgdNew(ibgSegment region);

/* constant shift: */
extern ibGeometry ibgdShift(ibGeometry gold,
					ibgFloat x[ibgDIM]);

/* linear transformation: */
extern ibGeometry ibgdLinearTrans(ibGeometry gold,
					ibgFloat a[ibgDIM][ibgDIM]);

	/* splitting using characteristic functions: */

/* splitting a region: */
extern ibGeometry ibgdSplitRegion(ibGeometry gold,
			ibgSegment		rnew,
			ibgSegment		rold,
			ibgFloat	(*func)(ibgPtObject fThis, ibgPoint *n),
		   	ibgPtObject	fThis,
			ibgFloat	Dfunc);

/* splitting a face: */
ibGeometry ibgdSplitFace(ibGeometry gold,
			ibgSegment		fnew,
			ibgSegment		fold,
			ibgFloat	(*func)(ibgPtObject fThis, ibgPoint *n),
			ibgPtObject	fThis,
	       		ibgFloat	Dfunc);

/* splitting a line: */
ibGeometry ibgdSplitLine(ibGeometry gold,
			ibgSegment		lnew,
			ibgSegment		lold,
			ibgFloat	(*func)(ibgPtObject fThis, ibgPoint *n),
			ibgPtObject	fThis,
			ibgFloat	Dfunc);

/* computing a function value: */
ibGeometry ibgdFunction(ibGeometry gold, int fnum,
			ibgFloat (*func)(void *fThis, ibgPoint *n),
			ibgPtObject fThis);

/* 1D-continuation outside the window: */
ibGeometry ibgdWindow(ibGeometry gold,
		ibgFloat	xmin[ibgDIM],
		ibgFloat	xmax[ibgDIM]);

/* using a grid as geometry description: */
ibGeometry ibgdGrid(ibGrid *gg, int fanz, int *flist);
ibGeometry ibgdBoundaryGrid(ibGrid *gg, ibgSegment outside);

/* reading polygons from a file and using them for geometry description */
ibGeometry ibgdPolygons(char* filename, ibgSegment outside,
		        ibgFloat *xmin, ibgFloat *xmax,
		        ibgFloat *ymin, ibgFloat *ymax);

#endif
