/* last edit: Ilja Schmelzer -------------- 18-OCT-1994 09:08:19.82	*/
/************************************************************************/
/*                                                                      */
/*  <<< I B G >>> - Intersection - Based Grid generation package 	*/
/*                                                                      */
/*  Version 1.1 by Ilja Schmelzer   schmelzer@iaas-berlin.d400.de       */
/*                                                                      */
/*  to be distributed under IBG license conditions (see "readme.ibg")	*/
/*                                                                      */
/************************************************************************/
/* <<< IBGDefault >>> - Intersection-Based Geometry - Default functions */
#include "ibgdefault.h"
#include "ibglib.h"
#include "ibgi.h"
#include "ibgd0.h"

ibgFloat ibgDMaxFace	= 0.1;	  /* distance for boundary test */
ibgFloat ibgDMaxLine	= 0.1;
ibgFloat ibgDMaxNode	= 0.1;
int	 ibgVBorderref	= 0;

int ibgDefaultFree(ibGeometry g)
{
 if(g != (ibGeometry)ibgNULL) free(g);
 return ibgSuccess;
}

/* bisection algorithm based on the Region-function: */
int ibgDefaultFaceWithEdge(ibGeometry g, ibgPoint *nint,
					ibgPoint *n1, ibgPoint *n2)
{ibgPoint c1,c2,cm,cf;
 ibgFloat delta=0;
 int m1=ibgpSegment(*n1), m2=ibgpSegment(*n2), d, rc;
 c1= *n1; c2= *n2;
 for(d=0;d<ibgDIM;d++){
	register ibgFloat dx = ibgpX(c1)[d] - ibgpX(c2)[d];
	delta += dx>0 ? dx : -dx;
 }
 while(delta>g->Delta){ delta /= 2;
     for(d=0;d<ibgDIM;d++){
	ibgpX(cm)[d] = 0.5*(ibgpX(c1)[d] + ibgpX(c2)[d]);
     }
     ibgiRegionOfPoint(g,&cm,&c1);
     if     (ibgpSegment(cm)==m1){
	if(delta>ibgDMaxFace){
	 	rc = ibgDefaultFaceWithEdge(g,nint,&c1,&cm);
		if(rc==ibgFaceFound){return ibgFaceFound;}
	 	rc = ibgDefaultFaceWithEdge(g,nint,&cm,&c2);
		return rc;
	}else if(m1==m2){
		ibgpType(*nint)	= ibgSNothing;
		return ibgNotFound;
	}
	c1 = cm;
     }else if(ibgpSegment(cm)==m2){
      	c2 = cm;
     }else             {
	c2 = cm;
	m2 = ibgpSegment(cm);
     }
 }   
 for(d=0;d<ibgDIM;d++){
	ibgpX(*nint)[d] = 0.5*(ibgpX(c1)[d] + ibgpX(c2)[d]);
 }
 ibgpSegment(*nint)=ibgDefaultFaceNr(m1,m2);
 ibgpType(*nint)=ibgSFace;
 return ibgFaceFound;
}

int ibgDefaultLineWithTriangle(ibGeometry g,
			ibgPoint *nint, ibgPoint *nface,
	      		ibgPoint *n1, ibgPoint *n2, ibgPoint *n3)
{ibgPoint n12,n23,n31,nold;
 int mf,m1,m2,m3,m4,b1,i,d,s,sn;
 ibgFloat dd,xm,xp;
 dd = 0;
#define max3(x1,x2,x3) x2>x3 ? x1>x2? x1:x2 : x1>x3? x1:x3
#define min3(x1,x2,x3) x2<x3 ? x1<x2? x1:x2 : x1<x3? x1:x3
 for(d=0;d<ibgDIM;d++) {
	xp = max3(ibgpX(*n1)[d],ibgpX(*n2)[d],ibgpX(*n3)[d]);
	xm = min3(ibgpX(*n1)[d],ibgpX(*n2)[d],ibgpX(*n3)[d]);
	dd += xp - xm;
 }
 if(dd < g->Delta) {
	for(d=0;d<ibgDIM;d++) {
		ibgpX(*nint)[d] = (ibgpX(*n1)[d]+ibgpX(*n2)[d]+ibgpX(*n3)[d])/3;
	}
	ibgiRegionOfPoint(g,nint,n1);
	ibgpSegment(*nint) = ibgDefaultLineNr;
	ibgpType(*nint) = ibgSLine;
	return ibgLineFound;
 }
 if(dd > ibgDMaxLine) goto linehalf;
 mf=ibgpSegment(*nface); m1=ibgpSegment(*n1); m2=ibgpSegment(*n2);
 if(ibgIncorrect!=ibgiStatusOfEdge(g,n1,nface)){
	if((m3=ibgpSegment(*n3))==m1){
		if(m2==m1) goto linehalf;
		ibgiFaceWithEdge(g,nint,n3,n2);
		if(ibgpSegment(*nint)==mf) return ibgiOnSide23;
		goto linehalf;
	}
	ibgiFaceWithEdge(g,nint,n1,n3);
	if(ibgpSegment(*nint)==mf) return ibgiOnSide31;
	goto linehalf;
 }else if(ibgIncorrect!=ibgiStatusOfEdge(g,n2,nface)){
 	if((m3=ibgpSegment(*n3))==m2){
		if(m1==m2) goto linehalf;
		ibgiFaceWithEdge(g,nint,n3,n1);
		if(ibgpSegment(*nint)==mf) return ibgiOnSide31;
		goto linehalf;
	}
	b1=ibgiFaceWithEdge(g,nint,n2,n3);
	if(ibgpSegment(*nint)==mf) return ibgiOnSide23;
	goto linehalf;
 }
linehalf:
 for(d=0;d<ibgDIM;d++) {
	ibgpX(n12)[d] = (ibgpX(*n1)[d]+ibgpX(*n2)[d])/2;
	ibgpX(n23)[d] = (ibgpX(*n2)[d]+ibgpX(*n3)[d])/2;
	ibgpX(n31)[d] = (ibgpX(*n3)[d]+ibgpX(*n1)[d])/2;
 }
 ibgiRegionOfPoint(g,&n12,n1);
 ibgiRegionOfPoint(g,&n23,n2);
 ibgiRegionOfPoint(g,&n31,n3);
 dd=0;
 for(d=0;d<ibgDIM;d++) {
	dd += (ibgpX(*nface)[d] - ibgpX(n12)[d])
	    * (ibgpX(   n12)[d] - ibgpX(*n1)[d]);
 }
 if(dd<0) s= -1; else s= -2;
 for(i=0;i<ibgdLSteps;i++){
	switch(s){
	case -1:
		sn=ibgDefaultLineWithTriangle(g,nint,nface,n1,&n12,&n31);
		switch(sn){
		case ibgiOnSide12:	s = ibgiOnSide12; break;
		case ibgiOnSide23:	s = -10; break;
		case ibgiOnSide31:	s = ibgiOnSide31; break;
		}break;
	case -2:
		sn=ibgDefaultLineWithTriangle(g,nint,nface,&n12,n2,&n23);
		switch(sn){
		case ibgiOnSide12:	s = ibgiOnSide12; break;
		case ibgiOnSide23:	s = ibgiOnSide23; break;
		case ibgiOnSide31:	s = -20; break;
		}break;
	case -23:
		sn=ibgDefaultLineWithTriangle(g,nint,&nold,&n12,&n31,n1);
		switch(sn){
		case ibgiOnSide12:	s = -10; break;
		case ibgiOnSide23:	s = ibgiOnSide31; break;
		case ibgiOnSide31:	s = ibgiOnSide12; break;
		}break;
	case -31:
		sn=ibgDefaultLineWithTriangle(g,nint,&nold,&n23,&n12,n2);
		switch(sn){
		case ibgiOnSide12:	s = -20; break;
		case ibgiOnSide23:	s = ibgiOnSide12; break;
		case ibgiOnSide31:	s = ibgiOnSide23; break;
		}break;
	case -12:
		sn=ibgDefaultLineWithTriangle(g,nint,&nold,&n31,&n23,n3);
		switch(sn){
		case ibgiOnSide12:	s = -30; break;
		case ibgiOnSide23:	s = ibgiOnSide23; break;
		case ibgiOnSide31:	s = ibgiOnSide31; break;
		}break;
	case -10:
		sn=ibgDefaultLineWithTriangle(g,nint,&nold,&n31,&n12,&n23);
		switch(sn){
		case ibgiOnSide12:	s = -23; break;
		case ibgiOnSide23:	s = -31; break;
		case ibgiOnSide31:	s = -12; break;
		}break;
	case -20:
		sn=ibgDefaultLineWithTriangle(g,nint,&nold,&n12,&n23,&n31);
		switch(sn){
		case ibgiOnSide12:	s = -31; break;
		case ibgiOnSide23:	s = -12; break;
		case ibgiOnSide31:	s = -23; break;
		}break;
	case -30:
		sn=ibgDefaultLineWithTriangle(g,nint,&nold,&n23,&n31,&n12);
		switch(sn){
		case ibgiOnSide12:	s = -12; break;
		case ibgiOnSide23:	s = -23; break;
		case ibgiOnSide31:	s = -31; break;
		}break;
	}
	if(sn == ibgLineFound){
		return ibgLineFound;
	}
  	if(s>0)	      		return s;
	nold = *nint;
 }
 ibgpSegment(*nint) = ibgDefaultLineNr; 
 ibgpType(*nint) = ibgSNothing;
 return ibgNotFound;
}

static int ibgpeq(ibGeometry g, ibgPoint *n1,ibgPoint *n2)
{ibgFloat *x1,*x2;int d;
 if(ibgpSegment(*n1)!=ibgpSegment(*n2)) return 1;
 x1=ibgpX(*n1); x2=ibgpX(*n2);
 for(d=0;d<ibgDIM;d++){register x;
	if((x = x1[d]-x2[d]) >  g->Delta) return 1;
	if((x              ) < -g->Delta) return 1;
 }
 return 0;
}

static int snnr[4][3]	=	{{1,2,3},{4,3,2},{1,3,4},{4,2,1}};
static int ssnr[4][3]	=	{{1,2,3},{0,3,2},{1,3,0},{0,2,1}};
static int snr[4]	=	{ibgiOnSide123,ibgiOnSide234,ibgiOnSide134,ibgiOnSide124};

int  ibgDefaultNodeInTetrahedron(ibGeometry g,
			ibgPoint *nint, ibgPoint *nline,
	  		ibgPoint *n1, ibgPoint *n2, ibgPoint *n3, ibgPoint *n4)
{
#define MAXPOINTS 6
#define MAXTRIANGLES 16
 ibgPoint point[MAXPOINTS],*nn1,*nn2,*n;
 ibgPoint *(sn[MAXTRIANGLES][3]);
 int	  (ss[MAXTRIANGLES][3]);
 int	  (sr[MAXTRIANGLES]);
 ibgPoint *(nn[5]);
 ibgPoint *n12,*n13,*n14,*n23,*n24,*n34,nold,nface;
 int	 d,i,sb[3],s0,t0,t1,t2,i0,i1,i2,st;
 ibgSegment m00,m01,m10,m11,ul;
 int	ibgNodeBorderRef0=ibgVBorderref;
 ibgFloat dd,d1,d2,d3,xm,xp;
 ibgFloat x0[ibgDIM],x1[ibgDIM],x2[ibgDIM],x3[ibgDIM];
 int first=1,slast,nlast,sh,s,rc;
/* Initialization */
/* diameter computation */
 dd = 0;
#define max4(x1,x2,x3,x4) x3>x4? x2>x3 ? x1>x2? x1:x2 : x1>x3? x1:x3:\
				 x2>x4 ? x1>x2? x1:x2 : x1>x4? x1:x4
#define min4(x1,x2,x3,x4) x3<x4? x2<x3 ? x1<x2? x1:x2 : x1<x3? x1:x3:\
				 x2<x4 ? x1<x2? x1:x2 : x1<x4? x1:x4
 for(d=0;d<ibgDIM;d++) {
	xp = max4(ibgpX(*n1)[d],ibgpX(*n2)[d],ibgpX(*n3)[d],ibgpX(*n4)[d]);
	xm = min4(ibgpX(*n1)[d],ibgpX(*n2)[d],ibgpX(*n3)[d],ibgpX(*n4)[d]);
	dd += xp - xm;
 }
 if(dd < g->Delta) {
	for(d=0;d<ibgDIM;d++) {
		ibgpX(*nint)[d] = (ibgpX(*n1)[d]+ibgpX(*n2)[d]
				+  ibgpX(*n3)[d]+ibgpX(*n4)[d])/4;
	}
	ibgiRegionOfPoint(g,nint,n1);
	ibgpSegment(*nint) = ibgDefaultLineNr;
 	ibgpType(*nint) = ibgSNode;
	return ibgNodeFound;
 }
 nn[1] = n1; nn[2] = n2; nn[3] = n3; nn[4] = n4;
 for(s0=0;s0<4;s0++){
	sr[s0]	 = snr[s0];
      	sn[s0][0] = nn[snnr[s0][0]];
    	sn[s0][1] = nn[snnr[s0][1]];
	sn[s0][2] = nn[snnr[s0][2]];
	ss[s0][0] = ssnr[s0][0];
	ss[s0][1] = ssnr[s0][1];
	ss[s0][2] = ssnr[s0][2];
 }
 slast = 4; nlast = 0;
nodebordertest:
/* weitere Verfeinerung des Randes */
 sb[0] = slast;
 sb[1] = slast + slast;
 sb[2] = slast + slast + slast;
 for(s0=0;s0<slast;s0++){
	sr[sb[0]+s0] = sr[sb[1]+s0] = sr[sb[2]+s0] = sr[s0];
	sn[sb[0]+s0][0] = sn[s0][0];
	sn[sb[1]+s0][1] = sn[s0][1];
	sn[sb[2]+s0][2] = sn[s0][2];
  	for(t0=0;t0<3;t0++){
		t1=(t0+1)%3; t2=(t0+2)%3;
	    	if((st=ss[s0][t0])>s0){n = &point[nlast++];
			for(d=0;d<ibgDIM;d++)
				ibgpX(*n)[d] = (ibgpX(*(sn[s0][t1]))[d]
					        +  ibgpX(*(sn[s0][t2]))[d])/2;
			ibgiRegionOfPoint(g,n,sn[s0][t1]);
		}else{
			for(i0=0;i0<3;i0++) if(ss[st][i0]==s0) break;
	       		n = sn[st][i0];
			i1=(i0+1)%3; i2=(i0+2)%3;
		      	ss[sb[t1]+s0][t0] = sb[i2]+st;
			ss[sb[t2]+s0][t0] = sb[i1]+st;
		      	ss[sb[i1]+st][t0] = sb[t2]+s0;
		      	ss[sb[i2]+st][t0] = sb[t1]+s0;
		}
	      	sn[s0][t0] = sn[sb[t1]+s0][t2] = sn[sb[t2]+s0][t1] = n;
	}
  	ss[sb[0]+s0][0] = s0;
	ss[sb[1]+s0][1] = s0;
	ss[sb[2]+s0][2] = s0;
 }
 for(s0=0;s0<slast;s0++){
  	ss[s0][0] = sb[0]+s0;
  	ss[s0][1] = sb[1]+s0;
  	ss[s0][2] = sb[2]+s0;
 }
 slast *= 4;
 if(first){first=0;
/* Unterteilung des Wuerfels */
	n23 = sn[0][0];
	n13 = sn[0][1];
	n12 = sn[0][2];
	n24 = sn[1][1];
	n34 = sn[1][2];
	n14 = sn[2][1];
 }
 if(dd>ibgDMaxNode) goto nodehalf;

 /* Durchlauf durch die Kanten zum Finden von Schnitten */
 ul = ibgpSegment(*nline);
 for(s0=0;s0<slast;s0++){
  	for(t0=0;t0<3;t0++){
	    	if((st=ss[s0][t0])>s0) continue;
		t1=(t0+1)%3; t2=(t0+2)%3;
		nn1 = sn[s0][t1];
		nn2 = sn[s0][t2];
		if(ibgpSegment(*n1)==ibgpSegment(*n2)) continue;
		rc = ibgiFaceWithEdge(g,&nface,nn1,nn2);
		rc = ibgiLineWithTriangle(g,nint,&nface,nn1,nn2,sn[s0][t0]);
		if(ul==ibgpSegment(*nint)){
			if(ibgpeq(g,nint,nline)) return sr[s0];
		}
		for(i0=0;i0<3;i0++) if(ss[st][i0]==s0) break;
		rc = ibgiLineWithTriangle(g,nint,&nface,nn2,nn1,sn[st][i0]);
		if(ul==ibgpSegment(*nint)){
			if(ibgpeq(g,nint,nline)) return sr[s0];
		}
	}
 }
/* refine the tetrahedron: */
nodehalf:
 if(ibgNodeBorderRef0){
	ibgNodeBorderRef0--; goto nodebordertest;
 }
 for(d=0;d<ibgDIM;d++){
	x1[d] = ibgpX(*n23)[d]   - ibgpX(*n4)[d];
	x2[d] = ibgpX(*n13)[d]   - ibgpX(*n4)[d];
	x3[d] = ibgpX(*n12)[d]   - ibgpX(*n4)[d];
	x0[d] = ibgpX(*nline)[d] - ibgpX(*n4)[d];
 }
 d1 = x0[0]*(x2[1]*x3[2]-x3[1]*x2[2])
    + x0[1]*(x2[2]*x3[0]-x3[2]*x2[0])
    + x0[2]*(x2[0]*x3[1]-x3[0]*x2[1]);
 d2 = x0[0]*(x3[1]*x1[2]-x1[1]*x3[2])
    + x0[1]*(x3[2]*x1[0]-x1[2]*x3[0])
    + x0[2]*(x3[0]*x1[1]-x1[0]*x3[1]);
 d3 = x0[0]*(x1[1]*x2[2]-x1[1]*x2[2])
    + x0[1]*(x1[2]*x2[0]-x1[2]*x2[0])
    + x0[2]*(x1[0]*x2[1]-x1[0]*x2[1]);
 if(d1>0)	s = -100;
 else if(d2>0)	s = -200;
 else if(d3>0)	s = -300;
 else 		s = -400;
 for(i=0;i<ibgdVSteps;i++){
	switch(s){
	case -100:
		sh=ibgDefaultNodeInTetrahedron(g,nint,nline,n1,n12,n13,n14);
		switch(sh){
			case ibgiOnSide123:	s = ibgiOnSide123;  break;
			case ibgiOnSide124:	s = ibgiOnSide124;  break;
			case ibgiOnSide134:	s = ibgiOnSide134;  break;
			case ibgiOnSide234:	s = -91;  break;
		}break;
	case -200:
		sh=ibgDefaultNodeInTetrahedron(g,nint,nline,n12,n2,n23,n24);
		switch(sh){
			case ibgiOnSide123:	s = ibgiOnSide123;  break;
			case ibgiOnSide124:	s = ibgiOnSide124;  break;
			case ibgiOnSide134:	s = -92;  break;
			case ibgiOnSide234:	s = ibgiOnSide234;  break;
		}break;
	case -300:
		sh=ibgDefaultNodeInTetrahedron(g,nint,nline,n13,n23,n3,n34);
		switch(sh){
			case ibgiOnSide123:	s = ibgiOnSide123;  break;
			case ibgiOnSide124:	s = -93;  break;
			case ibgiOnSide134:	s = ibgiOnSide134;  break;
			case ibgiOnSide234:	s = ibgiOnSide234;  break;
		}break;
	case -400:
		sh=ibgDefaultNodeInTetrahedron(g,nint,nline,n23,n13,n12,n24);
		switch(sh){
			case ibgiOnSide123:	s = ibgiOnSide123;  break;
			case ibgiOnSide124:	s = -23;  break;
			case ibgiOnSide134:	s = -20;  break;
			case ibgiOnSide234:	s = -21;  break;
		}break;
	case -19:
		sh=ibgDefaultNodeInTetrahedron(g,nint,&nold,n14,n13,n12,n1);
		switch(sh){
			case ibgiOnSide123:	s = -91;  break;
			case ibgiOnSide124:	s = ibgiOnSide134;  break;
			case ibgiOnSide134:	s = ibgiOnSide124;  break;
			case ibgiOnSide234:	s = ibgiOnSide123;  break;
		}break;
	case -29:
		sh=ibgDefaultNodeInTetrahedron(g,nint,&nold,n12,n23,n24,n2);
		switch(sh){
			case ibgiOnSide123:	s = -92;  break;
			case ibgiOnSide124:	s = ibgiOnSide123;  break;
			case ibgiOnSide134:	s = ibgiOnSide124;  break;
			case ibgiOnSide234:	s = ibgiOnSide234;  break;
		}break;
	case -39:
		sh=ibgDefaultNodeInTetrahedron(g,nint,&nold,n34,n23,n13,n3);
		switch(sh){
			case ibgiOnSide123:	s = -93;  break;
			case ibgiOnSide124:	s = ibgiOnSide234;  break;
			case ibgiOnSide134:	s = ibgiOnSide134;  break;
			case ibgiOnSide234:	s = ibgiOnSide123;  break;
		}break;
	case -49:
		sh=ibgDefaultNodeInTetrahedron(g,nint,&nold,n14,n24,n34,n4);
		switch(sh){
			case ibgiOnSide123:	s = -94;  break;
			case ibgiOnSide124:	s = ibgiOnSide124;  break;
			case ibgiOnSide134:	s = ibgiOnSide134;  break;
			case ibgiOnSide234:	s = ibgiOnSide234;  break;
		}break;
	case -91:
		sh=ibgDefaultNodeInTetrahedron(g,nint,&nold,n12,n13,n14,n24);
		switch(sh){
			case ibgiOnSide123:	s = -19;  break;
			case ibgiOnSide124:	s = -12;  break;
			case ibgiOnSide134:	s = ibgiOnSide124;  break;
			case ibgiOnSide234:	s = -14;  break;
		}break;
	case -21:
		sh=ibgDefaultNodeInTetrahedron(g,nint,&nold,n24,n13,n12,n14);
		switch(sh){
			case ibgiOnSide123:	s = -12;  break;
			case ibgiOnSide124:	s = -14;  break;
			case ibgiOnSide134:	s = ibgiOnSide124;  break;
			case ibgiOnSide234:	s = -19;  break;
		}break;
	case -41:
		sh=ibgDefaultNodeInTetrahedron(g,nint,&nold,n24,n14,n13,n12);
		switch(sh){
			case ibgiOnSide123:	s = -14;  break;
			case ibgiOnSide124:	s = ibgiOnSide124;  break;
			case ibgiOnSide134:	s = -12;  break;
			case ibgiOnSide234:	s = -19;  break;
		}break;
	case -92:
		sh=ibgDefaultNodeInTetrahedron(g,nint,&nold,n24,n23,n12,n13);
		switch(sh){
			case ibgiOnSide123:	s = -29;  break;
			case ibgiOnSide124:	s = -23;  break;
			case ibgiOnSide134:	s = -21;  break;
			case ibgiOnSide234:	s = ibgiOnSide123;  break;
		}break;
	case -12:
		sh=ibgDefaultNodeInTetrahedron(g,nint,&nold,n24,n13,n12,n23);
		switch(sh){
			case ibgiOnSide123:	s = -21;  break;
			case ibgiOnSide124:	s = -23;  break;
			case ibgiOnSide134:	s = -29;  break;
			case ibgiOnSide234:	s = ibgiOnSide123;  break;
		}break;
	case -32:
		sh=ibgDefaultNodeInTetrahedron(g,nint,&nold,n13,n23,n24,n12);
		switch(sh){
			case ibgiOnSide123:	s = -23;  break;
			case ibgiOnSide124:	s = ibgiOnSide123;  break;
			case ibgiOnSide134:	s = -21;  break;
			case ibgiOnSide234:	s = -29;  break;
		}break;
	case -93:
		sh=ibgDefaultNodeInTetrahedron(g,nint,&nold,n13,n23,n34,n24);
		switch(sh){
			case ibgiOnSide123:	s = -39;  break;
			case ibgiOnSide124:	s = -32;  break;
			case ibgiOnSide134:	s = -34;  break;
			case ibgiOnSide234:	s = ibgiOnSide234;  break;
		}break;
	case -23:
		sh=ibgDefaultNodeInTetrahedron(g,nint,&nold,n24,n23,n13,n34);
		switch(sh){
			case ibgiOnSide123:	s = -32;  break;
			case ibgiOnSide124:	s = ibgiOnSide234;  break;
			case ibgiOnSide134:	s = -34;  break;
			case ibgiOnSide234:	s = -39;  break;
		}break;
	case -43:
		sh=ibgDefaultNodeInTetrahedron(g,nint,&nold,n34,n24,n13,n23);
		switch(sh){
			case ibgiOnSide123:	s = -34;  break;
			case ibgiOnSide124:	s = ibgiOnSide234;  break;
			case ibgiOnSide134:	s = -39;  break;
			case ibgiOnSide234:	s = -32;  break;
		}break;
	case -94:
		sh=ibgDefaultNodeInTetrahedron(g,nint,&nold,n34,n24,n14,n13);
		switch(sh){
			case ibgiOnSide123:	s = -49;  break;
			case ibgiOnSide124:	s = -43;  break;
			case ibgiOnSide134:	s = ibgiOnSide134;  break;
			case ibgiOnSide234:	s = -41;  break;
		}break;
	case -14:
		sh=ibgDefaultNodeInTetrahedron(g,nint,&nold,n13,n14,n24,n34);
		switch(sh){
			case ibgiOnSide123:	s = -41;  break;
			case ibgiOnSide124:	s = ibgiOnSide134;  break;
			case ibgiOnSide134:	s = -43;  break;
			case ibgiOnSide234:	s = -49;  break;
		}break;
	case -34:
		sh=ibgDefaultNodeInTetrahedron(g,nint,&nold,n13,n24,n34,n14);
		switch(sh){
			case ibgiOnSide123:	s = -43;  break;
			case ibgiOnSide124:	s = -41;  break;
			case ibgiOnSide134:	s = ibgiOnSide134;  break;
			case ibgiOnSide234:	s = -49;  break;
		}break;
	}
	if(sh == ibgNodeFound){
		return ibgNodeFound;
  	}else if(s > 0){
	      	return s;
	}
	nold = *nint;
 }
 ibgpSegment(*nint) = ibgDefaultNodeNr; 
 ibgpType(*nint) = ibgSNothing;
 return ibgNotFound;
}
