/* last edit: Ilja Schmelzer -------------- 24-MAR-1995 17:51:53.58	*/
/************************************************************************/
/*                                                                      */
/*  <<< I B G >>> - Intersection - Based Grid generation package 	*/
/*                                                                      */
/*  Version 1.1 by Ilja Schmelzer   schmelzer@iaas-berlin.d400.de       */
/*                                                                      */
/*  to be distributed under IBG license conditions (see "readme.ibg")	*/
/*                                                                      */
/************************************************************************/
/* <<< IBGDSplitRegion >>> - Intersection-Based Geometry Definition
			      	Splitting a Region
*/
#define const
#include "ibglib.h"
#include "ibgd.h"
#include "ibgi.h"
#include "ibglib.h"
#include "ibgd0.h"
#include "ibgt.h"
#include "ibgdefault.h"

typedef struct _ibgSplitRegion{
	ibGeometryRec	g;
	ibGeometry	gold;
	ibgTopology	told;
	ibgSegment		rold,rnew,face;
	ibgFloat	(*func)(ibgPtObject fThis, ibgPoint *x);
	ibgPtObject	fThis;
	ibgFloat	epsilon;
}ibgSplitRegionRec,*ibgSplitRegion;

static int RegionOfPoint(ibGeometry geom, ibgPoint *nnew, const ibgPoint *nold)
{ibgSplitRegion splt= (ibgSplitRegion)geom;
 ibgiRegionOfPoint(splt->gold,nnew,nold);
 if( ibgpSegment(*nnew) == splt->rold)
 	if(splt->func(splt->fThis,nnew) > 0){
		ibgpSegment(*nnew) = splt->rnew;
	}
 return ibgRegionFound;
}

static ibgSegment splitChangeFaceNr(ibgSplitRegion splt,ibgSegment u)
{register int u0,u1;
 if((u0=u/ibgSMAX) == 0)	return u;
 u1 = u%ibgSMAX;
 if(splt->rold == u0){
 	if(splt->rnew > u1)	return ibgSMAX*splt->rnew+u1;
 	if(splt->rnew ==u1)	ibgfatal;
 	else			return ibgSMAX*u1+splt->rnew;
 }
 if(splt->rold == u1){
 	if(splt->rnew > u0)	return ibgSMAX*splt->rnew+u0;
 	if(splt->rnew ==u0)	ibgfatal;
 	else			return ibgSMAX*u0+splt->rnew;
 }
 ibgfatal;
 return u;
}

static ibgSegment splitCreateLineNr(ibgSplitRegion splt,ibgSegment u)
{int u0,u1;
 if((u0=u/ibgSMAX) == 0){
	ibgassert(ibgIncorrect !=ibgtStatusRF(&splt->g.top,splt->rold,u));
	if(ibgCorrect != ibgtStatusRF(&splt->g.top,splt->rnew,u)){
	      	return ibgDefaultLineNr2(splt->rold,splt->rnew);
	}else{
	      	return u;
	}
 }
 u1 = u%ibgSMAX;
 ibgassert((splt->rold == u0)||(splt->rold == u1));
 if(splt->rnew > u0) return ibgSMAX*(ibgSMAX*splt->rnew+u0)+u1;
 if(splt->rnew ==u0) return u;
 if(splt->rnew > u1) return ibgSMAX*(ibgSMAX*u0+splt->rnew)+u1;
 if(splt->rnew ==u1) return u;
 else                return ibgSMAX*(ibgSMAX*u0+u1)+splt->rnew;
}

static ibgSegment splitChangeLineNr(ibgSplitRegion splt,ibgSegment u)
{int u0,u1,u2,u3;
 if((u0=u/ibgSMAX) == 0)	return u;
 u1 = u0/ibgSMAX;
 if(u1 == 0){
	u1 = u %ibgSMAX;
	if(splt->rold == u0){
		if(splt->rnew > u1)	return ibgSMAX*splt->rnew+u1;
		if(splt->rnew ==u1)	ibgfatal;
		else			return ibgSMAX*u1+splt->rnew;
	}
	if(splt->rold == u1){
		if(splt->rnew > u0)	return ibgSMAX*splt->rnew+u0;
		if(splt->rnew ==u0)	ibgfatal;
		else			return ibgSMAX*u0+splt->rnew;
	}
 }
 u2 = u0%ibgSMAX;
 u3 = u %ibgSMAX;
 if(splt->rold == u1){
	if(splt->rnew > u2) return ibgSMAX*(ibgSMAX*splt->rnew+u2)+u3;
	if(splt->rnew ==u2) return (ibgSMAX*u2)+u3;
	if(splt->rnew > u3) return ibgSMAX*(ibgSMAX*u2+splt->rnew)+u3;
	if(splt->rnew ==u3) return (ibgSMAX*u2)+u3;
	else		   return ibgSMAX*(ibgSMAX*u2+u3)+splt->rnew;
 }
 if(splt->rold == u2){
	if(splt->rnew > u1) return ibgSMAX*(ibgSMAX*splt->rnew+u1)+u3;
	if(splt->rnew ==u1) return (ibgSMAX*u1)+u3;
	if(splt->rnew > u3) return ibgSMAX*(ibgSMAX*u1+splt->rnew)+u3;
	if(splt->rnew ==u3) return (ibgSMAX*u1)+u3;
	else		   return ibgSMAX*(ibgSMAX*u1+u3)+splt->rnew;
 }
 if(splt->rold == u3){
 	if(splt->rnew > u1) return ibgSMAX*(ibgSMAX*splt->rnew+u1)+u2;
	if(splt->rnew ==u1) return (ibgSMAX*u1)+u2;
	if(splt->rnew > u2) return ibgSMAX*(ibgSMAX*u1+splt->rnew)+u2;
	if(splt->rnew ==u2) return (ibgSMAX*u1)+u2;
	else		   return ibgSMAX*(ibgSMAX*u1+u2)+splt->rnew;
 }
 ibgfatal;
 return u;
}

static int SplitRegionZero(ibgPoint *nh, ibgPoint *n1, ibgPoint *n2,
			   	ibgSplitRegion splt)
{double ff,f1,f2,delta; ibgPoint c1,c2; int d,ic,dc;
 f1=(*splt->func)(splt->fThis,n1);
 f2=(*splt->func)(splt->fThis,n2);
 if(f1*f2>0)			{*nh= *n1;	return ibgError;}
 if(f1==0)			{*nh= *n1;	return ibgFaceFound;}
 if(f2==0)			{*nh= *n2;	return ibgFaceFound;}
 c1= *n1; c2= *n2; ic=0; dc=1;
 do{
     if(dc){dc=0; delta=0;
	for(d=0;d<ibgDIM;d++){
		register double dn	= (ibgpX(c1)[d] - ibgpX(c2)[d]);
		delta 			+= dn>0 ? dn : -dn;
		ibgpX(*nh)[d]		= (f2*ibgpX(c1)[d]
					-  f1*ibgpX(c2)[d]) / (f2-f1);
	}
     }else{ dc=1; delta=0;
	for(d=0;d<ibgDIM;d++){
		register double dn 	= (ibgpX(c1)[d] - ibgpX(c2)[d]);
		delta 			+= dn>0 ? dn : -dn;
		ibgpX(*nh)[d] 		= 0.5*(ibgpX(c1)[d] + ibgpX(c2)[d]);
	}
     }
     ibgiRegionOfPoint(splt->gold,nh,&c1);
     if(delta < splt->g.Delta)		return ibgFaceFound;
     if(ibgpSegment(*nh) != splt->rold)	return ibgNotFound; /* another region */
     ff=(splt->func)(splt->fThis,nh);
     if (ff > splt->epsilon)			ff += splt->epsilon;
     else if (ff < -splt->epsilon)		ff -= splt->epsilon;
     else				return ibgFaceFound;
     if(ff*f1>0){f1=ff;for(d=0;d<ibgDIM;d++) ibgpX(c1)[d]=ibgpX(*nh)[d];}
     else       {f2=ff;for(d=0;d<ibgDIM;d++) ibgpX(c2)[d]=ibgpX(*nh)[d];}
 }while(1);				return ibgFaceFound;
}

static int FaceWithEdge(ibGeometry This, ibgPoint *nint,
			const ibgPoint *n1, const ibgPoint *n2)
{ibgSplitRegion splt = ((ibgSplitRegion)This);
 ibgPoint nh,nb; int i,rc;
 if(ibgpSegment(*n1)==splt->rold){
	if(ibgpSegment(*n2)==splt->rnew){
		nb = *n2;
		ibgiRegionOfPoint(splt->gold,&nb,n1);
	       	if(splt->rold==ibgpSegment(nb))
			{ibgpSegment(nb)=splt->rnew;
			 rc=SplitRegionZero(&nh,n1,&nb,splt);}
		else	{rc=ibgNotFound; nh= *n2;}
	}else		{rc=ibgNotFound; nh= *n2;}
	while(rc==ibgNotFound){
		rc=ibgiFaceWithEdge(splt->gold,nint,n1,&nh);
		if(rc==ibgNotFound) return ibgNotFound;
		if((splt->func)(splt->fThis,nint)<=0)	return rc;
		rc=SplitRegionZero(&nh,n1,nint,splt);
	}
	*nint = nh;
	ibgpType(*nint) = ibgSFace;
	ibgpSegment(*nint) = splt->face;
	return ibgFaceFound;
 }else if(ibgpSegment(*n1)==splt->rnew){
	nh = *n1; rc = ibgiRegionOfPoint(splt->gold,&nh,n1);
 	if((rc==ibgError) || (splt->rnew != ibgpSegment(nh))){
		nb = nh;
		if(ibgpSegment(*n2) == splt->rold){
			rc = SplitRegionZero(&nh,&nb,n2,splt);
		}else{	nh = *n2; rc = ibgiRegionOfPoint(splt->gold,&nh,n2);
		 	if((rc==ibgError)) ibgpSegment(nh)=splt->rold;
			rc = ibgNotFound;
		}
	}else{
		rc = ibgiFaceWithEdge(splt->gold,nint,n1,n2);
		if(rc==ibgNotFound) return ibgNotFound;
		ibgassert(ibgIncorrect!=ibgtStatusRF(splt->told,splt->rnew,
							      ibgpSegment(*nint)));
		if(ibgIncorrect==ibgtStatusRF(splt->told,splt->rold,ibgpSegment(*nint)))
						return ibgFaceFound;
		if((splt->func)(splt->fThis,nint)<=0)	return ibgFaceFound;
		nb = *nint;
  		if(ibgpSegment(*n2)==splt->rold){
		 	rc = SplitRegionZero(&nh,&nb,n2,splt);
		}else{	rc = ibgNotFound; nh = *n2;
		}
	}
	while(rc==ibgNotFound){
		rc = ibgiFaceWithEdge(splt->gold,nint,&nb,&nh);
		if(rc==ibgNotFound) return ibgNotFound;
		if((splt->func)(splt->fThis,nint)>0){
		    if(ibgCorrect==ibgtStatusRF(splt->told,splt->rold
						,ibgpSegment(*nint))){
		    	ibgpSegment(*nint)=splitChangeFaceNr(splt,ibgpSegment(*nint));
		    }
		    return  ibgFaceFound;
		}
		rc=SplitRegionZero(&nh,&nb,nint,splt);
	}
	*nint = nh;
	ibgpType(*nint) = ibgSFace;
	ibgpSegment(*nint) = splt->face;
	return ibgFaceFound;
 }
 else if(ibgpSegment(*n2)==splt->rnew){
	nh = *n2; rc = ibgiRegionOfPoint(splt->gold,&nh,n2);
 	if(rc == ibgError) ibgpSegment(nh) = splt->rold;
      	rc = ibgiFaceWithEdge(splt->gold,nint,n1,&nh);
 }
 else	rc = ibgiFaceWithEdge(splt->gold,nint,n1,n2);
 if(rc==ibgNotFound) return ibgNotFound;

 if(ibgCorrect==ibgtStatusRF(splt->told,splt->rold,ibgpSegment(*nint))){
      	if((splt->func)(splt->fThis,nint)>0){
  		ibgpSegment(*nint) = splitChangeFaceNr(splt,ibgpSegment(*nint));
	}
 }
 return  ibgFaceFound;
}


#define ibgSplitFound	4567

static int SplitRegionLine0(ibgSplitRegion splt,
		      	ibgPoint *nint, ibgPoint *nface,
			ibgPoint *n1, ibgPoint *n2, ibgPoint *n3)
{ibgPoint n12,n23,n31,nold;
 ibGeometry g=splt->gold;
 int i,d,s,sn,rc;
 ibgFloat fi,f0,f1,f2,f3,p12,p23,p31,q12,q23,q31,dd,s0;
 f0 = (splt->func)(splt->fThis,nface);
 s0 = f0>0 ? 1 : -1; f0 *= s0;
 rc = ibgiLineWithTriangle(splt->gold,nint,nface,n1,n2,n3);
 fi = s0*(splt->func)(splt->fThis,nint);
 if(fi >= splt->epsilon)	return rc;
 if(rc == ibgLineFound)	return rc;
 if(fi >= -splt->epsilon)return ibgSplitFound;
 f1 = (splt->func)(splt->fThis,n1);
 f2 = (splt->func)(splt->fThis,n2);
 f3 = (splt->func)(splt->fThis,n3);
 if(f1*f2>=0){
	p12 = q12 = 0.5;
 }else{	p12 = f2/(f2-f1);
	q12 = f1/(f1-f2);
 }
 if(p12<0.1)	{p12=0.1; q12=0.9;}
 if(q12<0.1)	{q12=0.1; p12=0.9;}
 if(f2*f3>=0){
	p23 = q23 = 0.5;
 }else{	p23 = f3/(f3-f2);
	q23 = f2/(f2-f3);
 }
 if(p23<0.1)	{p23=0.1; q23=0.9;}
 if(q23<0.1)	{q23=0.1; p23=0.9;}
 if(f1*f3>=0){
	p31 = q31 = 0.5;
 }else{	p31 = f1/(f1-f3);
	q31 = f3/(f3-f1);
 }
 if(p31<0.1)	{p31=0.1; q31=0.9;}
 if(q31<0.1)	{q31=0.1; p31=0.9;}
 for(d=0;d<ibgDIM;d++){
	ibgpX(n12)[d]=(p12*ibgpX(*n1)[d]+q12*ibgpX(*n2)[d]);
	ibgpX(n23)[d]=(p23*ibgpX(*n2)[d]+q23*ibgpX(*n3)[d]);
	ibgpX(n31)[d]=(p31*ibgpX(*n3)[d]+q31*ibgpX(*n1)[d]);
 }
 ibgiRegionOfPoint(g,&n12,n1);
 ibgiRegionOfPoint(g,&n23,n2);
 ibgiRegionOfPoint(g,&n31,n3);
 dd=0;
 for(d=0;d<ibgDIM;d++) {
	dd += (ibgpX(*nface)[d] - ibgpX(n12)[d])
	    * (ibgpX(   n12)[d] - ibgpX(*n1)[d]);
 }
 if(dd<0) s= -1; else s= -2;
 for(i=0;i<ibgdLSteps;i++){
	switch(s){
	case -1:
		sn=SplitRegionLine0(splt,nint,nface,n1,&n12,&n31);
		switch(sn){
		case ibgiOnSide12:	s = ibgiOnSide12; break;
		case ibgiOnSide23:	s = -10; break;
		case ibgiOnSide31:	s = ibgiOnSide31; break;
		}break;
	case -2:
		sn=SplitRegionLine0(splt,nint,nface,&n12,n2,&n23);
		switch(sn){
		case ibgiOnSide12:	s = ibgiOnSide12; break;
		case ibgiOnSide23:	s = ibgiOnSide23; break;
		case ibgiOnSide31:	s = -20; break;
		}break;
	case -23:
		sn=SplitRegionLine0(splt,nint,&nold,&n12,&n31,n1);
		switch(sn){
		case ibgiOnSide12:	s = -10; break;
		case ibgiOnSide23:	s = ibgiOnSide31; break;
		case ibgiOnSide31:	s = ibgiOnSide12; break;
		}break;
	case -31:
		sn=SplitRegionLine0(splt,nint,&nold,&n23,&n12,n2);
		switch(sn){
		case ibgiOnSide12:	s = -20; break;
		case ibgiOnSide23:	s = ibgiOnSide12; break;
		case ibgiOnSide31:	s = ibgiOnSide23; break;
		}break;
	case -12:
		sn=SplitRegionLine0(splt,nint,&nold,&n31,&n23,n3);
		switch(sn){
		case ibgiOnSide12:	s = -30; break;
		case ibgiOnSide23:	s = ibgiOnSide23; break;
		case ibgiOnSide31:	s = ibgiOnSide31; break;
		}break;
	case -10:
		sn=SplitRegionLine0(splt,nint,&nold,&n31,&n12,&n23);
		switch(sn){
		case ibgiOnSide12:	s = -23; break;
		case ibgiOnSide23:	s = -31; break;
		case ibgiOnSide31:	s = -12; break;
		}break;
	case -20:
		sn=SplitRegionLine0(splt,nint,&nold,&n12,&n23,&n31);
		switch(sn){
		case ibgiOnSide12:	s = -31; break;
		case ibgiOnSide23:	s = -12; break;
		case ibgiOnSide31:	s = -23; break;
		}break;
	case -30:
		sn=SplitRegionLine0(splt,nint,&nold,&n23,&n31,&n12);
		switch(sn){
		case ibgiOnSide12:	s = -12; break;
		case ibgiOnSide23:	s = -23; break;
		case ibgiOnSide31:	s = -31; break;
		}break;
	}
	if(sn== ibgLineFound){
		return ibgLineFound;
	}else if(sn== ibgSplitFound){
		return ibgSplitFound;
  	}else if(s>0){
	      	return s;
	}
	nold = *nint;
 }
 ibgpSegment(*nint) = ibgDefaultLineNr;
 ibgpType(*nint) = ibgSNothing;
 ibgfatal;	/* besser man weiss es */
 return ibgNotFound;
}

/* nbase is here a region point over nface (in rold or in old rnew-part) */
static int SplitRegionLine1(ibgSplitRegion splt,
	   	      	ibgPoint *nint, ibgPoint *nface, ibgPoint *nbase,
			ibgPoint *n1, ibgPoint *n2, ibgPoint *n3)
{int rc,s12,s23,s31;
 ibgPoint nh;
 rc = ibgiFaceWithEdge(splt->gold,&nh,nbase,n2);
 if(rc==ibgNotFound){ rc = ibgiFaceWithEdge(splt->gold,&nh,n2,n3);
  if(rc==ibgNotFound){ rc = ibgiFaceWithEdge(splt->gold,&nh,n3,n1);
   if(rc==ibgNotFound){ rc = ibgiFaceWithEdge(splt->gold,&nh,n1,nbase);
    if(rc==ibgNotFound){;
    }else{
	rc = SplitRegionLine0(splt,nint,&nh,n1,n2,n3);
   	s12 = ibgiOnSide12; s23 = ibgiOnSide23; s31 = ibgiOnSide31;
    }
   }else{
	rc = SplitRegionLine0(splt,nint,&nh,n3,n1,n2);
   	s12 = ibgiOnSide31; s23 = ibgiOnSide12; s31 = ibgiOnSide23;
   }
  }else{
	rc = SplitRegionLine0(splt,nint,&nh,n2,n3,n1);
   	s12 = ibgiOnSide23; s23 = ibgiOnSide31; s31 = ibgiOnSide12;
  }
 }else{
	rc = SplitRegionLine0(splt,nint,&nh,n1,n2,n3);
   	s12 = ibgiOnSide12; s23 = ibgiOnSide23; s31 = ibgiOnSide31;
 }
 switch(rc){
 case ibgiOnSide12:	return s12;
 case ibgiOnSide23:	return s23;
 case ibgiOnSide31:	return s31;
 default:		return rc;
 }
}

static int LineWithTriangle(ibGeometry This, ibgPoint *nint, ibgPoint *nface,
		const ibgPoint *n1, const ibgPoint *n2, const ibgPoint *n3)
{ibgPoint nh1,nh2,nh3,nhf,nhb,nh;
 ibgSplitRegion splt = (ibgSplitRegion) This;
 int rc,d;
 ibgSegment fu;
 ibgFloat dd,dr,dx;
 nh1 = *n1; nh2 = *n2; nh3 = *n3; nhf = *nface;
 if(ibgpSegment(*n1) == splt->rnew){
	rc=ibgiRegionOfPoint(splt->gold,&nh1,n1);
	if(rc==ibgError) ibgpSegment(nh1) = splt->rold;
 }
 if(ibgpSegment(*n2) == splt->rnew){
	rc=ibgiRegionOfPoint(splt->gold,&nh2,n2);
	if(rc==ibgError) ibgpSegment(nh2) = splt->rold;
 }
 if(ibgpSegment(*n3) == splt->rnew){
	rc=ibgiRegionOfPoint(splt->gold,&nh3,n3);
	if(rc==ibgError) ibgpSegment(nh3) = splt->rold;
 }
 fu = ibgpSegment(nhf);
 if(ibgCorrect==ibgtStatusRF(&splt->g.top,splt->rnew,fu)){
	if(fu==splt->face){
		rc=ibgiRegionOfPoint(splt->gold,&nhf,&nh1);
	 	/* search along the new boundary */
		rc = SplitRegionLine1(splt,nint,nface,&nhf,&nh1,&nh2,&nh3);
		if(rc==ibgNotFound){
			rc = ibgDefaultLineWithTriangle(This,
		     				nint,nface,n1,n2,n3);
		 	if(rc != ibgLineFound) return rc;
			ibgwarning;
		/* very strange - Default has found a line, Line1 not. */
		}
		if(rc==ibgSplitFound){
			ibgpSegment(*nint) =
				splitCreateLineNr(splt,ibgpSegment(*nint));
			ibgpType(*nint) = ibgSLine;
			return ibgLineFound;
		}
		return rc;
	}
	dd = 0;
	nhb = nh1;
facerecover:
	rc = ibgiFaceWithEdge(splt->gold,&nh,&nhb,&nh2);
	if(rc!=ibgFaceFound){
		if((dd>0) && (dd<20*splt->g.Delta)){
      	/* uncertain old geometry description!!! use nh instead of nhf */
			ibgmessage(ibgMWUncGeom);
			goto facefound;
		}else{
			rc = ibgDefaultLineWithTriangle(This,
						nint,nface,n1,n2,n3);
		 	return rc;
		}
	}
	dd = dr = 0;
	for(d=0;d<ibgDIM;d++){
	   	dx = ibgpX(nh)[d]-ibgpX(nhf)[d];
		dd += dx>0 ? dx : -dx;
		dr += dx*(ibgpX(nh)[d]-ibgpX(nhb)[d]);
	}
	if(dd>2*splt->g.Delta){
		if(dr>=0){
			for(d=0;d<ibgDIM;d++){
			   	ibgpX(nhb)[d] = (ibgpX(nh)[d]+ibgpX(nhf)[d])/2;
			}
			ibgiRegionOfPoint(splt->gold,&nhb,&nh);
			goto facerecover;
		}else{
			;
		}
	}
facefound:
	nhf = nh;
	fu = ibgpSegment(nhf);
 }
 if(ibgCorrect==ibgtStatusRF(splt->told,splt->rold,fu)){
	rc = SplitRegionLine0(splt,nint,&nhf,&nh1,&nh2,&nh3);
 }else{
	rc = ibgiLineWithTriangle(splt->gold,nint,&nhf,&nh1,&nh2,&nh3);
 }

 if(rc==ibgLineFound){
    if(ibgCorrect==ibgtStatusRL(splt->told,splt->rold,ibgpSegment(*nint))){
    	if((splt->func)(splt->fThis,nint)>0){
  		ibgpSegment(*nint) = splitChangeLineNr(splt,ibgpSegment(*nint));
	}
    }
 }else if(rc==ibgSplitFound){
	ibgpSegment(*nint) = splitCreateLineNr(splt,ibgpSegment(*nint));
	ibgpType(*nint) = ibgSLine;
	return ibgLineFound;
 }else{
    if(ibgCorrect==ibgtStatusRF(splt->told,splt->rold,ibgpSegment(*nint))){
    	if((splt->func)(splt->fThis,nint)>0){
  		ibgpSegment(*nint) = splitChangeFaceNr(splt,ibgpSegment(*nint));
	}
    }
 }
 return  rc;
}

static int Free(ibGeometry This)
{ibgSplitRegion splt=((ibgSplitRegion)This);
 ibgdFree(splt->gold);
 return ibgSuccess;
}
ibGeometryClassRec	ibgSplitRegionClass;

ibGeometry ibgdSplitRegion(ibGeometry 	gold,
			ibgSegment		rnew,
			ibgSegment		rold,
			ibgFloat	(*func)(ibgPtObject fThis, ibgPoint *n),
			ibgPtObject	fThis,
			ibgFloat	Dfunc)
{ibgSplitRegion splt=(ibgSplitRegion) malloc(sizeof(ibgSplitRegionRec));
 ibgdInitialize((ibGeometry)splt,&ibgSplitRegionClass);
 ibgtCopy(&(splt->g.top),&(gold->top));
 splt->gold = gold;
 splt->told = &(gold->top);
 splt->rold = rold; splt->rnew=rnew;
 splt->face = ibgDefaultFaceNr(rold,rnew);
 splt->func = func; splt->fThis=fThis;
 splt->epsilon = 0.5*splt->g.Delta*Dfunc;
 ibgSplitRegionClass.RegionOfPoint	= RegionOfPoint;
 ibgSplitRegionClass.FaceWithEdge	= FaceWithEdge;
 ibgSplitRegionClass.LineWithTriangle	= LineWithTriangle;
 ibgSplitRegionClass.Free		= Free;
 return (ibGeometry) splt;
}
