/* last edit: Ilja Schmelzer -------------- 17-OCT-1994 19:38:50.25	*/
/************************************************************************/
/*                                                                      */
/*  <<< I B G >>> - Intersection - Based Grid generation package 	*/
/*                                                                      */
/*  Version 1.1 by Ilja Schmelzer   schmelzer@iaas-berlin.d400.de       */
/*                                                                      */
/*  to be distributed under IBG license conditions (see "readme.ibg")	*/
/*                                                                      */
/************************************************************************/
#define MODULENAME "ibgg"
#define PRINTLEVEL 2
#include "ibgg.h"
#include "ibglib.h"
#include "ibglimits.h"
#include "ibgc.h"
#include "ibgi.h"
#include "ibgd0.h"
#include "ibgdefault.h"
#include "ibgg0.h"
#include "ibgg0p.h"

#define cct(cc) ibgridCellType(ibgg,cc)
#define ccu(cc) ibgridCellSegment(ibgg,cc)
#define ccn(cc) ibgridCPointList(ibgg,cc)
#define ccs(cc)	ibgridCSideList(ibgg,cc)
#define ccdef(cc)	(ccu(cc)> 0)
#define ccout(cc)	(ccu(cc)==0)

#define ibggDelaunayNew 1
ibgPoints       ibggExternalPoints;
ibgPoints       pp;
int		ibggFace;
int		ibggLine;
int		ibggNode;
ibgFloat	ibggDmin[ibgDIM];
ibgFloat	ibggBLWait,ibggBLHalf,ibggBLObtuse,ibggBctg;
ibgFloat	ibggOref,ibggOinv,ibggOref2,ibggOmax;
ibgFloat	ibggBV;
ibgFloat	ibggPref;
int   		ibggPrefL;

ibGrid		ibgg;

ibGrid		ibgg0oldgrid;

int ibgridNRealloc(ibGrid *g)
{void *adr; int max;
 if(g->lastPoint<g->maxPoint)	return 1;
 ibgg = *g;
 max=3*ibgg.lastPoint/2 + 100;
 if((adr=realloc(ibgg.Point,max*sizeof(ibgPoint)))==ibgNULL){goto noverflow;}
 ibgg.Point = (ibgPoint *) adr;
 if(ibgg.PointCell!=ibgNULL){
 	if((adr=realloc(ibgg.PointCell,max*sizeof(int)))==ibgNULL)
	 	{goto noverflow;}
 	ibgg.PointCell = (int*)adr;
 }
 if(ibgridHasPointData(ibgg)){
	if((adr=realloc(ibgg.PointData,max*(ibgg.NDataSize)))==ibgNULL)
	 	{goto noverflow;}
	ibgg.PointData = adr;
 }
 ibgg.maxPoint=max-1;
 *g = ibgg;
 return 1;
noverflow:
 return 0;
}

int ibgridCNew(ibGrid *g, ibgCellType t)
{void *adr; int max,l;
 int c1;
 if(g->lastCell>=g->maxCell){
	max=3*g->lastCell/2 + 100;
	if((adr=realloc(g->CellType,max*sizeof(ibgCellType)))==ibgNULL)
		 	{g->lastCell--; ibgfatal; return 0;}
	g->CellType=(ibgCellType *)adr;
	if((adr=realloc(g->CellSegment,max*sizeof(int)))==ibgNULL)
		 	{g->lastCell--; ibgfatal; return 0;}
	g->CellSegment=(ibgSegment*)adr;
	if((adr=realloc(g->CPointFirst,max*sizeof(int)))==ibgNULL)
		 	{g->lastCell--; ibgfatal; return 0;}
	g->CPointFirst=(int*)adr;
	if((adr=realloc(g->CSideFirst,max*sizeof(int)))==ibgNULL)
		 	{g->lastCell--; ibgfatal; return 0;}
	g->CSideFirst=(int*)adr;
	if(ibgridHasCellData(*g)){
		if((adr=realloc(g->CellData,max*(g->CDataSize)))==ibgNULL)
			 	{g->lastCell--; ibgfatal; return 0;}
		g->CellData=adr;
	}
	g->maxCell = max-1;
 }
 c1 = ++(g->lastCell);
 (g->CPointFirst)[c1] = g->freeCPoint;
 g->freeCPoint += ibgcPoints(t);
 if(g->freeCPoint>=g->maxCPoint){
 	max=3*g->freeCPoint/2 + 100;
 	if((adr=realloc(g->CPoint,max*sizeof(int)))==ibgNULL)
 		{g->lastCell--; g->freeCPoint -= ibgcPoints(t);
 		 ibgfatal; return 0;}
  	g->CPoint=(int*)adr; g->maxCPoint=max-IBGCNMAX-1;
 }
 (g->CSideFirst)[c1] = g->freeCSide;
 switch(ibgcCODIM(t)){
 	case ibgSRegion:	g->freeCSide += (l=ibgcSides(t)); 	break;
 	case ibgSFace:	g->freeCSide += (l=ibgcSides(t) + 2);	break;
 	case ibgSLine:	g->freeCSide += (l=ibgcSides(t) + 1);	break;
 	case ibgSNode:	g->freeCSide += (l=ibgcSides(t) + 1);	break;
 }
 if(g->freeCSide>=g->maxCSide){
 	max=3*g->freeCSide/2 + 100;
 	if((adr=realloc(g->CSide,max*sizeof(int)))==ibgNULL)
 		{g->lastCell--; g->freeCPoint -= ibgcPoints(t);
 		 g->freeCSide -= l; ibgfatal; return 0;}
 	g->CSide=(int*)adr; g->maxCSide=max-IBGCSMAX-1;
 }
 g->CellType[c1] = t;
 g->CellSegment[c1] = 0;
 return c1;
}

int ibgridCORealloc(ibGrid *gg)
{void *adr,*adrn,*adrs; int max;
 if(gg->lastCOut>=gg->maxCOut){
	max=3*gg->lastCOut/2 + 100;
	if(gg->maxCOut==0){
		if((adrn=malloc(max*sizeof(int)))==ibgNULL)
			       				{ibgfatal;return 0;}
		if((adrs=malloc(max*sizeof(int)))==ibgNULL)
			       				{ibgfatal;return 0;}
	}else{
		if((adrn=realloc(gg->COutNext,max*sizeof(int)))==ibgNULL)
			       				{ibgfatal;return 0;}
		if((adrs=realloc(gg->COutside,max*sizeof(int)))==ibgNULL)
			       		     		{ibgfatal;return 0;}
	}
 	gg->COutNext = (int *) adrn;
 	gg->COutside = (int *) adrs;
      	gg->maxCOut  = max-1;
 }
 return 1;
}

static int er1,er2;

static int gcellFace(int c, ibgSegment m0, ibgSegment m1)
{int bl,br,dl,dr,i,*cnn=ccn(c),b0,b1,first=1;
 ibgCellType t=cct(c);
 ibgPoint *n;
 b0 = ibgDefaultFaceNr(m0,m1);
 if(m1==ibgOutsideRegionNr) return b0;
 for(i=0;i<ibgcPoints(t);i++){
      	n = &(ibgridPoint(ibgg,cnn[i]));
	if(ibgpType(*n)==ibgSRegion){
			/* innerer Knoten auf dem Rand */
		er1++; /* kleine Fehlermeldung */
		return ibgError;
	}else if(ibgpType(*n)==ibgSFace){
		if((b1=ibgpSegment(*n))!=b0){
			if(first==0) {
			 	er2++;/* kleine Fehlermeldung */
				return ibgError;
			}
			if(b1>=ibgSMAX){
			 	er2++; /* kleine Fehlermeldung */
				return ibgError;
			}
			b0=b1;
			first=0; continue;
		}
	}
 }
 return b0;
}

static int gcellSetFace(int c, ibgSegment m0, ibgSegment m1)
{ibgCellType t=cct(c);
 return	ibgDefaultFaceNr(m0,m1);
}

void ibgridMakeFaces(ibGrid *gg, ibGeometry g)
{int c0,co,c1,c2,cl,cl0,cl1,cl2,cs,cn,m,m0,m1,ni,n0,s,i,j,k,b0;
 int *cs0,*cs1,*cs2,*cn0,*cno,*cln,*css,*csn,*cnn,*cso,*pn,bs0,bs1,bso;
 ibgCellType t,ts,tn,tl,t0,t1,t2;
 int s0,s1,s2,p0,p1,p2;
 int bmindef=ibgDefaultFaceNr(1,0),bmax=0,*adr;
 ibgg = *gg;
 ibGeo = g;
 if(ibgg.gridDimension<1){
 	ibgg.firstFace=1; ibgg.lastFace=0;
	return;
 }
 er1 = er2 = 0;
 ibgg.firstFace= ibgg.lastCell+1;
/* creating ibgSFace-elements, their points, left and right neighbours: */
 cl0=ibgg.lastCell;
 for(c0=1;c0<=cl0;c0++) if(ccdef(c0)){
	if((m=ccu(c0))<=0) continue;
	t=cct(c0);
	for(s=0;s<ibgcSides(t);s++){
		cs=ccs(c0)[s];
		if(ccout(cs)) m1=ibgOutsideRegionNr; /* Au"senrand */
		else if(ibgcCODIM(cct(cs))!=ibgSRegion)	continue;
		else if(m<=(m1=ccu(cs)))		continue;
		switch(ibgcSSize(ts=cct(cs),s)){
			case 1:	tn=ibgc1Node;		break;
			case 2:	tn=ibgc2Edge;		break;
			case 3: tn=ibgc3Triangle;	break;
			case 4: tn=ibgc3Rectangle;	break;
		}
		ibgassert(ibgcPoints(tn)==ibgcSSize(t,s));
		cn = ibgridCNew(&ibgg,tn);
 		cct(cn)=tn;
		cnn=ccn(cn); csn=ccs(cn); css=ccs(cs); cn0=ccn(c0);
		for(i=0;i<ibgcPoints(tn);i++){
		      	cnn[i] = cn0[ibgcSPoint1(t,s,i)];
		}
		for(i=0;i<ibgcSides(tn);i++) csn[i]=0;
		csn[ibgcSExtR(tn)]=c0;
		csn[ibgcSExtL(tn)]=cs;
		ccs(c0)[s]=cn;
		for(i=0;i<ibgcSides(ts);i++) if(css[i]==c0) {css[i]=cn; break;}
		b0 = gcellFace(cn,m,m1);
		if(b0<0){ccu(cn)=b0=gcellSetFace(cn,m,m1);}
		else ccu(cn)=b0;
	}
 }
 ibgg.lastFace=ibgg.lastCell;
 *gg = ibgg;
}

void ibgridMakeLines(ibGrid *gg, ibGeometry g)
{int c0,co,c1,c2,cl,cl1,cl2,cs,cn,m,m0,ni,n0,s,i,j,k,b0;
 int *cs0,*cs1,*cs2,*cn0,*cno,*cln,*css,*csn,*cnn,*cso,*pn,bs0,bs1,bso;
 ibgCellType t,ts,tn,tl,t0,t1,t2;
 int s0,s1,s2,p0,p1,p2,first,bl,br,dl,dr;
 int bmindef=ibgDefaultFaceNr(1,0),bmax=0,*adr;
 if(ibgg.gridDimension<2){
	ibgg.firstFace=1; ibgg.lastFace=0;
	return;
 }
 ibgg  = *gg;
 ibGeo = g;
 ibgg.firstLine=ibgg.lastCell+1;
 for(c0=ibgg.firstFace;c0<=ibgg.lastFace;c0++){t0=cct(c0);
   for(s0=0;s0<ibgcSides(t0);s0++){
	if((cs0=ccs(c0))[s0]!=0) continue;
	n0=(cn0=ccn(c0))[ibgcSPoint1(t0,s0,0)];
/* search of the first (after c0) ibgSFace-element c1: */
	co=c0; c1=cs0[ibgcSExtR(t0)];
	while(ibgcCODIM(t1=cct(c1))==ibgSRegion){
		cso=ccs(c1);cno=ccn(c1);
	 	for(s1=0;s1<ibgcSides(t1);s1++) if(cso[s1]==co) break;
		ibgassert(s1<ibgcSides(t1));
		for(p1=0;p1<ibgcSSize(t1,s1);p1++)
			if(cno[ibgcSPoint1(t1,s1,p1)]==n0) break;
	    	co=c1;c1=cso[ibgcSSide(t1,s1,p1)];
	}
/* search of orientation in c1 */
	if(co==(cs1=ccs(c1))[ibgcSExtL(t1)]){
		for(s1=0;s1<ibgcSides(t1);s1++)
			if(ccn(c1)[ibgcSPoint1(t1,s1,0)]==n0) break;
		ibgassert(s1<ibgcSides(t1));
		co=c1;c2=cs1[ibgcSExtR(t1)];
	}else if(co==cs1[ibgcSExtR(t1)]){
		for(s1=0;s1<ibgcSides(t1);s1++)
			if(ccn(c1)[ibgcSPoint1(t1,s1,1)]==n0) break;
		ibgassert(s1<ibgcSides(t1));
	      	co=c1;c2=cs1[ibgcSExtL(t1)];
	}else ibgfatal;
	ibgassert(c0!=c1);
	first=1;
	do{
/* search of the next ibgSFace-element c2: */
    	    while(ibgcCODIM(t2=cct(c2))==ibgSRegion){
	    	cso=ccs(c2);cno=ccn(c2);
	 	for(s2=0;s2<ibgcSides(t2);s2++) if(cso[s2]==co) break;
		ibgassert(s2<ibgcSides(t2));
		for(p2=0;p2<ibgcSSize(t2,s2);p2++)
			if(ccn(c2)[ibgcSPoint1(t2,s2,p2)]==n0) break;
		co=c2;c2=cso[ibgcSSide(t2,s2,p2)];
	    }
/* search of orientation in c2 */
	    if(co==ccs(c2)[ibgcSExtL(t2)]){
	    	for(s2=0;s2<ibgcSides(t2);s2++)
			if(ccn(c2)[ibgcSPoint1(t2,s2,0)]==n0) break;
		ibgassert(s2<ibgcSides(t2));
		co=c2;c2=ccs(c2)[ibgcSExtR(t2)];
	    }else if(co==ccs(c2)[ibgcSExtR(t2)]){
	    	for(s2=0;s2<ibgcSides(t2);s2++)
			if(ccn(c2)[ibgcSPoint1(t2,s2,1)]==n0) break;
		ibgassert(s2<ibgcSides(t2));
	      	co=c2;c2=ccs(c2)[ibgcSExtL(t2)];
	    }else ibgfatal;
	    if(first) {
		if(co==c0) if(ccu(c0)==ccu(c1)){/* typical - no new element */
			ccs(c0)[s0]=c1; ccs(c1)[s1]=c0; goto nexts0;
	     	}first=0;
/* creating a new ibgSLine-element: */
		if(ibgg.gridDimension==3) tl=ibgc3Edge; else tl=ibgc2Node;
		cl  = ibgridCNew(&ibgg,tl);
		cln = ccn(cl); cct(cl) = tl;
		m0  = -1;
		for(i=0;i<ibgcPoints(tl);i++){
		      	cln[i] = ni = ccn(c0)[ibgcSPoint1(t0,s0,i)];
			if(ibgpType(ibgridPoint(ibgg,ni))==ibgSLine){
				if(m0== -1){
					m0 = ibgpSegment(ibgridPoint(ibgg,ni));
				}else if(ibgpSegment(ibgridPoint(ibgg,ni))!=m0){
					m0 = ibgDefaultLineNr;
				}
			}
		}
		if(m0== -1) m0=ibgDefaultLineNr;
		ccu(cl) = m0;
/* default space is only for one external neighbour ibgcSExtF */
  		bs0 = ibgridCONew(&ibgg); if(bs0==0) return;
  		bs1 = ibgridCONew(&ibgg); if(bs1==0) return;
		ccs(cl)[ibgcSExtF(tl)] = bs0;
		ibgg.COutside[bs0] = c0; ccs(c0)[s0] = cl;
		ibgg.COutside[bs1] = c1; ccs(c1)[s1] = cl;
		ibgg.COutNext[bs0] = bs1;
	    }
	    if(co==c0) {ibgg.COutNext[bs1] = bs0; break;}
	    bso = bs1; bs1 = ibgridCONew(&ibgg); if(bs1==0) return;
	    ibgg.COutNext[bso] = bs1;
	    ibgg.COutside[bs1] = co; ccs(co)[s2] = cl;
	}while(1);
nexts0:	;
    }
 }
 ibgg.lastLine=ibgg.lastCell;
 *gg = ibgg;
}
void ibgridMakeNodes(ibGrid *gg, ibGeometry g)
{int c0,co,c1,c2,cl,cl0,cl2,cs,cn,m,m0,ni,n0,s,i,j,k,b0;
 int *cs0,*cs1,*cs2,*cn0,*cno,*cln,*css,*csn,*cnn,*cso,*pn,bs0,bs1,bso;
 ibgCellType t,ts,tn,tl,t0,t1,t2;
 int s0,s1,s2,p0,p1,p2,first,bl,br,dl,dr;
 int bmindef=ibgDefaultFaceNr(1,0),bmax=0,*adr;
 if(ibgg.gridDimension<3){
	ibgg.firstNode=1; ibgg.lastNode=0;
	return;
 }
 ibgg  = *gg;
 ibGeo = g;
 ibgg.firstNode=ibgg.lastCell+1;
 pn=(int*)calloc(ibgg.lastPoint+1,sizeof(int));
 cl2=ibgg.lastLine;
 for(c0=ibgg.firstLine;c0<=cl2;c0++){t0=cct(c0);
     for(s0=0;s0<ibgcPoints(t0);s0++){
	n0=ccn(c0)[s0];
	if(pn[n0]==0){pn[n0]= -c0; continue;}
	if(pn[n0]< 0){
		c1 = -pn[n0];
		for(s1=0;s1<2;s1++) if(ccn(c1)[s1]==n0) break;
		ibgassert(s1<2);
		ccs(c1)[s1] = c0; ccs(c0)[s0] = c1;
		pn[n0] = c0; continue;
	}
	c1 = pn[n0]; t1=cct(c1);
	if(ibgcCODIM(t1)==ibgSLine){
		cl = ibgridCNew(&ibgg,ibgc3Node); if(cl==0) return;
		ccn(cl)[0]=n0;
		bs0 = ibgridCONew(&ibgg); if(bs0==0) return;
		ccs(cl)[ibgcSExtF(ibgc3Node)] = bs0;
		ibgg.COutside[bs0] = c0;
		ibgg.COutNext[bs0] = bs0;
		for(s1=0;s1<2;s1++) if(ccn(c1)[s1]==n0) break;
		ibgassert(s1<2);
		c2=ccs(c1)[s1];
		for(s2=0;s2<2;s2++) if(ccn(c2)[s2]==n0) break;
		ibgassert(s2<2);
		ccs(c0)[s0]=cl; ccs(c1)[s1]=cl; ccs(c2)[s2]=cl;
		pn[n0] = cl;
	}else if(ibgcCODIM(t1)==ibgSNode){
		bs0 = ibgridCONew(&ibgg); if(bs0==0) return;
		ccs(c0)[s0]=c1;
		bso = ccs(cl)[ibgcSExtF(ibgc3Node)];
		ibgg.COutside[bs0] = c0;
		ibgg.COutNext[bs0] = ibgg.COutNext[bso];
		ibgg.COutNext[bso] = bs0;
	}else ibgfatal;
     }
 }
 for(n0=1;n0<=ibgg.lastPoint;n0++) if(pn[n0]<0) {
		cl = ibgridCNew(&ibgg,ibgc3Node); if(cl==0) return;
		ccn(cl)[0]=n0;
		c0 = -pn[n0];
		if     (ccn(c0)[0]==n0) ccs(c0)[0]=cl;
		else if(ccn(c0)[1]==n0) ccs(c0)[1]=cl;
		else ibgfatal;
		ccs(cl)[ibgcSExtF(ibgc3Node)]=c0;
		bs0 = ibgridCONew(&ibgg); if(bs0==0) return;
		ccs(cl)[ibgcSExtF(ibgc3Node)] = bs0;
		ibgg.COutside[bs0] = c0;
		ibgg.COutNext[bs0] = bs0;
 }
 free(pn);
/* Test des Gitters auf Korrektheit */
 for(c0=1;c0<=ibgg.lastCell;c0++) if(ccdef(c0)) {t0=cct(c0);
  	for(s0=0;s0<ibgcSides(t0);s0++){
		cs=ccs(c0)[s0];
		ibgassert(cs>0);
		ibgassert(ibgcCODIM(t0)<=ibgcCODIM(cct(cs)));
		if(ibgcCODIM(t0)==ibgcCODIM(cct(cs))){
			ibgassert(ccu(c0)==ccu(cs));
		}else if(ibgcCODIM(cct(cs))==ibgSFace){
		}
	}
 }
 ibgg.lastNode = ibgg.lastCell;
 *gg = ibgg;
}


void ibGridFunction(ibGrid *gg, int fnum,
	 	 	ibgFloat	(*func)(ibgPtObject fThis, ibgPoint *x),
			ibgPtObject	fThis)
{int nn;
 for(nn=ibgridFirstPoint(*gg);nn<=ibgridLastPoint(*gg);nn++){
	ibgpF(ibgridPoint(*gg,nn))[fnum] = func(fThis,&ibgridPoint(*gg,nn));
 }
}

ibGrid *ibGridCopy(ibGrid *grid)
{int i;
 ibGrid *gg=(ibGrid*)malloc(sizeof(ibGrid));
 *gg = ibgg = *grid;
 gg->Point	= (ibgPoint*)malloc(IBGPOINTSIZE*(gg->maxPoint+1));
 gg->CellType	= (ibgCellType*)malloc(sizeof(ibgCellType)*(gg->maxCell+1));
 gg->CellSegment	= (ibgSegment*)malloc(sizeof(int)*(gg->maxCell+1));
 gg->CPointFirst	= (int*)malloc(sizeof(int)*(gg->maxCell+1));
 gg->CSideFirst	= (int*)malloc(sizeof(int)*(gg->maxCell+1));
 gg->CPoint	= (int*)malloc(sizeof(int)*(gg->maxCPoint+1));
 gg->CSide	= (int*)malloc(sizeof(int)*(gg->maxCSide+1));
 gg->COutNext	= (int*)malloc(sizeof(int)*(gg->maxCOut+1));
 gg->COutside	= (int*)malloc(sizeof(int)*(gg->maxCOut+1));
 for(i=gg->firstPoint;i<=gg->lastPoint;i++){
	gg->Point[i] = ibgg.Point[i];
 }
 for(i=gg->firstCell;i<=gg->lastCell;i++){
	gg->CellType[i]		= ibgg.CellType[i];
	gg->CellSegment[i]	= ibgg.CellSegment[i];
	gg->CPointFirst[i]	= ibgg.CPointFirst[i];
	gg->CSideFirst[i]	= ibgg.CSideFirst[i];
 }
 for(i=0;i<gg->freeCPoint;i++){
		gg->CPoint[i]	= ibgg.CPoint[i];
 }
 for(i=0;i<gg->freeCSide;i++){
		gg->CSide[i]	= ibgg.CSide[i];
 }
 for(i=0;i<=gg->lastCOut;i++){
		gg->COutside[i]	= ibgg.COutside[i];
		gg->COutNext[i]	= ibgg.COutNext[i];
 }
 return gg;
}

ibGrid *ibGridFree(ibGrid *gg)
{
 if((gg->Point)		!= ibgNULL) free(gg->Point);
 if((gg->CellType)	!= ibgNULL) free(gg->CellType);
 if((gg->CellSegment)	!= ibgNULL) free(gg->CellSegment);
 if((gg->CPointFirst)	!= ibgNULL) free(gg->CPointFirst);
 if((gg->CSideFirst)	!= ibgNULL) free(gg->CSideFirst);
 if((gg->CPoint)		!= ibgNULL) free(gg->CPoint);
 if((gg->CSide)		!= ibgNULL) free(gg->CSide);
 if((gg->COutNext)	!= ibgNULL) free(gg->COutNext);
 if((gg->COutside)	!= ibgNULL) free(gg->COutside);
 if((gg->CellData)	!= ibgNULL) free(gg->CellData);
 if((gg->PointData)	!= ibgNULL) free(gg->PointData);
 if((gg->PointCell)	!= ibgNULL) free(gg->PointCell);
 free(gg);
 return ibgNULL;
}

ibggRefPar ibggRef;

static ibgFloat	ibggDminOld[3];

ibGrid	*ibGridGenerate(ibGeometry g,
	       	unsigned maxPoint,
      		unsigned nx, ibgFloat *x, unsigned fx,
		unsigned ny, ibgFloat *y, unsigned fy,
	    	unsigned nz, ibgFloat *z, unsigned fz,
	     	void (*RefineRegion)(ibgPtObject This, ibgPoint *point,
			   	      	ibgFloat *length),
		ibgPtObject pRefineRegion,
		void (*RefineFace)(ibgPtObject This, ibgPoint *point,
				      	ibgFloat *normal, ibgFloat *tangential),
	     	ibgPtObject pRefineFace,
	     	void (*RefineLine)(ibgPtObject This, ibgPoint *point,
			   	      	ibgFloat *normal, ibgFloat *tangential),
	     	ibgPtObject pRefineLine,
	     	void (*RefineNode)(ibgPtObject This, ibgPoint *point,
			   	      	ibgFloat *length),
	     	ibgPtObject pRefineNode,
		ibgBoolean (*RefineEdge) (ibgPtObject This,
					ibgPoint *n1, ibgPoint *n2),
		ibgPtObject pRefineEdge
	)
{ibGrid *gg; int i,rc;
 ibGrid0 *g0;
 clock_t time0=clock();
 clock_t timeo,timen;
 long int timed;
 ibgg0SetRefPar(&ibggRef,
		RefineRegion,pRefineRegion,
		RefineFace,pRefineFace,
		RefineLine,pRefineLine,
		RefineNode,pRefineNode,
		RefineEdge,pRefineEdge);
 if(nx<1)	nx=1;
 if(ny<1)	ny=1;
 if(nz<1)	nz=1;
/* if here is access violation the input is incorrect: */
 for(i=1;i<nx;i++) if(x[i]<=x[i-1]+g->Delta) goto incorrinput;
 for(i=1;i<ny;i++) if(y[i]<=y[i-1]+g->Delta) goto incorrinput;
 for(i=1;i<nz;i++) if(z[i]<=z[i-1]+g->Delta) goto incorrinput;
 if(nx==1){
	if(ny>1)	goto incorrinput;
	if(nz>1)	goto incorrinput;
 }else{
	if(g->Delta/(x[nx-1]-x[0]) < 10*IBG_FLT_EPSILON) goto deltatoosmall;
 }
 if(ny==1){
	if(nz>1)	goto incorrinput;
 }else{
	if(g->Delta/(y[ny-1]-y[0]) < 10*IBG_FLT_EPSILON) goto deltatoosmall;
 }
 if(nz==1){;
 }else{
	if(g->Delta/(z[nz-1]-z[0]) < 10*IBG_FLT_EPSILON) goto deltatoosmall;
 }
 ibgmsgreset;
 ibgmessage(ibgMIStart);
 ibgmsgreset;
 g0 = ibgg0CoarseGrid(maxPoint,nx,x,fx,ny,y,fy,nz,z,fz);
 if(g0==ibgNULL)	goto nogrid;
/*
 for(i=0;i<ibgDIM;i++){
	ibggDminOld[i] = ibggDmin[i];
	if(ibggDmin[i]>ibggDefaultLineNormal)ibggDmin[i]=ibggDefaultLineNormal;
 }
*/
 timed = (1000*(long int)((timeo=clock()) - time0))/CLOCKS_PER_SEC;
 ibgmsgreset;
 ibgmessage(ibgMIRefine);
 rc = ibgg0RegionOfPoint(g0,g);
 if(rc==ibgError){
	ibgg0Free(g0);
	goto nogrid;
 }

 rc = ibgg0Refine(g0,g,&ibggRef);
 if(rc==ibgError){
	goto nogrid;
 }
 timed = (1000*(long int)((timen=clock()) - timeo))/CLOCKS_PER_SEC;
 if(ibgMStatus == ibgMIRefine) ibgmsg(" %d ms;",timed);
 timeo=timen;
 ibgmsgreset;
 ibgmessage(ibgMIBShift);
 rc = ibgg0Boundary(g0,g,&ibggRef,maxPoint/2);
 if(rc==ibgError){
	goto nogrid;
 }
 timed = (1000*(long int)((timen=clock()) - timeo))/CLOCKS_PER_SEC;
 if(ibgMStatus == ibgMIBShift)	ibgmsg(" %d ms;",timed);
 timeo = timen;
 ibgmsgreset;
 ibgmessage(ibgMIDelaun);
 if (ibggDelaunayNew) {
     rc = ibgg0getPoints(&pp,g0);
     rc = ibgAppendPoints(&pp,&ibggExternalPoints);
     gg = (ibGrid*) malloc(sizeof(ibGrid));
     rc = ibggDelaunay(gg,g0->gg->gridDimension,&pp);
 }else{
     rc = ibgg0Delaunay(g0);
     if(rc==ibgError){
         goto nogrid;
     }
     timed = (1000*(long int)((timen=clock()) - timeo))/CLOCKS_PER_SEC;
     if(ibgMStatus == ibgMIDelaun)	ibgmsg(" %d ms;",timed);
     timeo = timen;
     ibgmsgreset;
     ibgmessage(ibgMIBGridG);
     ibgflush;
/*  gg = ibgg0Correct(g0); */
     gg = g0->gg;
     if(gg==ibgNULL){
         goto nogrid;
     }
 }
 ibgridMakeRegions(gg,g);
 ibgridMakeFaces(gg,g);
 ibgridMakeLines(gg,g);
 ibgridMakeNodes(gg,g);
/*
 for(i=0;i<ibgDIM;i++)	ibggDmin[i] = ibggDminOld[i];
*/
 timed = (1000*(long int)((timen=clock()) - timeo))/CLOCKS_PER_SEC;
 if(ibgMStatus == ibgMIBGridG)	ibgmsg(" %d ms;",timed);
 ibgmsgreset;
 ibgmessage(ibgMISuccess);
 timed = (1000*(long int)((timen=clock()) - time0))/CLOCKS_PER_SEC;
 ibgmsg(" with %d points, %d cells, in %d ms",
       	ibgg.lastPoint,ibgg.lastCell,timed);
 ibgg0oldgrid = *gg;
 ibgmsgreset;
 return gg;
incorrinput:
 ibgmessage(ibgMEIncCoarse);
 goto nogrid;
deltatoosmall:
 ibgmessage(ibgMESmallDelta);
 goto nogrid;
nogrid:
 ibgmsgreset;
 ibgmessage(ibgMINoGrid);
 return ibgNULL;
}

/* default refinement parameters and routines: */
ibgFloat ibggDefaultRegionLength,ibggDefaultNodeLength;
ibgFloat ibggDefaultFaceNormal,ibggDefaultFaceTangential;
ibgFloat ibggDefaultLineNormal,ibggDefaultLineTangential;

void ibggDefaultRefineRegion(ibgPtObject This, ibgPoint *point,
			   	      	ibgFloat *length)
{*length = ibggDefaultRegionLength;}

ibgBoolean ibggDefaultRefineEdge (ibgPtObject This, ibgPoint *n1, ibgPoint *n2)
{return ibgFalse;}

void ibggDefaultRefineFace(ibgPtObject This, ibgPoint *nface,
			   	      	ibgFloat *normal, ibgFloat *tangential)
{*normal	= ibggDefaultFaceNormal;
 *tangential  	= ibggDefaultFaceTangential;
}

void ibggDefaultRefineLine(ibgPtObject This, ibgPoint *nline,
			   	      	ibgFloat *normal, ibgFloat *tangential)
{*normal	= ibggDefaultLineNormal;
 *tangential  	= ibggDefaultLineTangential;
}

void ibggDefaultRefineNode(ibgPtObject This, ibgPoint *point,
			   	      	ibgFloat *length)
{*length = ibggDefaultNodeLength;}

void ibGridInit()
{ibgcInit();
/* initialization of parameters from the documentation: */
 ibggDmin[0]= 0.001;
 ibggDmin[1]= 0.001;
 ibggDmin[2]= 0.001;
 ibggDefaultRegionLength	= ibgInfty;
 ibggDefaultFaceNormal		= ibgInfty;
 ibggDefaultFaceTangential	= ibgInfty;
 ibggDefaultLineNormal		= 0.01;
 ibggDefaultLineTangential	= ibgInfty;
 ibggDefaultNodeLength	= ibgInfty;
/* initialization of other parameters: */
 ibggExternalPoints.Point = malloc(100*sizeof(ibgPoint));
 ibggExternalPoints.Near  = malloc(100*sizeof(int));
 ibggExternalPoints.maxPoint = 99;
 ibggExternalPoints.lastPoint = 0;
 ibggExternalPoints.firstPoint = 1;
 ibggFace=1;
 ibggLine=1;
 ibggNode=1;
 ibggBLWait=1.01;
 ibggBLHalf=2.01;
 ibggBLObtuse=1.01;
 ibggBctg=0.1;
 ibggOref=0.49;
 ibggOmax=0.0;
 ibggBV=1.01;	/* internal: 1 + rounding error */
 ibggPref=2.01;
 ibggPrefL=1;
/* defining dependend parameters: */
 ibggOinv=1.0/ibggOref;
 ibggOref2=ibggOref*ibggOref;
}
