/* last edit: Ilja Schmelzer -------------- 24-MAR-1995 19:53:26.67	*/
/************************************************************************/
/*                                                                      */
/*  <<< I B G >>> - Intersection - Based Grid generation package 	*/
/*                                                                      */
/*  Version 1.1 by Ilja Schmelzer   schmelzer@iaas-berlin.d400.de       */
/*                                                                      */
/*  to be distributed under IBG license conditions (see "readme.ibg")	*/
/*                                                                      */
/************************************************************************/
#include "ibgi.h"
#include "ibgg0.h"
#include "ibgd0.h"
#include "ibgg.h"
#include "ibgdefault.h"
#include "ibglib.h"

#include "ibgg0p.h"

	ibGeometry	ibGeo;	/* current geometry */
	ibGrid0	ibgg0;	/* current internal grid */

#define ibgFound 0
#define ibgNotFound (-1)
#define ibgrefreg(n1,norm)		\
	(*ibggRef.RefineRegion)(ibggRef.pRefineRegion,n1,norm)
#define ibgrefface(n1,norm,tang)	\
	(*ibggRef.RefineFace)(ibggRef.pRefineFace,n1,norm,tang)
#define ibgrefline(n1,norm,tang)	\
	(*ibggRef.RefineLine)(ibggRef.pRefineLine,n1,norm,tang)
#define ibgrefvert(n1,norm)		\
	(*ibggRef.RefineNode)(ibggRef.pRefineNode,n1,norm)
#define ibgrefedge(n1,n2) (*ibggRef.RefineEdge)(ibggRef.pRefineEdge,n1,n2)

/* local creation procedures: */
static int    	g0BNew0();

/* local boundary shift procedures: */
static int	g0bshift(int nn, int firstcall);
static void	g0bjump(int n0);
static int	g0breg(int bb);

/* local refinement procedures: */
static void	g1nref(int nn);
static int	g1eref(int nu, unsigned ro);
static void	g2nref(int nn);
static int	g2eref(int nu, unsigned ro);
static int	g2elot(int nn, unsigned rr, unsigned ro);
static void	g3nref(int nn);
static int	g3eref(int nu, unsigned ro);
static int	g3elot(int nn, unsigned rr, unsigned ro);
static int	g3elo2(int nn, unsigned rr, unsigned ro);
static int	g3qref2(int qq, unsigned ro);
static int	g3qref1(int qq, unsigned ro);
static int	g3qref(int qq);

int g3EDelete(int cb, int eb);

/* dimension-dependend data */
#define MAXR1 2
#define MAXO1 0

#define MAXR2 4
#define MAXO2 2
#define MAXQ2 0

#define MAXR3 6
#define MAXO3 4
#define MAXQ3 8

static ibgFloat ibgNorm,ibgtang;
/* <<< dimension-dependend arrays >>> */
/* dim == 3: */
#define MAXP3 5	/* P: the number after O */
enum q3richt{nnn,pnn,npn,nnp,npp,pnp,ppn,ppp,ooo};
enum n3richt{x3p,y3p,z3p,x3n,y3n,z3n,o3o};
static enum q3richt g3rrq[MAXR3][MAXR3]=
{{ooo,ppp,pnp,ooo,pnn,ppn},{ppn,ooo,ppp,npp,ooo,npn},{ppp,npp,ooo,nnp,pnp,ooo},
 {ooo,npn,npp,ooo,nnp,nnn},{pnp,ooo,nnp,nnn,ooo,pnn},{pnn,ppn,ooo,npn,nnn,ooo}};

static enum q3richt g3roq[MAXR3][MAXP3]=
{{ppp,ppn,pnn,pnp,ppp},{ppp,npp,npn,ppn,ppp},{ppp,pnp,nnp,npp,ppp},
 {npp,npn,nnn,nnp,npp},{pnp,nnp,nnn,pnn,pnp},{ppn,pnn,nnn,npn,ppn}};

static enum n3richt g3ror[MAXR3][MAXP3]=
{{y3p,z3n,y3n,z3p,y3p},{z3p,x3n,z3n,x3p,z3p},{x3p,y3n,x3n,y3p,x3p},
 {y3p,z3n,y3n,z3p,y3p},{z3p,x3n,z3n,x3p,z3p},{x3p,y3n,x3n,y3p,x3p}};

static enum n3richt g3rrr[MAXR3][MAXR3]=
{{o3o,z3p,y3n,o3o,z3n,y3p},{z3n,o3o,x3p,z3p,o3o,x3n},{y3p,x3n,o3o,y3n,x3p,o3o},
 {o3o,z3n,y3p,o3o,z3p,y3n},{z3p,o3o,x3n,z3n,o3o,x3p},{y3n,x3p,o3o,y3p,x3n,o3o}};

/* g3roq[r][o] is the octant of r, g3ror[r][o-1] and g3ror[r][o] */
/* g3rrq[r][o] is the positive octant of r,o;	*/
/* g3rrr[r][o] is the positive third direction to r,o;	*/

enum q3richt g3qother[MAXQ3]= {ppp,npp,pnp,ppn,pnn,npn,nnp,nnn};

/* dim == 2: */
enum q2richt{nn2,pn2,np2,pp2};
enum n2richt{x2p,y2p,x2n,y2n};
static enum q2richt g2roq[MAXR2][MAXO2]=
{{pp2,pn2},{np2,pp2},{np2,nn2},{nn2,pn2}};

static enum n2richt g2ror[MAXR2][MAXO2]=
{{y2p,y2n},{x2p,x2n},{y2p,y2n},{x2p,x2n}};

/* g2roq[r][o] is the quadrant of r and g2ror[r][o] */

#if MAXQ2 > 0
     	static enum q2richt g2qother[MAXQ2]= {pp2,np2,pn2,nn2};
#endif

int ibgg0getPoints(ibgPoints *p, ibGrid0 *g)
{
 ibgg0= *g; ibgg = *(g->gg);
 g0BNShift();
 g0QFree(g);
 p->spaceDimension = ibgg.spaceDimension;
 p->firstPoint     = 1;
 p->lastPoint      = ibgg.lastPoint;
 p->maxPoint       = ibgg.maxPoint;
 p->Point          = ibgg.Point;
 p->Near           = ibgg0.no;
}

/* allocation routines: */
int g0CAlloc(int max)
{void *adr;int i;
 if((adr=malloc(max*sizeof(ibgCellType)))==ibgNULL) {goto overflow;}
 ibgg.CellType = (ibgCellType*) adr;
 if((adr=malloc(max*sizeof(ibgSegment)))==ibgNULL) {goto overflow;}
 ibgg.CellSegment = (ibgSegment*) adr;
 if((adr=malloc(max*sizeof(int)))==ibgNULL) {goto overflow;}
 ibgg.CPointFirst = (int*) adr;
 if((adr=malloc(max*sizeof(int)))==ibgNULL) {goto overflow;}
 ibgg.CSideFirst = (int*) adr;
 if((adr=malloc(2*max*sizeof(int)))==ibgNULL) {goto overflow;}
 ibgg.CPoint = (int*) adr;
 if((adr=malloc(2*max*sizeof(int)))==ibgNULL) {goto overflow;}
 ibgg.CSide = (int*) adr;
 if((adr=malloc(max*sizeof(tcv)))==ibgNULL) {goto overflow;}
 ibgg.CellData	= (tcv*) adr;
 ibgg.CDataSize	= sizeof(tcv);
 if((adr=malloc(ibgg.maxPoint*sizeof(int)))==ibgNULL) {goto overflow;}
 ibgg.PointCell	= (int*) adr;
 ibgg.maxCell	= max-1;
 ibgg.maxCSide	= 2*max-IBGCSMAX-1;
 ibgg.maxCPoint	= 2*max-IBGCNMAX-1;
 ibgg.maxCOut	= 0;
 ibgg.lastCOut	= 0;
 ibgg.freeCSide	= 0;
 ibgg.freeCPoint	= 0;
 ibgg.firstCell   = 1;
 ibgg.firstRegion = 1;
 ibgg.firstFace   = 1;
 ibgg.firstLine   = 1;
 ibgg.firstNode = 1;
 ibgg.lastCell   = 0;
 ibgg.lastRegion = 0;
 ibgg.lastFace   = 0;
 ibgg.lastLine   = 0;
 ibgg.lastNode = 0;
 for(i=0;i<IBGCTMAX;i++) ibgg.CFree[i] = ibgridNoCell;
 ibgg.NFree = ibgridNoPoint;
 return 1;
overflow:
 ibgfatal;
 return 0;
}

static int g0NAlloc(int last, int max)
{void *adr; int d;
 ibgg.lastPoint=last;
 if(last+10 >= max) max = last+100;
 max++;
 if((adr=malloc(max*sizeof(ibgPoint)))==ibgNULL) {goto overflow;}
 ibgg.Point = (ibgPoint *) adr;
 ibgg.CDataSize	= sizeof(tcv);
 ibgg0.nx = (char *) &(((ibgPoint *)adr)->xx[0]);
 ibgg0.nt = (char *) &(((ibgPoint *)adr)->tt);
 ibgg0.nu = (char *) &(((ibgPoint *)adr)->uu);
 ibgg0.nf = (char *) &(((ibgPoint *)adr)->ff[0]);
#ifdef GLEVEL
 ibgg0.nl = (char *) &(((ibgPoint *)adr)->ll[0]);
#endif
 if((adr=malloc(max*sizeof(int)))==ibgNULL)	{goto overflow;}
 ibgg0.nstack = (int*) adr;
 for(d=0;d<ibgg.gridDimension;d++){
 	if((adr=malloc(max*sizeof(int)))==ibgNULL)	{goto overflow;}
 	ibgg0.ndp[d] = (int*)adr;
 	ibgg0.ndr[d] = (int*)adr;
 	if((adr=malloc(max*sizeof(int)))==ibgNULL)	{goto overflow;}
 	ibgg0.ndn[d] = (int*)adr;
 	ibgg0.ndr[ibgg.gridDimension+d] = (int*)adr;
 }
 if((adr=malloc(max*sizeof(int)))==ibgNULL)		{goto overflow;}
 ibgg0.no = (int*)adr;
 ibgg.maxPoint=max-1;
 return 1;
overflow:
 ibgfatal;
 return 0;
}

/* reallocation of all parts of ibgg0: */

int g0NRealloc()
{void *adr; int max; int d;
 if(ibgg.lastPoint<ibgg.maxPoint)	return 1;
 if(ibgridNRealloc(&ibgg)==0)	return 0;
 ibgg0.nx = (char *) &(ibgg.Point->xx[0]);
 ibgg0.nt = (char *) &(ibgg.Point->tt);
 ibgg0.nu = (char *) &(ibgg.Point->uu);
 ibgg0.nf = (char *) &(ibgg.Point->ff[0]);
#ifdef GLEVEL
 ibgg0.nl = (char *) &(ibgg.Point->ll[0]);
#endif
 if((adr=realloc(ibgg0.nstack,(ibgg.maxPoint+1)*sizeof(int)))==ibgNULL)
        	       				{goto overflow;}
 ibgg0.nstack = (int*)adr;
 for(d=0;d<ibgg0.maxq;d++){
        if((adr=realloc(ibgg0.nq[d],(ibgg.maxPoint+1)*sizeof(int)))==ibgNULL)
        	       				{goto overflow;}
        ibgg0.nq[d] = (int*)adr;
 }
 for(d=0;d<ibgg0.maxd;d++){
        if((adr=realloc(ibgg0.ndp[d],(ibgg.maxPoint+1)*sizeof(int)))==ibgNULL)
        	       				{goto overflow;}
        ibgg0.ndp[d] = (int*)adr;
        ibgg0.ndr[d] = (int*)adr;
        if((adr=realloc(ibgg0.ndn[d],(ibgg.maxPoint+1)*sizeof(int)))==ibgNULL)
               	       				{goto overflow;}
        ibgg0.ndn[d] = (int*)adr;
        ibgg0.ndr[ibgg.gridDimension+d] = (int*)adr;
 }
 if(ibgg0.no!=ibgNULL){
        if((adr=realloc(ibgg0.no,(ibgg.maxPoint+1)*sizeof(int)))==ibgNULL)
        	       	       				{goto overflow;}
        ibgg0.no = (int*)adr;
 }
 if(ibgg0.nb!=ibgNULL){
        if((adr=realloc(ibgg0.nb,(ibgg.maxPoint+1)*sizeof(int)))==ibgNULL)
        	       	       				{goto overflow;}
        ibgg0.nb = (int*)adr;
 }
 return 1;
overflow:
 ibgfatal;
 return 0;
}

static int g0BAlloc(int max)
{void *adr;
 if((adr=malloc(max*sizeof(ibgPoint)))==ibgNULL)	{goto overflow;}
 ibgg0.bPoint = (ibgPoint *) adr;
 ibgg0.bx = (char *) &(((ibgPoint *)adr)->xx[0]);
 ibgg0.bt = (char *) &(((ibgPoint *)adr)->tt);
 ibgg0.bu = (char *) &(((ibgPoint *)adr)->uu);
 ibgg0.bf = (char *) &(((ibgPoint *)adr)->ff[0]);
 if((adr=malloc(max*sizeof(int)))==ibgNULL)	{goto overflow;}
 ibgg0.bn = (int *) adr;
 if((adr=malloc(max*sizeof(int)))==ibgNULL)	{goto overflow;}
 ibgg0.bstack = (int *) adr;
 if((adr=calloc((ibgg.maxPoint+1),sizeof(int)))==ibgNULL){goto overflow;}
 ibgg0.nb = (int *) adr;
 ibgg0.bfree	= -1;
 ibgg0.lastBound	= 0;
 ibgg0.maxBound	= max-1;
 return 1;
overflow:
 ibgfatal;
 return 0;
}

static int g0BRealloc()
{void *adr; int max;
 if(ibgg0.lastBound<ibgg0.maxBound) return 1;
 max=3*ibgg0.lastBound/2;
 if((adr=realloc(ibgg0.bPoint,max*sizeof(ibgPoint)))==ibgNULL)	{goto overflow;}
 ibgg0.bPoint = (ibgPoint *) adr;
 ibgg0.bx = (char *) &(((ibgPoint *)adr)->xx[0]);
 ibgg0.bt = (char *) &(((ibgPoint *)adr)->tt);
 ibgg0.bu = (char *) &(((ibgPoint *)adr)->uu);
 ibgg0.bf = (char *) &(((ibgPoint *)adr)->ff[0]);
 if((adr=realloc(ibgg0.bn,max*sizeof(int)))==ibgNULL)	{goto overflow;}
 ibgg0.bn = (int *) adr;
 if((adr=realloc(ibgg0.bstack,max*sizeof(int)))==ibgNULL)	{goto overflow;}
 ibgg0.bstack = (int *) adr;
 ibgg0.maxBound=max-1;
 return 1;
overflow:
 ibgfatal;
 return 0;
}

int g0QFree()
{int q,d;
 if(ibgg0.maxq==0) return 0;
 ibgg0.maxq = 0;
 for(q=0;q<ibgg0.maxq;q++){
        if(ibgg0.qn[q] != ibgNULL)	free(ibgg0.qn[q]);
        if(ibgg0.nq[q] != ibgNULL)	free(ibgg0.nq[q]);
 }
 for(d=0;d<ibgg0.maxd;d++){
        if(ibgg0.qdp[d] != ibgNULL)	free(ibgg0.qdp[d]);
        if(ibgg0.qdn[d] != ibgNULL)	free(ibgg0.qdn[d]);
        if(ibgg0.ndp[d] != ibgNULL)	free(ibgg0.ndp[d]);
        if(ibgg0.ndn[d] != ibgNULL)	free(ibgg0.ndn[d]);
 }
 if(ibgg0.qn0 != ibgNULL)		free(ibgg0.qn0);
 ibgg0.maxd = 0;
 ibgg0.maxq = 0;
 return 1;
}

static int g0QAlloc(int last, int max)
{void *adr; int q,d;
 ibgg0.lastQuad = 0;
 ibgg0.maxQuad  = 0;
 if(ibgg0.maxq==0) return 1;
 if(last >= max) max = last+1;
 max++;
 for(q=0;q<ibgg0.maxq;q++){
        if((adr=malloc(max*sizeof(int)))==ibgNULL)
        	{g0QFree(); goto overflow; }
        ibgg0.qn[q] = (int *) adr;
        if((adr=malloc(max*sizeof(int)))==ibgNULL)
        	{g0QFree(); goto overflow; }
        ibgg0.nq[q] = (int *) adr;
 }
/*
 for(d=0;d<ibgg.gridDimension;d++){
        if((adr=malloc(max*sizeof(int)))==ibgNULL)
        	{g0QFree(); goto overflow; }
        ibgg0.qdp[d] = adr;
        ibgg0.qdr[d] = adr;
        if((adr=malloc(max*sizeof(int)))==ibgNULL)
        	{g0QFree(); goto overflow; }
        ibgg0.qdn[d] = adr;
        ibgg0.qdr[ibgg.gridDimension+d] = adr;
 }
 if((adr=malloc(max*sizeof(int)))==ibgNULL)
        	{g0QFree(); goto overflow; }
 ibgg0.qn0 = adr;
*/
 ibgg0.lastQuad = last;
 ibgg0.maxQuad  = max-1;
 return 1;
overflow:
 ibgfatal;
 return 0;
}

static int g0QRealloc()
{void *adr; int q,d,max;
 if(ibgg0.lastQuad<ibgg0.maxQuad) return 1;
 max = 3*ibgg0.lastQuad/2;
 for(q=0;q<ibgg0.maxq;q++){
        if((adr=realloc(ibgg0.qn[q],max*sizeof(int)))==ibgNULL){goto overflow;}
        ibgg0.qn[q] = (int *) adr;
 }
/*
 for(d=0;d<ibgg.gridDimension;d++){
        if((adr=realloc(ibgg0.qdp[d],max*sizeof(int)))==ibgNULL){goto overflow;}
        ibgg0.qdp[d] = adr;
        ibgg0.qdr[d] = adr;
        if((adr=realloc(ibgg0.qdn[d],max*sizeof(int)))==ibgNULL){goto overflow;}
        ibgg0.qdn[d] = adr;
        ibgg0.qdr[ibgg.gridDimension+d] = adr;
 }
 if((adr=realloc(ibgg0.qn0,max*sizeof(int)))==ibgNULL)	{goto overflow;}
 ibgg0.qn0 = adr;
*/
 ibgg0.maxQuad=max-1;
 return 1;
overflow:
 ibgfatal;
 return 0;
}

static int g0BNew0()
{int b;
 if((b=ibgg0.bfree)>0){
        ibgassert(cbn(b)== -100000);
        ibgg0.bfree = ibgg0.bstack[b];
        ibgassert((ibgg0.bfree>0)?(cbn(ibgg0.bfree)== -100000):1);
        return b;
 }
 if(ibgg0.lastBound < ibgg0.maxBound){
        cbn(b = ++ibgg0.lastBound) = 0;
        return b;
 }
 if(g0BRealloc()){
        cbn(b = ++ibgg0.lastBound) = 0;
        return b;
 }
 return 0;
}

int g0NNew0()
{int n;
 if((n=ibgg0.nfree)>0){
        ibgg0.nfree = ibgg0.nstack[n]; return n;
 }
 if(ibgg.lastPoint < ibgg.maxPoint){
        return ++ibgg.lastPoint;
 }
 if(g0NRealloc()){
        return ++ibgg.lastPoint;
 }
 return 0;
}

/* shift of point coordinates to boundary coordinates, free of boundary data */
int g0BNShift()
{int nn,bb;
 ibgPoint nd;
 ibgSegmentType t;
/* Speicherung der boundary-Koordinaten auf den Ursprungsknoten: */
 for(nn=1;nn<=ibgg.lastPoint;nn++) if(cnb(nn)!=0){
        bb = cnb(nn);
        *cnd(nn) = *cbd(bb);
 }
 free(ibgg0.nb); free(ibgg0.bn); free(ibgg0.bPoint);
 free(ibgg0.bstack);
 return ibgSuccess;
}

static int g0nfirst,g0nstb,g0nstl;
static int g0bfirst,g0bstb,g0bstl;

static int g0bshift(int nn, int firstcall)
{int m,bm,nu,no,nh,nr,lv,lv0,r0,d,u,r,o,bb,bo,bu,ne,be,rc;
 ibgFloat dx,dx0,dxmin;
 int lb[3],rb[3],lb0;
 if(cnb(nn)!=0) return 0;
beg:
/* Feststellung der Vorzugsrichtung r0 zum Verschieben des Knotens */
 m=cnu(nn); lv0=0; lb0=0; dxmin=big1;
 for(d=0;d<cgdim;d++){nu=cndn(nn,d);
notest:	if(cndef(no=cndp(nn,d))){
	        dx=cnx(no)[d]-cnx(nn)[d]; if(dx<dxmin) dxmin=dx;
	        if(cnb(no)==0){if(cnu(no)==m) goto nutest; lv=3;}
	        else if(cbt(bo=cnb(no))==ibgSFace) {
	        	lb0=1;
	        	if(ibgIncorrect!=ibgiStatusOfEdge(ibGeo,cnd(nn),cbd(bo))){
	        		if(cbx(bo)[d]<cnx(no)[d]) goto nutest;
	        	}
	        	if(m==cnu(no)) goto nutest;
	        	lv=2;
	        }else{	lb0=1; goto nutest;
	        }
	        if(nu==outside) lv=1;
	        if(lv>lv0)	{r0=d; dx0=dx; lv0=lv; ne=no;}
	        else if(lv==lv0){if(dx<dx0) {r0=d; dx0=dx; ne=no;}}
	}
nutest:	if(cndef(nu)){
	        dx=cnx(nn)[d]-cnx(nu)[d]; if(dx<dxmin) dxmin=dx;
	        if(cnb(nu)==0){if(cnu(nu)==m) continue; lv=3;}
	        else if(cbt(bu=cnb(nu))==ibgSFace) {
	        	lb0=1;
	        	if(ibgIncorrect !=
	        		ibgiStatusOfEdge(ibGeo,cnd(nn),cbd(bu))){
	        			if(cbx(cnb(nu))[d]>cnx(nu)[d]) continue;
	        	}
	        	if(m==cnu(nu)) continue;
	        	lv=2;
	        }else{	lb0=1; continue;
	        }
	        if(no==outside) lv=1;
	        if(lv>lv0)	{r0=cgdim+d; dx0=dx; lv0=lv; ne=nu;}
	        else if(lv==lv0){if(dx<dx0) {r0=cgdim+d; dx0=dx; ne=nu;}}
	}
 }
/* lv0=1 -> Verschiebung vom Au"senrand weg; lv0=2 -> Verschiebung zu face; */
 if(lv0>0){
	if(r0>=cgdim)	{d=r0-cgdim; r=r0; no=nn; nu=cndn(nn,d);}
	else    	{r=r0+cgdim; d=r0; nu=nn; no=cndp(nn,d);}
	if(lv0==1){
	        if(firstcall)	goto nextbsh;
	        if(cnb(ne)==0){
	        	if(cndr(no,d)==outside){
	        	    if(cndr(nu,r)==outside){
	        		if(cndef(ibgg0.eref(nu,d))) goto nextbsh;
	         		 return 0;
	        	    }
	        	}
	        }else{
	        	if(cndef(ibgg0.eref(nu,d))) goto nextbsh;
	        	 return 0;
	        }
	}
	if(dx0>ibggBLWait*dxmin && firstcall)	goto nextbsh;
	if(dx0>ibggBLHalf*dxmin){
	        if(firstcall)	goto nextbsh;
    	        if(cndef(ibgg0.eref(nu,d))) goto nextbsh;
	}
	bb = g0BNew; if(bb==0) return 0;
   	rc = ibgiFaceWithEdge(ibGeo,cbd(bb),cnd(nn),cnd(ne));
	ibgassert(rc==ibgFaceFound);
	bm = cbu(bb);
	if(cnb(ne)==0){
	        if(cndr(no,d)==outside){
/* Au"senrand no mu"s Au"senrand bleiben !!! */
	            ibgassert(cndr(nu,r)!=outside);
	            if(cbx(bb)[d]>=cnx(no)[d]-2*ibGeo->Delta){
	        	ibgassert(cnb(no)==0);
	        	cnb(no)=bb; cbn(bb)=no;
	             	cbx(bb)[d]=cnx(no)[d];
	        	if(no==nn) rc=0; else rc=nn;
	            }else{
	        	ibgassert(cnb(nu)==0);
	        	cnb(nu)=bb; cbn(bb)=nu;
	        	if(nu==nn) rc=0; else rc=nn;
	            }
	        }else if(cndr(nu,r)==outside){
/* Au"senrand nu mu"s Au"senrand bleiben !!! */
	            if(cbx(bb)[d]<=cnx(nu)[d]+2*ibGeo->Delta){
	        	ibgassert(cnb(nu)==0);
	        	cnb(nu)=bb; cbn(bb)=nu;
	        	cbx(bb)[d]=cnx(nu)[d];
	        	if(nu==nn) rc=0; else rc=nn;
	            }else{
	        	ibgassert(cnb(no)==0);
	        	cnb(no)=bb; cbn(bb)=no;
	        	if(no==nn) rc=0; else rc=nn;
	            }
	        }else if(cbx(bb)[d]<0.5*(cnx(nu)[d]+cnx(no)[d])){
	        	ibgassert(cnb(nu)==0);
	         	cnb(nu)=bb; cbn(bb)=nu;
	        	if(nu==nn) rc=0; else rc=nn;
	        }else{
	        	ibgassert(cnb(no)==0);
	        	cnb(no)=bb; cbn(bb)=no;
	        	if(no==nn) rc=0; else rc=nn;
	        }
 	}else{	/* the other point ne is already boundary point */
	        be=cnb(ne);
	        ibgassert(lv0>1);
	        rc=0;
	        if((cbx(bb)[d]-cnx(nn)[d])/(cbx(be)[d]-cnx(nn)[d])
	        			>=ibggBV){
	         	g0BDel(bb);
	        }else if(cbu(bb)==cbu(be)){
	        	if((cbx(bb)[d]-cnx(nn)[d])*(cbx(be)[d]-cnx(ne)[d])
	        			< 0.0){
/* that means, be is not compatible with cnu(nn)!!! */
	        		ibgmessage(ibgMWIncFFace);
	        		ibgmsg(" %d not face of %d ",cbu(bb),cnu(nn));
	        	 	g0BDel(bb);
	        	}else{
	        		ibgassert(cnb(nn)==0);
	        	 	cnb(nn)=bb; cbn(bb)=nn;
	        	}
	        }else{
	        	ibgassert(cnb(nn)==0);
	         	cnb(nn)=bb; cbn(bb)=nn;
	        }
	        if(r0<cgdim){
#ifdef GDEBUG
	              	if(cbx(bb)[d]>cbx(be)[d]){
	        		ibgassert(cbx(be)[d]>=cnx(ne)[d]-ibGeo->Delta);
	        		ibgassert(cbx(bb)[d]<=cnx(ne)[d]+ibGeo->Delta);
	        	}
#endif
	        	if(cbx(bb)[d]>cbx(be)[d]-ibGeo->Delta)
				cbx(bb)[d]=cbx(be)[d]-ibGeo->Delta;
		}else{
#ifdef GDEBUG
		      	if(cbx(bb)[d]<cbx(be)[d]){
				ibgassert(cbx(be)[d]<=cnx(ne)[d]+ibGeo->Delta);
				ibgassert(cbx(bb)[d]>=cnx(ne)[d]-ibGeo->Delta);
			}
#endif
			if(cbx(bb)[d]<cbx(be)[d]+ibGeo->Delta)
				cbx(bb)[d]=cbx(be)[d]+ibGeo->Delta;
		}
	}
	ibgassert(cbx(bb)[d]>=cnx(nu)[d]-ibGeo->Delta);
	ibgassert(cbx(bb)[d]<=cnx(no)[d]+ibGeo->Delta);
	if(cbx(bb)[d]<cnx(nu)[d]+ibGeo->Delta) cbx(bb)[d]=cnx(nu)[d]+ibGeo->Delta;
	if(cbx(bb)[d]>cnx(no)[d]-ibGeo->Delta) cbx(bb)[d]=cnx(no)[d]-ibGeo->Delta;
	for(o=0;o<ibgDIM;o++) if(o!=d){
		ibgassert(cbx(bb)[o]<cnx(nu)[o]+ibGeo->Delta);
		ibgassert(cbx(bb)[o]>cnx(nu)[o]-ibGeo->Delta);
		cbx(bb)[o]=cnx(nu)[o]; /* Rundungsfehler-Korrektur */
	}
	for(o=0;o<ibgg0.maxo;o++){
		r=cror(d,o);
		if(cnund(nr=cndr(nn,r)))	continue;
		if(cnb(nr)!=0)		continue;
	 	if(cnu(nr)==m)		continue;
/* including nr into the stack */
		if(nr>g0nfirst)		continue;
		if(ibgg0.nstack[nr]!=0)	continue;
		if(firstcall){
			if(nr>nn)	continue;
		}
		if(g0nstb<=0)		   g0nstb  = nr;
		else		ibgg0.nstack[g0nstl] = nr;
		ibgg0.nstack[nr] = -1;       g0nstl  = nr;
	}
	if(rc==nn) goto nextbsh;
 }else if(lb0){
	if(firstcall)	goto nextbsh;
	for(d=0;d<cgdim;d++) lb[d] = 0;
	for(r=0;r<ibgg0.maxr;r++)	if(cndef(no=cndr(nn,r))){
		if(cnb(no)==0)		continue;
		bm=cnb(no);
	   	for(d=0;d<cgdim;d++) if(d!=rposit(r)){
	 		if(cnx(no)[d]!=cbx(bm)[d]) {lb[d] = 1; rb[d] = r;}
		}
	}
	for(r=0;r<ibgg0.maxr;r++)	if(cnnothing(cndr(nn,r))){
		if(lb[rposit(r)]==0)	continue;
	  	if(cnund(ibgg0.elo2(nn,r,rb[rposit(r)]))){
			ibgfatal;
		}
	}
 }
 return 0;
nextbsh:
 if(nn <= g0nfirst && ibgg0.nstack[nn]==0){
	if(g0nstb<=0)	g0nstb = nn;
	else		ibgg0.nstack[g0nstl] = nn;
	ibgg0.nstack[nn] = -1;	   g0nstl  = nn;
 }
 return 0;
error:
 ibgfatal; return 0;
}



static void g0bjump(int n0)
{int r,r0,d,d0,nr,b,nr0,no,b0,i,rc;
 ibgFloat dx,dd,dx0,dd0=big1;
 int m=cnu(n0);
 for(r=0;r<ibgg0.maxr;r++){
	if(cnund(nr=cndr(n0,r)))	continue;
	if(cnb(nr)==0){
	        if(m==cnu(nr))		continue;
	        if(nr<=g0nfirst && ibgg0.nstack[nr]==0){
	        	if(g0nstb<=0)	g0nstb = nr;
	        	else		ibgg0.nstack[g0nstl] = nr;
	        	ibgg0.nstack[nr] = -1;	   g0nstl  = nr;
	        }
	        continue;
	}
	b=cnb(nr); d=rposit(r);
	if(r==d){
	        if(cbx(b)[d]>0.5*(cnx(nr)[d]+cnx(n0)[d])-ibGeo->Delta) continue;
	        dx = cbx(b)[d] - cnx(n0)[d];
	        dd = cnx(nr)[d] - cnx(n0)[d];
	}else{	if(cbx(b)[d]<0.5*(cnx(nr)[d]+cnx(n0)[d])+ibGeo->Delta) continue;
	        dx = cnx(n0)[d] - cbx(b)[d];
	        dd = cnx(n0)[d] - cnx(nr)[d];
	}
	if(dx<0)	{nr0=nr; b0=b; dd0=dd; r0=r; dx0 = dx; break;}
	if(dd<dd0)	{nr0=nr; b0=b; dd0=dd; r0=r; dx0 = dx; }
 }
 if(dd0==big1){
	if(n0<=g0nfirst) g0bshift(n0,1);
moverflow:      return;
 }
 d0 = rposit(r0);
#ifdef GDEBUG
 if(cbt(cnb(nr0))==ibgSFace){
 	for(d=0;d<cgdim;d++) if(d!=d0) ibgassert(cnx(n0)[d]==cbx(b0)[d]);
 }
#endif
/* special regularizations if dx0 is near 0: */
 if(dx0<0){
	if(cbt(cnb(nr0))==ibgSFace)	if(cnu(n0)!=cnu(nr0)){
		 				/* error in b0-computation!!!*/
		rc = ibgiFaceWithEdge(ibGeo,cbd(b0),cnd(n0),cnd(nr0));
		ibgassert(rc==ibgFaceFound);
		for(d=0;d<cgdim;d++)if(d!=d0)(cbx(b0)[d]=cnx(n0)[d]);
		if(rispos(r0)){
	      		if(cbx(b0)[d0]<cnx(n0)[d0]+ibGeo->Delta)
			   cbx(b0)[d0]=cnx(n0)[d0]+ibGeo->Delta;
			if(cbx(b0)[d0]>cnx(nr0)[d0]-ibGeo->Delta)
			   cbx(b0)[d0]=cnx(nr0)[d0]-ibGeo->Delta;
		}else{
			if(cbx(b0)[d0]>cnx(n0)[d0]-ibGeo->Delta)
			   cbx(b0)[d0]=cnx(n0)[d0]-ibGeo->Delta;
			if(cbx(b0)[d0]<cnx(nr0)[d0]+ibGeo->Delta)
			   cbx(b0)[d0]=cnx(nr0)[d0]+ibGeo->Delta;
		}
		goto shift;
      	}
 	if(rispos(r0)){
		if(cbx(b0)[d0]>cnx(n0)[d0]-ibGeo->Delta)
		   cbx(b0)[d0]=cnx(n0)[d0]-ibGeo->Delta;
	}else{
      		if(cbx(b0)[d0]<cnx(n0)[d0]+ibGeo->Delta)
		   cbx(b0)[d0]=cnx(n0)[d0]+ibGeo->Delta;
	}
 }else{
	if(cbt(cnb(nr0))==ibgSFace)	if(cnu(n0)==cnu(nr0)){
				 		/* error in b0-computation!!!*/
		no = cndr(n0,rother(r0));
		ibgassert(cndef(no));
		ibgassert(cnu(no)!=cnu(n0));
		rc = ibgiFaceWithEdge(ibGeo,cbd(b0),cnd(n0),cnd(no));
		ibgassert(rc==ibgFaceFound);
		for(d=0;d<cgdim;d++)if(d!=d0)(cbx(b0)[d]=cnx(n0)[d]);
		if(rispos(r0)){
			if(cbx(b0)[d0]>cnx(n0)[d0]-ibGeo->Delta)
			   cbx(b0)[d0]=cnx(n0)[d0]-ibGeo->Delta;
			if(cbx(b0)[d0]<cnx(no)[d0]+ibGeo->Delta)
			   cbx(b0)[d0]=cnx(no)[d0]+ibGeo->Delta;
		}else{
	      		if(cbx(b0)[d0]<cnx(n0)[d0]+ibGeo->Delta)
			   cbx(b0)[d0]=cnx(n0)[d0]+ibGeo->Delta;
			if(cbx(b0)[d0]>cnx(no)[d0]-ibGeo->Delta)
			   cbx(b0)[d0]=cnx(no)[d0]-ibGeo->Delta;
		}
		goto shift;
	}
	if(rispos(r0)){
      		if(cbx(b0)[d0]<cnx(n0)[d0]+ibGeo->Delta)
		   cbx(b0)[d0]=cnx(n0)[d0]+ibGeo->Delta;
	}else{
		if(cbx(b0)[d0]>cnx(n0)[d0]-ibGeo->Delta)
		   cbx(b0)[d0]=cnx(n0)[d0]-ibGeo->Delta;
	}
 }
shift:
 ibgassert(cnb(nr0)==b0);
 cnb(nr0)= 0;
 ibgassert(cnb(n0)==0);
 cnb(n0) = b0;		cbn(b0)  = n0;
 if(b0<=g0bfirst && ibgg0.bstack[b0]==0){
	if(g0bstb<=0)	g0bstb = b0;
	else{
		ibgassert(ibgg0.bstack[g0bstl] == -1);
		ibgg0.bstack[g0bstl] = b0;
	}
	ibgg0.bstack[b0] = -1;       g0bstl  = b0;
 }
 g0bjump(nr0);
}

/* regularising around the Face-point nn: */

/* potential bug: really we have to test the position of b1 on the edge. */
#define g0boundtest(d,n,bm)\
 if(!d){if(cnb(n)==0);\
	else if(cbt(cnb(n))==ibgSFace){if(cbu(cnb(n))==bm) d=1; else d= -1;}\
	else d=1;}\
 else if(cbt(cnb(n))>=ibgSLine){d=1;}

static int g0breg(int bb)
{int nn,bo,bu,b3,bh,br,dd,d,dr,di,ro,ru,rr,r1,r2,o,o1,qq,qo,nm,bm,om,ur;
 ibgSegmentType tnn;
 int upper0,upper,lower,rc,hoch;
 int no,nu,noo,nuu,noh,nuh,nor,nur,nr,n1,n2,n3,n4,nv,nv0,nv1,nv2,nv3,nvo,nvr;
 ibgFloat dx,dy,dl,del,du,dx1,dx2;
 nn = cbn(bb);
 switch(cbt(bb)){
 	case ibgSFace:
/* find shift direction dd, point no (oben), shift length dx, edge length dl */
 ibgrefface(cbd(bb),&ibgNorm,&ibgtang);
 for(dd=0;dd<cgdim;dd++){
	if((dx=cbx(bb)[dd]-cnx(nn)[dd])==0) continue;
	if(dx>0){ro=dd; ru=cgdim+dd;
		no=cndr(nn,ro); ibgAssert(cndef(no));
		dl=cnx(no)[dd]-cnx(nn)[dd];
		break;
	}else	{ro=cgdim+dd; ru=dd; dx = -dx;
                no=cndr(nn,ro); ibgAssert(cndef(no));
		dl=cnx(nn)[dd]-cnx(no)[dd];
		break;
	}
 }
 ibgassert(cndef(no));
 ibgassert(dx>0);
 if(dl>ibgNorm) {if(cndef(ibgg0.eref(nn,ro))) goto nextbreg;}
/* test correctness of segments */
 bm = cbu(bb); nm = cnu(nn);
 if(ibgiStatusOfEdge(ibGeo,cnd(nn),cbd(bb))==ibgIncorrect)
  	{if(cndef(ibgg0.eref(nn,ro))) goto nextbreg;}
 if(cndef(nu=cndr(nn,ru))){
   if(cnb(nu)!=0){
	if(cnu(nu)!=nm){
		if(ibgiStatusOfEdge(ibGeo,cnd(nn),cbd(cnb(nu)))==ibgIncorrect){
			if(cndef(ibgg0.eref(nn,ru))) goto nextbreg;
		}
	}
   }
 }
/* here we start to detect if we have found another intersection with
   the boundary bm going around the rectangle starting in lower/upper
   direction: lower/upper == 0 - unknown; 1 - found; -1 - not found; */
 if(cnb(no)==0){
	if(ibgiStatusOfEdge(ibGeo,cnd(no),cbd(bb))==ibgIncorrect){
		if(cndef(ibgg0.eref(nn,ro))) goto nextbreg;
		ibgmessage(ibgMWGridCoarse);
	 	upper0 = -1;
	}else{
	 	upper0 = 0;
	}
 }else if(cbt(cnb(no))==ibgSFace){
	if(rispos(ro)){
	  if(cbx(cnb(no))[dd]>cnx(no)[dd]){
/* potential error: refinement too often */
		if(ibgiStatusOfEdge(ibGeo,cnd(no),cbd(bb))==ibgIncorrect){
			if(cndef(ibgg0.eref(nn,ro))) goto nextbreg;
			ibgmessage(ibgMWGridCoarse);
		 	upper0 = -1;
		}else{
		 	upper0 = 0;
		}
	  }else upper0 = 1;
	}else if(cbx(cnb(no))[dd]<cnx(no)[dd]){
		if(ibgiStatusOfEdge(ibGeo,cnd(no),cbd(bb))==ibgIncorrect){
			if(cndef(ibgg0.eref(nn,ro))) goto nextbreg;
		     	ibgmessage(ibgMWGridCoarse);
			upper0 = -1;
		}else{
			upper0 = 0;
		}
	}else	upper0 = 1;
	if(upper0==1){
/* Anderer Schnitt des Rechtecks mit Rand gefunden. Ist es der Richtige? */
		if(bm==cbu(cnb(no))) upper0 = 1;
	 	else if(ibgIncorrect!=
			ibgiStatusOfEdge(ibGeo,cbd(bb),cbd(cnb(no)))){
/* Verschiedene Raender, aber gleiches Nachbargebiet vorhanden */
		     	upper0 = -1;
		}else{
			if(cndef(ibgg0.eref(nn,ro))) goto nextbreg;
		     	ibgmessage(ibgMWGridCoarse);
		     	upper0 = -1;
		}
	}
 }else{	upper0 = 1;
 }
/* look into the sides of the edge: */
 if(cgdim==3)		qq=cnq(nn,croq(ro,0));
 else if(cgdim<2)	return 0;
 for(o=1;o<=ibgg0.maxo;o++){
	if(cgdim==3){	qo=qq; if(qo==(qq=cnq(nn,croq(ro,o)))) continue;
			nur=cndr(nn,rr=cror(ro,o-1));
	}else if((nur=cndr(nn,rr=cror(ro,o-1)))==outside) continue;
/* we have found a side of a cuboid. Let's go around it. We find the
   corners nuu,noo,nur,nor. If there are refined region points on the edges
   or other dangerous situations (obtuse angles)
   we regularise by refinement and goto nextbreg;
*/
	upper=upper0; lower = 0;
	if(cndef(nur)){
		nuu=nn;
		if(rr<cgdim){dr=rr; dy=cnx(nur)[dr]-cnx(nuu)[dr];}
  		else  {dr=rr-cgdim; dy=cnx(nuu)[dr]-cnx(nur)[dr];}
		if(dy>ibgtang){
				if(cndef(ibgg0.eref(nuu,rr))) goto nextbreg;
		}
		if(dy*dy<ibggBLObtuse*dx*(dl-dx)){
/* nn shifted along the longer edge and gives an obtuse angle: */
				if(cndef(ibgg0.eref(nn,ro))) goto nextbreg;
		}
		if(cnu(nur)!=nm){
		    if(cnb(nur)==0){
/* the side nuu-nur intersects a boundary that is not detected.
   there can be an obtuse angle on the other side. */
			if(dy<dl) goto changeface;
			if(cnund(cndr(nuu,ru))){
			  	if(cndef(ibgg0.elot(nuu,ru,rr))) goto nextbreg;
			}
			if(cnund(cndr(nur,ru))){
			  	if(cndef(ibgg0.elot(nur,ru,rr))) goto nextbreg;
				if(cgdim==3){
					if(g3qref2(cnq(nuu,croq(ru,o)),rr))
								goto nextbreg;
				}
			     	ibgmessage(ibgMWGridCoarse);
			}
			if(ibggBctg*dy<dx){
				if(dy>1.5*dl||cndr(nur,rr)==outside){
				 	if(cndef(ibgg0.eref(nuu,rr))) goto nextbreg;
				}
				bb=g0BNew; if(bb==0) return 0;
			   	rc = ibgiFaceWithEdge(ibGeo,cbd(bb),cnd(nur),cnd(nuu));
				ibgassert(rc==ibgFaceFound);
				if(rr<cgdim){
				 	ibgassert(cbx(bb)[dr]>=cnx(nuu)[dr]-ibGeo->Delta);
					ibgassert(cbx(bb)[dr]<=cnx(nur)[dr]+ibGeo->Delta);
					if(cbx(bb)[dr]<cnx(nuu)[dr]+ibGeo->Delta)
						cbx(bb)[dr]=cnx(nuu)[dr]+ibGeo->Delta;
					if(cbx(bb)[dr]>cnx(nur)[dr]-ibGeo->Delta)
						cbx(bb)[dr]=cnx(nur)[dr]-ibGeo->Delta;
				}else{
				 	ibgassert(cbx(bb)[dr]<=cnx(nuu)[dr]+ibGeo->Delta);
					ibgassert(cbx(bb)[dr]>=cnx(nur)[dr]-ibGeo->Delta);
					if(cbx(bb)[dr]>cnx(nuu)[dr]-ibGeo->Delta)
						cbx(bb)[dr]=cnx(nuu)[dr]-ibGeo->Delta;
					if(cbx(bb)[dr]<cnx(nur)[dr]+ibGeo->Delta)
						cbx(bb)[dr]=cnx(nur)[dr]+ibGeo->Delta;
				}
				for(o1=0;o1<ibgDIM;o1++) if(o1!=dr){
					ibgassert(cbx(bb)[o1]<cnx(nuu)[o1]+ibGeo->Delta);
					ibgassert(cbx(bb)[o1]>cnx(nuu)[o1]-ibGeo->Delta);
					cbx(bb)[o1]=cnx(nuu)[o1];
				}
				if(cndr(nur,rr)==outside) {g0BDel(bb);}
				else{
					ibgassert(cnb(nur)==0);
				      	cnb(nur) = bb; cbn(bb) = nur;
/* potential bug - better put nur into g0breg-stack and return */
					while(g0breg(bb));
					goto nextbreg;
				}
		    	} /* of ctg*dy < dx */
  		    }else if((cbt(cnb(nur))==ibgSFace) &&
			(ibgiStatusOfEdge(ibGeo,cnd(nur),cbd(bb))==ibgIncorrect)){
			br = cnb(nur);
			ur = cbu(br);
			if(ibgiStatusOfEdge(ibGeo,cnd(nn),cbd(br))==ibgIncorrect){
				if(dy<dl) goto changeface;
				if(cndef(ibgg0.eref(nuu,rr))) goto nextbreg;
			}
			if(cnund(cndr(nuu,ru))){
			  	if(cndef(ibgg0.elot(nuu,ru,rr))) goto nextbreg;
			}
		    }
		} /* of cnu(nur)!=nm */
		if(0){
changeface:	    rc = ibgiFaceWithEdge(ibGeo,cbd(bb),cnd(nn),cnd(nur));
		    ibgassert(rc==ibgFaceFound);
		    for(d=0;d<cgdim;d++){
			if(d!=dr){
				cbx(bb)[d] = cnx(nn)[d];
			}else if(rispos(rr)){
		      		if(cbx(bb)[dr]<cnx(nn)[dr]+ibGeo->Delta)
				   cbx(bb)[dr]=cnx(nn)[dr]+ibGeo->Delta;
				if(cbx(bb)[dr]>cnx(nur)[dr]-ibGeo->Delta)
				   cbx(bb)[dr]=cnx(nur)[dr]-ibGeo->Delta;
			}else{
				if(cbx(bb)[dr]>cnx(nn)[dr]-ibGeo->Delta)
				   cbx(bb)[dr]=cnx(nn)[dr]-ibGeo->Delta;
				if(cbx(bb)[dr]<cnx(nur)[dr]+ibGeo->Delta)
				   cbx(bb)[dr]=cnx(nur)[dr]+ibGeo->Delta;
			}
		    }
 		    goto nextbreg;
		}
		if(cnund(nor=cndr(no,rr))){
			if(cnb(no)==0){
				if(cndef(ibgg0.elot(no,rr,ro))) goto nextbreg;
				ibgfatal;
			}
			noo = cndr(no, ro); ibgassert(cndef(noo));
			nor = cndr(noo,rr); ibgassert(cndef(nor));
			g0boundtest(upper,noo,bm);
		}else	noo = no;
	}else{	nuu = cndr(nn,ru);	ibgassert(cndef(nuu));
	        g0boundtest(lower,nn,bm);
/* the main edge is refined. */
		du = cnx(nn)[dd]-cnx(nuu)[dd]; if(du<0) du = -du;
		nur= cndr(nuu,rr);	ibgassert(cndef(nur));
		if(rispos(rr)){dr=rr; dy=cnx(nur)[dr]-cnx(nuu)[dr];}
		else    {dr=rr-cgdim; dy=cnx(nuu)[dr]-cnx(nur)[dr];}
		if(dy>ibgtang){
			      	if(cndef(ibgg0.eref(nuu,rr))) goto nextbreg;
		}
		if(dy*dy<ibggBLObtuse*(du+dx)*(dl-dx)){
/* an obtuse angle on the other side possible. */
			      	if(cndef(ibgg0.elot(nn,rr,ro))) goto nextbreg;
		}
		noo = no; nor=cndr(noo,rr); ibgassert(cndef(nor));
		g0boundtest(lower,nuu,bm);
	}
	g0boundtest(upper,nor,bm);
	g0boundtest(lower,nur,bm);
/* now nuu, noo and nur are defined. Let's find the rest: */
	if(cnund(nr=cndr(nur,ro))){
		if(cnb(nur)==0){
			if(cndef(ibgg0.elot(nur,ro,rr))) goto nextbreg; ibgfatal;
		}
		nur = cndr(nuh=nur,rr); ibgassert(cndef(nur));
		g0boundtest(lower,nur,bm);
		nr  = cndr(nur,ro); ibgassert(cndef(nr));
	}else	nuh = nuu;
	if(cnund(   cndr(nor,ru))){
		if(cnb(nor)==0){
			if(cndef(ibgg0.elot(nor,ru,rr))) goto nextbreg; ibgfatal;
		}
		nor = cndr(noh=nor,rr); ibgassert(cndef(nor));
		g0boundtest(upper,nor,bm);
	}else	noh = noo;
	if(nor!=nr){
		if(cnb(nr)==0){
/* es kann auch generell elot gemacht werden: */
			if(!lower || dx > dl/2){
				if(cndef(ibgg0.elot(nr,rother(rr),ro))) goto nextbreg;
				ibgfatal;
			}
		}
		g0boundtest(upper,nr,bm);
		ibgassert(cndr(nr,ro)==nor);
	}
	g0boundtest(lower,nr,bm);
/* now all corners nuu, noo, nor, nur are defined. If there are refinement
points on the edges, they are boundary points (nn,nuh,nr,noh).
If a good intersection was found we can continue with another side. */
	if(lower==1 || upper==1) continue;
/* potential bug: we really have to count the number of intersections. */
	if(!lower){
	       	g0boundtest(lower,nor,bm);
		g0boundtest(lower,noh,bm); if(lower==1) continue;
	}
	if(!upper){
	       	g0boundtest(upper,nur,bm);
		g0boundtest(upper,nuh,bm); if(upper==1) continue;
	}
/* the correct other intersection is not found. search for Edge-intersection */
	b3=g0BNew; if(b3==0) return 0;
   	rc=ibgiLineWithRectangle(ibGeo,cbd(b3),cbd(bb)
			,cnd(noo),cnd(nuu),cnd(nur),cnd(nor));
	if(rc!=ibgLineFound) {g0BDel(b3); continue;}
	if(ro<cgdim){
	      	if(rr<cgdim)	{n1=noo; n2=nuu; n3=nur; n4=nor;}
		else		{n1=nor; n2=nur; n3=nuu; n4=noo;}
	}else{
	      	if(rr<cgdim)	{n1=nuu; n2=noo; n3=nor; n4=nur;}
		else		{n1=nur; n2=nor; n3=noo; n4=nuu;}
	}
	if(cnx(n1)[dd]-cnx(n2)[dd]>cnx(n3)[dr]-cnx(n2)[dr])	hoch=1;
	else						hoch=0;
/* find the quarter of the rectangle containing the Edge point: */
	dx1=(cbx(b3)[dd]-cnx(n2)[dd])/(cnx(n1)[dd]-cnx(n2)[dd]);
	dx2=(cbx(b3)[dr]-cnx(n2)[dr])/(cnx(n3)[dr]-cnx(n2)[dr]);
	if(dx1<0.5){r1=dd;
	      	if(dx2<0.5)	{nv0=n2; nv1=n1; nv2=n3; nv3=n4; r2=dr;}
		else {dx2=1-dx2; nv0=n3; nv1=n4; nv2=n2; nv3=n1; r2=dr+cgdim;}
	}else{	r1=dd+cgdim; dx1=1-dx1;
	      	if(dx2<0.5)	{nv0=n1; nv1=n2; nv2=n4; nv3=n3; r2=dr;}
		else {dx2=1-dx2; nv0=n4; nv1=n3; nv2=n1; nv3=n2; r2=dr+cgdim;}
	}
/* find point to shift in the quarter: */
	nv=nv0;
	if((nvo=cndr(nv0,r1))!=nv1)
		if(dx1>0.25) nv=nvo;
	if((nvr=cndr(nv0,r2))!=nv2){
		if(dx2>0.25){
			if(nv==nvo){
				if(hoch) nv=nvr;
			}else nv=nvr;
		}
	}
	if(cndr(nv,rother(r1))==outside){
		nv = nvo;
		if(cndr(nv,rother(r2))==outside){
			nv = nv3;
		}
	}else	if(cndr(nv,rother(r2))==outside){
		nv = nvr;
	}
/* point nv found, let's shift him: */
	ibgassert(cndef(nv));
	bh = cnb(nv);
    	cnb(nv) = b3; cbn(b3) = nv;
	if(bh!=0){
		ibgassert(cbt(bh)==ibgSFace);
		{g0BDel(bh);}
	}
	if(rispos(r1)){
	      	ibgassert(cbx(b3)[dd]>=cnx(nv0)[dd]-ibGeo->Delta);
		if(cbx(b3)[dd]<cnx(nv0)[dd]+ibGeo->Delta) cbx(b3)[dd]=cnx(nv0)[dd]+ibGeo->Delta;
	}else{
	      	ibgassert(cbx(b3)[dd]<=cnx(nv0)[dd]+ibGeo->Delta);
		if(cbx(b3)[dd]>cnx(nv0)[dd]-ibGeo->Delta) cbx(b3)[dd]=cnx(nv0)[dd]-ibGeo->Delta;
	}
	if(rispos(r2)){
	      	ibgassert(cbx(b3)[dr]>=cnx(nv0)[dr]-ibGeo->Delta);
		if(cbx(b3)[dr]<cnx(nv0)[dr]+ibGeo->Delta) cbx(b3)[dr]=cnx(nv0)[dr]+ibGeo->Delta;
	}else{
	      	ibgassert(cbx(b3)[dr]<=cnx(nv0)[dr]+ibGeo->Delta);
		if(cbx(b3)[dr]>cnx(nv0)[dr]-ibGeo->Delta) cbx(b3)[dr]=cnx(nv0)[dr]-ibGeo->Delta;
	}
 	if(cbx(b3)[dd]>cnx(nv)[dd]){
		if(cbx(b3)[dd]<cnx(nv)[dd]+ibGeo->Delta) cbx(b3)[dd]=cnx(nv)[dd]+ibGeo->Delta;
	}else{
		if(cbx(b3)[dd]>cnx(nv)[dd]-ibGeo->Delta) cbx(b3)[dd]=cnx(nv)[dd]-ibGeo->Delta;
	}
 	if(cbx(b3)[dr]>cnx(nv)[dr]){
		if(cbx(b3)[dr]<cnx(nv)[dr]+ibGeo->Delta) cbx(b3)[dr]=cnx(nv)[dr]+ibGeo->Delta;
	}else{
		if(cbx(b3)[dr]>cnx(nv)[dr]-ibGeo->Delta) cbx(b3)[dr]=cnx(nv)[dr]-ibGeo->Delta;
	}
	for(di=0;di<cgdim;di++){
		if(di!=dd && di!=dr){
			ibgassert(cbx(b3)[di]<cnx(nv)[di]+ibGeo->Delta);
			ibgassert(cbx(b3)[di]>cnx(nv)[di]-ibGeo->Delta);
			cbx(b3)[di] = cnx(nv)[di];
		}
	}
/* potential bug - better put nv into g0breg-stack and return */
	while(g0breg(b3));
	if(bh==bb) return 0;
	goto nextbreg;
 }
 return 0;
	case ibgSLine:
 ibgrefline(cbd(bb),&ibgNorm,&ibgtang);
 for(d=0;d<cgdim;d++){
	if(cbx(bb)[d]!=cnx(nn)[d]){
		continue;
	}else{
		ibgassert(cgdim==3);
		for(o=0;o<MAXO3;o++){int qr;
			if((qq=cnq(nn,croq(d,o)))==(qr=cnq(nn,croq(3+d,o)))){
				if(qq>0){
					if(g3qref2(qq,d)) goto nextbreg;
					ibgfatal;
				}
			}
			if(qq>0){
			    if(cnx(cqn(qq,croq(d,o)))[d]-cnx(nn)[d]>ibgtang){
				if(g3qref2(qq,d)) goto nextbreg;
			    }else{
				if(g3qref1(qq,d)) goto nextbreg;
			}}
			if(qr>0){
			    if(cnx(nn)[d]-cnx(cqn(qr,croq(3+d,o)))[d]>ibgtang){
				if(g3qref2(qr,d)) goto nextbreg;
			    }else{
				if(g3qref1(qr,d)) goto nextbreg;
			}}
		}
	}
 }
 for(rr=0;rr<ibgg0.maxr;rr++){
	d  = rposit(rr);
	if(cnund(no = cndr(nn,rr))) continue;
	if(cnb(no)!=0){
	  	bo = cnb(no);
		del= 0;
		dx = cbx(bo)[0]-cbx(bb)[0]; if(dx<0) del -= dx; else del += dx;
		dx = cbx(bo)[1]-cbx(bb)[1]; if(dx<0) del -= dx; else del += dx;
		dx = cbx(bo)[2]-cbx(bb)[2]; if(dx<0) del -= dx; else del += dx;
		if(del<6*ibGeo->Delta){
/* wenn es hier knallt noch mal "uberlegen ob es so richtig ist: */
			if(( rispos(rr)&&(cbx(bo)[rr]<cnx(no)[rr]))
			|| (!rispos(rr)&&(cbx(bo)[ d]>cnx(no)[ d]))){
				if(cbt(bo)==ibgSNode){
					ibgassert(cnb(nn)==bb);
					cnb(nn) = 0; g0BDel(bb);
					g0bshift(nn,0); bb=bo;
		 		}else{
					ibgassert(cnb(no)==bo);
					cnb(no) = 0; g0BDel(bo);
					g0bshift(no,1);
				}
	 	 		goto nextbreg;
			} /* else error??? */
		}
	}
 }
 for(d=0;d<cgdim;d++){
	if(cnx(nn)[d]<cbx(bb)[d]){
	  if(cndef(no = cndp(nn,d)))
  	      	if(cnx(no)[d]-cnx(nn)[d]>ibgNorm)
		 	{if(cndef(ibgg0.eref(nn,d))) goto nextbreg;}
	  if(cndef(nu = cndn(nn,d)))
  	      	if(cnx(nn)[d]-cnx(nu)[d]>ibgNorm)
		 	{if(cndef(ibgg0.eref(nu,d))) goto nextbreg;}
	}else if(cnx(nn)[d]>cbx(bb)[d]){
	  if(cndef(no = cndn(nn,d)))
  	      	if(cnx(nn)[d]-cnx(no)[d]>ibgNorm)
		 	{if(cndef(ibgg0.eref(no,d))) goto nextbreg;}
	  if(cndef(nu = cndp(nn,d)))
  	      	if(cnx(nu)[d]-cnx(nn)[d]>ibgNorm)
		 	{if(cndef(ibgg0.eref(nn,d))) goto nextbreg;}
	}else	continue;
	for(o=0;o<ibgg0.maxo;o++){
		ro = ibgg0.ror[d][o];
		{register int d1 = rposit(ro);
		 if(cnx(nn)[d1]==cbx(bb)[d1]) continue;
		}
		if(cnund(cndr(nn,ro)))
		 	{if(cnund(ibgg0.elot(nn,ro,d)))ibgwarning; else goto nextbreg;}
		if(cndef(no)) if(cnund(cndr(no,ro)))
		 	{if(cnund(ibgg0.elot(no,ro,d)))ibgwarning; else goto nextbreg;}
		if(cndef(nu)) if(cnund(cndr(nu,ro)))
		 	{if(cnund(ibgg0.elot(nu,ro,d)))ibgwarning; else goto nextbreg;}
	}
 }
 return 0;
	case ibgSNode:
 return 0;
 }
nextbreg:
 if(bb<=g0bfirst && ibgg0.bstack[bb]==0){
	if(g0bstb<=0)	g0bstb = bb;
	else{
		ibgassert(ibgg0.bstack[g0bstl] == -1);
		ibgg0.bstack[g0bstl] = bb;
	}
	ibgg0.bstack[bb] = -1;       g0bstl  = bb;
 }
 return 0;
moverflow: ibgfatal; return 0;
}


static int g1eref(int nu, unsigned ro)
{int b,d,dh,ru;
 int no,nh,nur,nor,nr,po,pu;
 if(rispos(ro))	{d=ro;		ru=ro+cgdim;	no=cndr(nu,ro);}
 else		{d=ro-cgdim;	ru=ro;ro=d;	no=nu;nu=cndr(no,ru);}

/* Eigentliche Halbierung der Kante: */
 nh=g0NNew; if(nh==0) {return nothing;}
 cndr(nu,ro) = nh; cndr(nh,ru) = nu; cndr(no,ru) = nh; cndr(nh,ro) = no;
 if(ibgg0.nb!=ibgNULL) {cnb(nh) = 0;}

/* Definieren der Koordinaten und Level: */
 for(dh=0;dh<ibgDIM;dh++){
        if(d!=dh){
        	cnx(nh)[d] = cnx(nu)[d];
#ifdef GLEVEL
         	cnl(nh)[dh] = cnl(nu)[dh];
#endif
        }else{	cnx(nh)[dh] = 0.5*(cnx(nu)[dh]+cnx(no)[dh]);
                cndp(nh,dh) = cndn(nh,dh) = nothing;
#ifdef GLEVEL
        	cnl(nh)[dh] = cnl(nu)[dh]+1;
        	if(cnl(nh)[dh]<=cnl(no)[dh]) cnl(nh)[dh] = cnl(no)[dh]+1;
#endif
        }
 }

/* Definieren der Daten auf dem neuen Knoten: */
 ibgiRegionOfPoint(ibGeo,cnd(nh),cnd(nu));

if(ibgg0.nb!=ibgNULL) {g0bjump(nh);}
 return nh;
}/* end of g1eref */

ibgFloat g1cgh;
static ibgFloat g1cgo,g1cgu;

static int ibgg0New(ibGrid0 *g0)
{int n;
 ibgg0= *g0; ibgg = *(g0->gg);
 g0BNShift();
 g0QFree();
 g0CAlloc(ibgg.lastPoint);
 ibgg.lastCell  = ibgg.lastPoint;
 for(n=1;n<=ibgg.lastPoint;n++){
  	ibgg.CPointFirst[n] = n;
	ibgg.CSideFirst[n] = 0;
	ibgg.CPoint[n] = n;
 	cct(n) = ibgc0Node;
 	ccu(n) = 0;
 }
 *ibgg0.gg = ibgg;
 *g0 = ibgg0;
 return ibgSuccess;
}

static int ibgg1New(ibGrid0 *g0)
{ibGrid *gg;
 ibgPoint nd; ibgSegmentType t; int nn,bb,c,d,na,ne;
 ibgFloat xm,dx,big;
 int n,c0,*cn0,*cs0,nc;
 ibgg0= *g0; ibgg = *(g0->gg);
 g0BNShift();
 g0QFree();
 g0CAlloc(ibgg.lastPoint+10);
 ibgg.lastCell=ibgg.lastPoint+1;
 c0=0; cn0=ibgg.CPoint; cs0=ibgg.CSide; nc=0; na=1;
 for(n=1;n<=ibgg.lastPoint;n++){
	c0++;
	*(cn0++) = n;
	*(cs0++) = cndr(n,1);
 	if((*(cn0++) = *(cs0++) = cndr(n,0))==outside) ne=n;
	ibgg.CPointFirst[c0] = ibgg.CSideFirst[c0] = nc;
	nc += 2;
 	cct(c0) = ibgc1Edge;
 	ccu(c0) = 0;
 }
 ibgassert(cndn(na,0)==outside);
 ibgassert(cndp(ne,0)==outside);
 ibgg.lastPoint += 2; if(g0NRealloc()==0) return ibgError;
 xm=0.5*(cgmin(0)+cgmax(0)); dx=(cgmax(0)-cgmin(0));
 big=0.01*dx*dx/ibGeo->Delta;
 cnx(ibgg.lastPoint-1)[0] = xm-big;
 cnx(ibgg.lastPoint  )[0] = xm+big;
 cnu(ibgg.lastPoint-1)    = nothing;
 cnu(ibgg.lastPoint  )    = nothing;
 cnt(ibgg.lastPoint-1)    = ibgSRegion;
 cnt(ibgg.lastPoint  )    = ibgSRegion;
 c0++;
 ibgg.CPointFirst[c0] = ibgg.CSideFirst[c0] = nc;
 nc += 2;
 cct(c0) = ibgc1Edge;
 ccn(c0)[1] = na;
 ccs(c0)[1] = na;		ccs(na)[0] = c0;
 ccn(c0)[0] = ibgg.lastPoint-1;	ccs(c0)[0] = 0;
 ccn(ne)[1] = ibgg.lastPoint;	ccs(ne)[1] = 0;
 ibgassert(nc<ibgg.maxCPoint);
 ibgassert(nc<ibgg.maxCSide);
 ibgassert(c0<ibgg.maxCell);
 ibgg.freeCPoint = ibgg.freeCSide = nc;
 ibgg.lastCell  = c0;
 ibgassert(c0==ibgg.lastPoint-1);
 *ibgg0.gg = ibgg;
 *g0 = ibgg0;
 return ibgSuccess;
}

/* Lot auf Kante in Richtung rr in <rr,ro>-Fl"ache. */
static int g2elot(int nn, unsigned rr, unsigned ro)
{
 int no,nu,nor,nur,noru,nuro,norr,nurr,dd;
 no =cndr(nn,ro);		if(cnund(no))	return nothing;
 nu =cndr(nn,rother(ro));	if(cnund(nu))	return nothing;
 nor=cndr(no,rr);		if(cnund(nor))	return nothing;
 nur=cndr(nu,rr);		if(cnund(nur))	return nothing;
 noru=cndr(nor,rother(ro));
 if(cnnothing(noru)){
#ifdef GLEVEL
 	dd=rposit(rr); if(cnl(no)[dd]>=cnl(nor)[dd]) return nothing;
#endif
	norr=cndr(nor,rr);
	if(cnund(norr))		return nothing;
	noru=cndr(norr,rother(ro));
	if(cnnothing(noru))	return nothing;
	if(noru==nur)		return g2eref(noru,ro);
	if(noru==cndr(nur,rr)){
		cndr(nor,rother(ro)) = nur;
		cndr(nur,ro)	  = nor;
				return g2eref(nur,ro);
	}
	nurr=cndr(nur,rr);
	if(cnund(nurr))		return nothing;
	if(noru==cndr(nurr,ro)){
		cndr(nor,rother(ro)) = nur;
		cndr(nur,ro)	  = nor;
				return g2eref(nur,ro);
	}
	else 			return nothing;
 }
 if(noru==nur)			return g2eref(noru,ro);
 if(noru==cndr(nur,rr))		return g2eref(noru,ro);
 if(noru==cndr(nur,ro)){
 	cndr(nn,rr) = noru;
 	cndr(noru,rother(rr)) = nn;
				return noru;
 }
#ifdef GLEVEL
 dd=rposit(rr); if(cnl(nu)[dd]>=cnl(nur)[dd]) return nothing;
#endif
 nurr=cndr(nur,rr);
 if(cnund(nurr))		        return nothing;
 if(noru==cndr(nurr,ro)){
 	cndr(nn,rr) = noru;
	cndr(noru,rother(rr)) = nn;
				return noru;
 }
 else	 			return nothing;
}

static int g2eref(int nu, unsigned ro)
{int o,u,b,bu,d,dh,dr,eo,e1,rr,r1,ru;
 int no,nh,nur,nor,nr,noo,nuu;
 ibgFloat dx,dy;
 if(rispos(ro))	{d=ro;		ru=ro+cgdim;	no=cndr(nu,ro);}
 else		{d=ro-cgdim;	ru=ro;ro=d;	no=nu;nu=cndr(no,ru);}
 if((cnx(no)[d]-cnx(nu)[d])<=ibggDmin[d]) return nothing;

/* Schaffung der Voraussetzungen f"ur eine Unterteilung: */
 o=cror(d,0);
 if(cnund(cndr(nu,o))) g2elot(nu,o,d);
 if(cnund(cndr(no,o))) g2elot(no,o,d);
 u=rother(o);
 if(cnund(cndr(nu,u))) g2elot(nu,u,d);
 if(cnund(cndr(no,u))) g2elot(no,u,d);

/* Eigentliche Halbierung der Kante: */
 nh=g0NNew;
 cndr(nh,0) = cndr(nh,1) = cndr(nh,2) = cndr(nh,3) = nothing;
 cndr(nu,ro) = nh; cndr(nh,ru) = nu; cndr(no,ru) = nh; cndr(nh,ro) = no;
 if(ibgg0.nb!=ibgNULL) {cnb(nh) = 0;}

/* Definieren der Koordinaten und Level: */
 cnx(nh)[o] = cnx(nu)[o];
 cnx(nh)[2] = cnx(nu)[2];
 cnx(nh)[d] = 0.5*(cnx(nu)[d]+cnx(no)[d]);
#ifdef GLEVEL
 cnl(nh)[o] = cnl(nu)[o];
 cnl(nh)[2] = 0;
 cnl(nh)[d] = cnl(nu)[d]+1;
 if(cnl(nh)[d]<=cnl(no)[d]) cnl(nh)[d] = cnl(no)[d]+1;
#endif

/* Definieren der Daten auf dem neuen Knoten: */
 ibgiRegionOfPoint(ibGeo,cnd(nh),cnd(nu));

/* Hilfsstack der Reihenfolge der Knotenerzeugung: */
 ibgg0.no[nh]=nu;

/* Herstellen der Verbindungen zu den direkt sichtbaren Nachbarn: */
 for(o=0;o<MAXO2;o++){rr=g2ror[ro][o];
        nur=cndr(nu,rr);
        if(nur==outside)	cndr(nh,rr) = outside;
        else{
        	ibgassert(cndef(nur));
        	nr=cndr(nur,ro);
        	if(cnnothing(nr)) {
        		nur=cndr(nur,rr);	ibgassert(cndef(nur));
        		nr =cndr(nur,ro);	ibgassert(cndef(nr));
        	}
        	nor=cndr(no,rr); ibgassert(cndef(nor));
        	if(nr==nor)		 cndr(nh,rr) = -nr;
        	else if(nr==cndr(nor,rr))	 cndr(nh,rr) = -nr;
        	else {
        		ibgassert((cndr(nr,ro)==nor)||(cndr(nr,ro)==cndr(nor,rr)));
        		cndr(nr,rother(rr)) = nh; cndr(nh,rr) = nr;
        	}
        }
 }
 dx=cnx(no)[d]-cnx(nu)[d];
 for(o=0;o<MAXO2;o++){rr=g2ror[ro][o]; dr=rposit(rr);
        if(cnnothing(nr=cndr(nh,rr))){
        	nr=cndr(nu,rr);
        	if(rr>=cgdim)	dy=(cnx(nu)[dr]-cnx(nr)[dr])/dx;
        	else		dy=(cnx(nr)[dr]-cnx(nu)[dr])/dx;
        	if(dy<ibggOref) g2elot(nh,rr,ro);
        }else{
        	if(rr>=cgdim)	dy=(cnx(nu)[dr]-cnx(nr)[dr])/dx;
        	else		dy=(cnx(nr)[dr]-cnx(nu)[dr])/dx;
        	if(dy>ibggOinv){
        		if(cnund(cndr(nr=cndr(nu,rr),ro)))
        			g2elot(nr,ro,rr);
        		else if(cnund(cndr(nr=cndr(no,rr),ru)))
        			g2elot(nr,ru,rr);
        	}
        }
 }
 if(cndef(noo=cndr(no,ro))){
        if(cnx(noo)[d]-cnx(no)[d]>ibggPref*dx)
        			g2eref(no,ro);
 }
 if(cndef(nuu=cndr(nu,ru))){
        if(cnx(nu)[d]-cnx(nuu)[d]>ibggPref*dx)
        			g2eref(nuu,ro);
 }
 if(ibgg0.nb!=ibgNULL) {g0bjump(nh);}
 return nh;
}/* end of g2eref */


ibgFloat g2cgh=1.0;
static ibgFloat g2cgo,g2cgu,g2ctg,g2tan,g2hlf;

/* Die eigentliche Halbierungroutine eines Wrfels.  Halbiert nur wenn
alle Halbierungspunkte in dieser Richtung bereits vorhanden sind. */

int g3qref0(int qq, unsigned ro)
{
 int nu,no,nh,n0,n1,n2,n3,nr,nl,nur,nhr,nor,nurr,ou,oo,oh;
 int qn,o,w,eo,eu,fo,fu,go,gu;
 int rr,ru=rother(ro);
/* Test der Ausf"uhrbarkeit */
 if(cnund(qq)) return 0;
 nu = cqn(qq,(croq(ru,3)));  no = cqn(qq,(int)croq(ro,3));
 for(o=0;o<MAXO3;o++){
        ou = nu; nu = cqn(qq,croq(ru,o));
        oo = no; no = cqn(qq,croq(ro,o));
        if(no==cndr(nu,ro))	return 0;
        if((nor=cndr(oo,cror(ro,o)))==no) continue;
        if((nur=cndr(ou,cror(ru,o)))==nu) continue;
/* why the following is possible? Unknown, but has happened.
        if(nor==nothing) return 0;
        if(nur==nothing) return 0;
	*/
        if(nor==cndr(nur,ro))	return 0;
 }
/* Eigentliche Unterteilung des Quaders: */
 qn = g0QNew;
 eu = croq(ru,3); eo = croq(ro,3);
 nu = cqn(qq,eu); nh = cndr(nu,ro); ibgAssert(cndef(nh));
 no = cqn(qq,eo); ibgAssert(cndr(nh,ro) == no);
 fu = cqother(eo); fo = cqother(eu);
 for(o=0;o<MAXO3;o++){
        ou = nu; gu = fu; oh = nh; go = fo;
        eu = croq(ru,o);  eo = croq(ro,o);
        fu = cqother(eo); fo = cqother(eu);
        nu = cqn(qq,eu);  nh = cndr(nu,ro); ibgAssert(cndef(nh));
	no = cqn(qq,eo);  ibgAssert(cndr(nh,ro) == no);
	cqn(qn,eu) = nu;  cqn(qn,eo) = nh; cqn(qq,eu) = nh;
        cnq(nu,fo) = qn;  cnq(nh,fu) = qn; ibgAssert(cnq(nh,fo)==qq);
	rr = cror(ru,o);
	if((nhr=cndr(oh,rr))!=nh) {
	    ibgAssert(cndef(nhr));
            ibgAssert(cndr(nhr,rr)==nh);
	    cnq(nhr,fu)=qn; cnq(nhr,gu)=qn;
	}
	if((nur=cndr(ou,rr))!=nu) {cnq(nur,fo)=qn; cnq(nur,go)=qn;
	        if(o>1)continue;
	        if(cnund(nurr=cndr(nur,cror(ru,o+1)))) continue;
	        for(w=0;w<MAXO3;w++){
	        	if(cnq(nurr,croq(ro,w))==qq)
	        	   cnq(nurr,croq(ro,w))= qn;
	        }
	}
 }
/* evtl. Eintragung einer Verbindung auf der Halbierungsfl"ache: */
 n0 = cqn(qq,croq(ru,0));
 n1 = cqn(qq,croq(ru,1));
 n2 = cqn(qq,croq(ru,2));
 n3 = cqn(qq,croq(ru,3));
 if(n1==(nr=cndr(n0,cror(ru,1)))){
	if(n2==(nr=cndr(n1,cror(ru,2)))) return 1;
	if(n0==(nl=cndr(n3,cror(ru,0)))) return 1;
	cndr(nr,cror(ru,3)) = nl;
	cndr(nl,cror(ru,1)) = nr;	 return 1;}
 else	if(n3==(nl=cndr(n2,cror(ru,3)))){
	if(n2==(nr=cndr(n1,cror(ru,2)))) return 1;
	if(n0==(nl=cndr(n3,cror(ru,0)))) return 1;
	cndr(nr,cror(ru,3)) = nl;
	cndr(nl,cror(ru,1)) = nr;	 return 1;}
 else  {cndr(nr,cror(ru,2)) = nl;
	cndr(nl,cror(ru,0)) = nr;	 return 1;}
}

static int g3qref2(int qq, unsigned ro)
{
 int nu,no,nn,np,nur,nor,ou,oo;
 int o,ru=rother(ro);
 if(cnund(qq)) return 0;
 nn = cqn(qq,ibgg0.firstq);		np = cqn(qq,ibgg0.lastq);
 nu = cqn(qq,(croq(ru,3)));	no = cqn(qq,(int)croq(ro,3));
 for(o=0;o<MAXO3;o++){
 	ou = nu; nu = cqn(qq,croq(ru,o));
 	oo = no; no = cqn(qq,croq(ro,o));
 	if(no==cndr(nu,ro)){
 		g3eref(nu,ro);
 	 	if(nn!=cqn(qq,ibgg0.firstq)) return 1;
 		if(np!=cqn(qq,ibgg0.lastq)) return 1;
 	}
 	if((nor=cndr(oo,cror(ro,o)))==no) continue;
 	if((nur=cndr(ou,cror(ru,o)))==nu) continue;
        if(nor==nothing) return 0;
	if(nur==nothing) return 0;
  	if(nor==cndr(nur,ro)){
 		g3eref(nur,ro);
 	 	if(nn!=cqn(qq,ibgg0.firstq)) return 1;
 		if(np!=cqn(qq,ibgg0.lastq)) return 1;
 	}
 }
 return g3qref0(qq,ro);
}

static int g3qref1(int qq, unsigned ro)
{
 int nu,no,nur,nor,ou,oo;
 int o,ru=rother(ro);
 if(cnund(qq)) return 0;
 nu = cqn(qq,(croq(ru,3)));  no = cqn(qq,(int)croq(ro,3));
 for(o=0;o<MAXO3;o++){
 	ou = nu; nu = cqn(qq,croq(ru,o));
 	oo = no; no = cqn(qq,croq(ro,o));
        if(no!=cndr(nu,ro))	return g3qref2(qq,ro);
        if((nor=cndr(oo,cror(ro,o)))==no) continue;
        if((nur=cndr(ou,cror(ru,o)))==nu) continue;
        if(nor!=cndr(nur,ro))	return g3qref2(qq,ro);
 }
 return 0;
}

/* Die zentrale Kantenverfeinerungsroutine */

static int g3eref(int nu, unsigned ro)
{int qq,qo,o,b,d,dh,dr,eo,e1,rr,r1,ru;
 int no,nh,nur,nurr,nor,noo,nuu,nr,lo;
 ibgFloat dx,dy;
 if(rispos(ro))	{d=ro;		ru=ro+cgdim;	no=cndr(nu,ro);}
 else		{d=ro-cgdim;	ru=ro;ro=d;	no=nu;nu=cndr(no,ru);}
 if(cnund(no)) return nothing;
/* Schaffung der Voraussetzungen f"ur eine Unterteilung: */
 for(o=0;o<MAXO3;o++){
        if(cnund(qq= cnq(nu,lo=croq(ro,o))))continue;
        if     (qq==cnq(no,lo))
        	{if(!g3qref2(qq,ro)) return nothing;}
        else if(qq==cnq(nu,croq(ru,o)))
        	{if(!g3qref2(qq,ro)) return nothing;}
 }
/* Hartes externes Verbot einer weiteren Unterteilung: */
 if((cnx(no)[d]-cnx(nu)[d])<=ibggDmin[d]) return nothing;
/* Eigentliche Halbierung der Kante: */
 nh=g0NNew; if(nh==0) {return nothing;}
 cndr(nu,ro) = nh; cndr(nh,ru) = nu; cndr(no,ru) = nh; cndr(nh,ro) = no;
 if(ibgg0.nb!=ibgNULL) {cnb(nh) = 0;}
/* Definieren der Koordinaten und Level: */
 for(dh=0;dh<ibgDIM;dh++){
        if(d!=dh){
        	cnx(nh)[dh] = cnx(nu)[dh];
                cndp(nh,dh) = cndn(nh,dh) = nothing;
#ifdef GLEVEL
         	cnl(nh)[dh] = cnl(nu)[dh];
#endif
        }else{	cnx(nh)[dh] = 0.5*(cnx(nu)[dh]+cnx(no)[dh]);
#ifdef GLEVEL
        	cnl(nh)[dh] = cnl(nu)[dh]+1;
        	if(cnl(nh)[dh]<=cnl(no)[dh]) cnl(nh)[dh] = cnl(no)[dh]+1;
#endif
        }
 }
/* Definieren der Daten auf dem neuen Knoten: */
 ibgiRegionOfPoint(ibGeo,cnd(nh),cnd(nu));

/* Hilfsstack der Reihenfolge der Knotenerzeugung: */
 ibgg0.no[nh]=nu;
/* Definieren der nq-Zeiger: */
 for(o=0;o<MAXO3;o++){
        eo = croq(ro,o);
        cnq(nh,eo) = (cnq(nh,croq(ru,o)) = cnq(nu,eo));
 }
/* Herstellen der Verbindungen zu den direkt sichtbaren Nachbarn: */
 for(o=0;o<MAXO3;o++){
        rr = cror(ro,o);
        if(cnund(nur=cndr(nu,rr)))		{cndr(nh,rr)=  nur;continue;}
        if(cnund(nr=cndr(nur,ro))){
        	if(cnq(nh ,croq(ro,o)) !=
        	   cnq(nur,croq(ro,o)))		{cndr(nh,rr)= -nur;continue;}
        	if(cnund(nurr=cndr(nur,rr)))	{cndr(nh,rr)= -nur;continue;}
        	if(cnund(nr  =cndr(nurr,ro))) 	{cndr(nh,rr)= -nur;continue;}
        }
        if(cnund(nor=cndr(no,rr)))		{cndr(nh,rr)=  nor;continue;}
        if(nr==nor)		 		{cndr(nh,rr)= -nor;continue;}
        if(nr==cndr(nor,rr))			{cndr(nh,rr)= -nor;continue;}
        if((cndr(nr,ro)!=nor)&&(cndr(nr,ro)!=cndr(nor,rr)))
        					{cndr(nh,rr)= -nor;continue;}
        eo=croq(ro,o);	e1=croq(ro,o+1);
        if(cnq(nu,eo)==cnq(nu,e1))		{cndr(nh,rr)= -nor;continue;}
        cndr(nr,rother(rr)) = nh; cndr(nh,rr) = nr;
 }
/* Ende der eigentlichen Verfeinerung. Die direkten Nachbarn sind definiert */
/* Kontrolle m"oglicher Quaderhalbierungen: */
 qq=cnq(nh,croq(ro,3)); /* 3 == MAXO-1 */
 for(o=0;o<MAXO3;o++){qo=qq;
        if(qo==(qq=cnq(nh,croq(ro,o)))) continue;
        if(g3qref0(qq,ro))	qq=cnq(nh,croq(ro,o));
 }
 if(ibgg0.nb!=ibgNULL) {g0bjump(nh);}
 return nh;
}/* end of g3eref */

static int g3elot(int nn, unsigned rr, unsigned ro)
{
 int no,nu,nor,nur,noru,nuro,norr,nurr,dd;
 no =cndr(nn,ro);		if(cnund(no))	return nothing;
 nu =cndr(nn,rother(ro));	if(cnund(nu))	return nothing;
 nor=cndr(no,rr);		if(cnund(nor))	return nothing;
 nur=cndr(nu,rr);		if(cnund(nur))	return nothing;
 noru=cndr(nor,rother(ro));
 if(cnnothing(noru)){
#ifdef GLEVEL
 	dd=posit(rr); if(cnl(no)[dd]>=cnl(nor)[dd]) return nothing;
#endif
	norr=cndr(nor,rr);
	if(cnund(norr))		return nothing;
	noru=cndr(norr,rother(ro));
	if(cnnothing(noru))	return nothing;
	if(noru==nur)		return g3eref(noru,ro);
	if(noru==cndr(nur,rr)){
/*		cndr(nor,rother(ro)) = nur;
		cndr(nur,ro)	  = nor;
				return g3eref(nur,ro);
*/				return nothing;
	}
	nurr=cndr(nur,rr);
	if(cnund(nurr))		return nothing;
	if(noru==cndr(nurr,ro)){
/*		cndr(nor,rother(ro)) = nur;
		cndr(nur,ro)	  = nor;
				return g3eref(nur,ro);
*/				return nothing;
	}
	else 			return nothing;
 }
 if(noru==nur)			return g3eref(noru,ro);
 if(noru==cndr(nur,rr))		return g3eref(noru,ro);
 if(noru==cndr(nur,ro)){
 	cndr(nn,rr) = noru;
 	cndr(noru,rother(rr)) = nn;
				return noru;
 }
#ifdef GLEVEL
 dd=posit(rr); if(cnl(nu)[dd]>=cnl(nur)[dd]) return nothing;
#endif
 nurr=cndr(nur,rr);
 if(cnund(nurr))			return nothing;
 if(noru==cndr(nurr,ro)){
 	cndr(nn,rr) = noru;
	cndr(noru,rother(rr)) = nn;
			   	return noru;
 }
 else	 			return nothing;
}

static int g3elo2(int nn, unsigned rr, unsigned ro)
{int rc,q0;
 if(cndef(rc=g3elot(nn,rr,ro)))	return rc;
 q0=cnq(nn,ibgg0.rrq[rr][ro]);
 ibgassert(q0==cnq(nn,ibgg0.rrq[ro][rr]));
 g3qref2(q0,ibgg0.rrr[ro][rr]);
 if(cndef(rc=cndr(nn,rr)))	return rc;
 if(cndef(rc=g3elot(nn,rr,ro)))	return rc;
 if(cndef(rc=g3elot(nn,rr,ibgg0.rrr[ro][rr])))	return rc;
 ibgfatal;
}

/* parameter-controlled refinement of a cuboid */
static int g3qref(int qq)
{int nn,np,rc=0;
 ibgFloat dx,dy,dz;
beg:
 nn = cqn(qq,ibgg0.firstq); np = cqn(qq,ibgg0.lastq);
 dx = cnx(np)[0]-cnx(nn)[0];
 dy = cnx(np)[1]-cnx(nn)[1];
 dz = cnx(np)[2]-cnx(nn)[2];
/* Omax - maximal allowed anisotropy of a cuboid */
 if(dx>ibggOmax*dy)	{if((rc=g3qref2(qq,0))>0) goto beg;}
 if(dx>ibggOmax*dz)	{if((rc=g3qref2(qq,0))>0) goto beg;}
 if(dy>ibggOmax*dz)	{if((rc=g3qref2(qq,1))>0) goto beg;}
 if(dy>ibggOmax*dx)	{if((rc=g3qref2(qq,1))>0) goto beg;}
 if(dz>ibggOmax*dx)	{if((rc=g3qref2(qq,2))>0) goto beg;}
 if(dz>ibggOmax*dy)	{if((rc=g3qref2(qq,2))>0) goto beg;}
 return rc;
}

static void g3nref(int nn)
{int ref=1,i,q,d,u,o,rv,dv,rq,dq,e,no,noo,nu,nuu,nr,nur,nor,qq,qo;
 ibgFloat dx,dy,dz;
 ibgFloat xx[ibgDIM],lmax;
 ibgrefreg(cnd(nn),&lmax);
/* isotropic refinement (maxlength) in positive directions: */
 xx[0]=cnx(nn)[0]+lmax;
 xx[1]=cnx(nn)[1]+lmax;
 xx[2]=cnx(nn)[2]+lmax;
 while(ref){ref=0;
	for(d=0;d<cgdim;d++){
		if(cndef(no=cndr(nn,d))){
		       	if(cnx(no)[d]<=xx[d])		continue;
			if(cndef(g3eref(nn,d))) ref=1;
		}else{
			e =croq(d,0);
			if(cnund(qq=cnq(nn,e)))		continue;
		       	if(cnx(cqn(qq,e))[d]<=xx[d])	continue;
			if(g3qref2(qq,d)) ref=1;
		}
	}
 }
/* anisotropic refinement in positive directions: */
 for(d=0;d<cgdim;d++){
 	if(cnund(no=cndr(nn,d)))	continue;
	while(ibgrefedge(cnd(nn),cnd(no))) if(cnund(no=g3eref(nn,d))) break;
 }
/* isotropic refinement (maxlength) in negative directions: */
 xx[0]=cnx(nn)[0]-lmax;
 xx[1]=cnx(nn)[1]-lmax;
 xx[2]=cnx(nn)[2]-lmax;
 while(ref){ref=0;
	for(d=0;d<cgdim;d++){
		if(cndef(no=cndr(nn,u=cgdim+d))){
		       	if(cnx(no)[d]>=xx[d])		continue;
			if(cndef(g3eref(no,d))) ref=1;
		}else{
			e=croq(u,0);
			if(cnund(qq=cnq(nn,e)))		continue;
		       	if(cnx(no=cqn(qq,e))[d]>=xx[d])	continue;
			if(g3qref2(qq,d)) ref=1;
		}
  	}
 }
/* geometric refinement tests for refined edges: */
 for(d=0;d<cgdim;d++){u=cgdim+d;
	if(cnund(no=cndr(nn,d))) continue;
	if(cnund(nu=cndr(nn,u))) continue;
	dx=cnx(no)[d]-cnx(nu)[d];
	qq=cnq(nn,croq(d,0)); /* 3 == MAXO-1 */
	for(o=0;o<MAXO3;o++){qo=qq;
		if(cnund(qq = cnq(nn,e=croq(d,o+1)))){
			if(cnund(qo))				continue;
			if(qo!=cnq(nn,croq(u,o)))		continue;
       		}else  	if(qq!=cnq(nn,croq(u,o+1)))		continue;
/* nu-no is a refined edge of qq --- start tests for qq: */
		if(qq!=qo){
		      	if(!cnnothing(cndr(nn,rv=cror(d,o))))	continue;
			nr=cndr(nu,rv);
			if(rv<cgdim){
				dv = rv;
				dy = (cnx(nr)[dv]-cnx(nu)[dv])/dx;
			}else{	dv = rv-cgdim;
				dy = (cnx(nu)[dv]-cnx(nr)[dv])/dx;
			}
			if(dy<ibggOref) g3elot(nn,rv,d);
			else if(dy>ibggOinv){
				nor=cndr(no,rv); ibgassert(cndef(nor));
				if(cnund(cndr(nor,u))) g3elot(nor,u,rv);
				nur=cndr(nu,rv); ibgassert(cndef(nur));
				if(cnund(cndr(nur,d))) g3elot(nur,d,rv);
			}
		}else{
			nr=cqn(qq,e);
		      	rv=cror(d,o);
			if(rv<cgdim){
				dv = rv;
				dy = (cnx(nr)[dv]-cnx(nu)[dv])/dx;
			}else{	dv = rv-cgdim;
				dy = (cnx(nu)[dv]-cnx(nr)[dv])/dx;
			}
		      	rq=cror(d,o+1);
			if(rq<cgdim){
				dq = rq;
				dz = (cnx(nr)[dq]-cnx(nu)[dq])/dx;
			}else{	dq = rq-cgdim;
				dz = (cnx(nu)[dq]-cnx(nr)[dq])/dx;
			}
			if(dy<ibggOref){
				if(dz<ibggOref)
			      		g3qref2(qq,d);
				else{
					g3qref2(qq,dq);
					g3elot(nn,rv,d);
				}
			}else if(dy*dy<ibggOref2*(1.0+4*dz*dz)){
				if(dz<ibggOref)
			      		g3qref2(qq,d);
				else{
					g3qref2(qq,dq);
					g3elot(nn,rv,d);
				}
			}else if(dy>ibggOinv){
				g3qref1(qq,dq);
			}
		}
	}
 }
/* refinement of parallel neighbour edges in the directions d and u:
	ibggPrefL - number of neighbours to control
	ibggPref  - maximal ratio of length of the edge to dx
*/
 for(d=0;d<cgdim;d++){
	if(cnund(no = cndr(nn,d))) continue;
	dx = cnx(no)[d] - cnx(nn)[d];
	u  = cgdim + d;
	e  = croq(d,0);
	for(i=0;i<ibggPrefL;i++){
		if(cnund(noo=cndr(no,d))){
		   	if(cnund(qq=cnq(no,e)))	break;
			noo=cqn(qq,e);
			while(cnx(noo)[d]-cnx(no)[d]>ibggPref*dx){
				if(!g3qref2(qq,d)) break;
				noo=cqn(qq=cnq(no,e),e);
				ibgassert(cndef(noo));
			}
  			if(cnund(noo=cndr(no,d))) noo=cqn(qq=cnq(no,e),e);
		}else{
			while(cnx(noo)[d]-cnx(no)[d]>ibggPref*dx){
				if(cnund(noo=g3eref(no,d))) break;
			}
		}
		no = noo;
	}
	nu = nn;
	e  = croq(u,0);
	for(i=0;i<ibggPrefL;i++){
		if(cnund(nuu=cndr(nu,u))){
		   	if(cnund(qq=cnq(nu,e)))	break;
			nuu=cqn(qq,e);
			while(cnx(nu)[d]-cnx(nuu)[d]>ibggPref*dx){
				if(!g3qref2(qq,d)) break;
				nuu=cqn(qq=cnq(nu,e),e);
				ibgassert(cndef(nuu));
			}
			if(cnund(nuu=cndr(nu,u))) nuu=cqn(qq=cnq(nu,e),e);
		}else{
			while(cnx(nu)[d]-cnx(nuu)[d]>ibggPref*dx){
				if(cnund(nuu=g3eref(nu,u))) break;
			}
		}
		nu = nuu;
	}
 }
 for(d=0;d<cgdim;d++){
	u  = cgdim + d;
	if(cnund(nu = cndr(nn,u))) continue;
	dx = cnx(nn)[d] - cnx(nu)[d];
	e  = croq(u,0);
	for(i=0;i<ibggPrefL;i++){
		if(cnund(nuu=cndr(nu,u))){
		   	if(cnund(qq=cnq(nu,e)))	break;
			nuu=cqn(qq,e);
			while(cnx(nu)[d]-cnx(nuu)[d]>ibggPref*dx){
				if(!g3qref2(qq,d)) break;
				nuu=cqn(qq=cnq(nu,e),e);
				ibgassert(cndef(nuu));
			}
  			if(cnund(nuu=cndr(nu,u))) nuu=cqn(qq=cnq(nu,e),e);
		}else{
			while(cnx(nu)[d]-cnx(nuu)[d]>ibggPref*dx){
				if(cnund(nuu=g3eref(nu,u))) break;
			}
		}
		nu = nuu;
	}
	no = nn;
	e  = croq(d,0);
	for(i=0;i<ibggPrefL;i++){
		if(cnund(noo=cndr(no,d))){
		   	if(cnund(qq=cnq(no,e)))	break;
			noo=cqn(qq,e);
			while(cnx(noo)[d]-cnx(no)[d]>ibggPref*dx){
				if(!g3qref2(qq,d)) break;
				noo=cqn(qq=cnq(no,e),e);
				ibgassert(cndef(noo));
			}
			if(cnund(noo=cndr(no,d))) noo=cqn(qq=cnq(no,e),e);
		}else{
			while(cnx(noo)[d]-cnx(no)[d]>ibggPref*dx){
				if(cnund(noo=g3eref(no,d))) break;
			}
		}
		no = noo;
	}
 }
}

static void g2nref(int nn)
{int ref=1,i,d,u,o,rv,dv,rq,dq,e,no,noo,nu,nuu,nr,nur,nor;
 ibgFloat dx,dy,dz;
 ibgFloat xx[ibgDIM],lmax;
 ibgrefreg(cnd(nn),&lmax);
/* isotropic refinement (maxlength) in positive directions: */
 xx[0]=cnx(nn)[0]+lmax;
 xx[1]=cnx(nn)[1]+lmax;
 while(ref){ref=0;
	for(d=0;d<cgdim;d++){
		if(cndef(no=cndr(nn,d))){
		       	if(cnx(no)[d]<=xx[d])		continue;
			if(cndef(g2eref(nn,d))) ref=1;
		}
	}
 }
/* anisotropic refinement in positive directions: */
 for(d=0;d<cgdim;d++){
 	if(cnund(no=cndr(nn,d)))	continue;
	while(ibgrefedge(cnd(nn),cnd(no))) if(cnund(no=g2eref(nn,d))) break;
 }
/* isotropic refinement (maxlength) in negative directions: */
 xx[0]=cnx(nn)[0]-lmax;
 xx[1]=cnx(nn)[1]-lmax;
 while(ref){ref=0;
	for(d=0;d<cgdim;d++){
		if(cndef(no=cndn(nn,d))){
		       	if(cnx(no)[d]>=xx[d])		continue;
			if(cndef(g2eref(no,d))) ref=1;
		}
  	}
 }
/* geometric refinement tests for refined edges: */
 for(d=0;d<cgdim;d++){u=cgdim+d;
	if(cnund(no=cndr(nn,d))) continue;
	if(cnund(nu=cndr(nn,u))) continue;
	dx=cnx(no)[d]-cnx(nu)[d];
	for(o=0;o<MAXO2;o++){
	      	if(!cnnothing(cndr(nn,rv=cror(d,o))))	continue;
		nr=cndr(nu,rv);
		if(rv<cgdim){
			dv = rv;
			dy = (cnx(nr)[dv]-cnx(nu)[dv])/dx;
		}else{	dv = rv-cgdim;
			dy = (cnx(nu)[dv]-cnx(nr)[dv])/dx;
		}
		if(dy<ibggOref) g2elot(nn,rv,d);
		else if(dy>ibggOinv){
			nor=cndr(no,rv); ibgassert(cndef(nor));
			if(cnund(cndr(nor,u))) g2elot(nor,u,rv);
			nur=cndr(nu,rv); ibgassert(cndef(nur));
			if(cnund(cndr(nur,d))) g2elot(nur,d,rv);
		}
	}
 }
/* refinement of parallel neighbour edges in the directions d and u:
	ibggPrefL - number of neighbours to control
	ibggPref  - maximal ratio of length of the edge to dx
*/
 for(d=0;d<cgdim;d++){
	if(cnund(no = cndr(nn,d))) continue;
	dx = cnx(no)[d] - cnx(nn)[d];
	u  = cgdim + d;
	for(i=0;i<ibggPrefL;i++){
		if(cnund(noo=cndr(no,d))){
		}else{
			while(cnx(noo)[d]-cnx(no)[d]>ibggPref*dx){
				if(cnund(noo=g2eref(no,d))) break;
			}
		}
		no = noo;
	}
	nu = nn;
	for(i=0;i<ibggPrefL;i++){
		if(cnund(nuu=cndr(nu,u))){
		}else{
			while(cnx(nu)[d]-cnx(nuu)[d]>ibggPref*dx){
				if(cnund(nuu=g2eref(nu,u))) break;
			}
		}
		nu = nuu;
	}
 }
 for(d=0;d<cgdim;d++){
	u  = rnegat(d);
	if(cnund(nu = cndr(nn,u))) continue;
	dx = cnx(nn)[d] - cnx(nu)[d];
	for(i=0;i<ibggPrefL;i++){
		if(cnund(nuu=cndr(nu,u))){
		}else{
			while(cnx(nu)[d]-cnx(nuu)[d]>ibggPref*dx){
				if(cnund(nuu=g2eref(nu,u))) break;
			}
		}
		nu = nuu;
	}
	no = nn;
	for(i=0;i<ibggPrefL;i++){
		if(cnund(noo=cndr(no,d))){
		}else{
			while(cnx(noo)[d]-cnx(no)[d]>ibggPref*dx){
				if(cnund(noo=g2eref(no,d))) break;
			}
		}
		no = noo;
	}
 }
}

static void g1nref(int nn)
{int ref=1,i,q,d=0,u,o,rv,dv,rq,dq,e,no,noo,nu,nuu,nr,nur,nor,qq,qo;
 ibgFloat dx,dy,dz;
 ibgFloat xx[ibgDIM],lmax;
/* isotropic refinement (maxlength) in positive directions: */
 ibgrefreg(cnd(nn),&lmax);
 xx[0]=cnx(nn)[0]+lmax;
 while(ref){ref=0;
	if(cndef(no=cndp(nn,0))){
	       	if(cnx(no)[d]<=xx[0])		continue;
		if(g1eref(nn,0)) ref=1;
	}
 }
/* anisotropic refinement in positive directions: */
 if(cndef(no=cndp(nn,0))){
 	while(ibgrefedge(cnd(nn),cnd(no))) if(cnund(no=g1eref(nn,0))) break;
 }
/* isotropic refinement (maxlength) in negative directions: */
 xx[0]=cnx(nn)[0]-lmax;
 while(ref){ref=0;
	if(cndef(no=cndn(nn,0))){
	       	if(cnx(no)[0]>=xx[0])		continue;
		if(g1eref(no,0)) ref=1;
  	}
 }
/* refinement of parallel neighbour edges in the directions d and u:
	ibggPrefL - number of neighbours to control
	ibggPref  - maximal ratio of length of the edge to dx
*/
 if(cndef(no = cndr(nn,0))){
	dx = cnx(no)[0] - cnx(nn)[0];
	u  = rnegat(d);
	for(i=0;i<ibggPrefL;i++){
		if(cndef(noo=cndr(no,d))){
			while(cnx(noo)[d]-cnx(no)[d]>ibggPref*dx){
				if(cnund(noo=g1eref(no,d))) break;
			}
		}
		no = noo;
	}
	nu = nn;
	for(i=0;i<ibggPrefL;i++){
		if(cndef(nuu=cndr(nu,u))){
			while(cnx(nu)[d]-cnx(nuu)[d]>ibggPref*dx){
				if(cnund(nuu=g1eref(nu,u))) break;
			}
		}
		nu = nuu;
	}
 }
 u  = rnegat(d);
 if(cndef(nu = cndr(nn,u))){
	dx = cnx(nn)[d] - cnx(nu)[d];
	for(i=0;i<ibggPrefL;i++){
		if(cnund(nuu=cndr(nu,u))){
		}else{
			while(cnx(nu)[d]-cnx(nuu)[d]>ibggPref*dx){
				if(cnund(nuu=g1eref(nu,u))) break;
			}
		}
		nu = nuu;
	}
	no = nn;
	for(i=0;i<ibggPrefL;i++){
		if(cnund(noo=cndr(no,d))){
		}else{
			while(cnx(noo)[d]-cnx(no)[d]>ibggPref*dx){
				if(cnund(noo=g1eref(no,d))) break;
			}
		}
		no = noo;
	}
 }
}

/* Delaunay-Triangulierung des Gitters */
static ibgFloat big;

static int ibgg2find(ibgFloat *xx, int c0)
{register int nl,*cn0;
 ibgCellType t;
 int i,sl;
 ibgFloat vv,x1,x2,y1,y2;
 ibgFloat* ff[IBGCSMAX];
 register ibgFloat *fj;
 if(ccu(c0)<0){
	c0 = 1;
	while (ccu(c0)<0) c0++;
 }
beg:
 ibgassert(ccu(c0)>=0);
 nl=ibgcPoints(t=cct(c0)); cn0=ccn(c0);
 for(i=0;i<nl;i++) ff[i]=cnx(cn0[i]);
 for(i=0;i<nl;i++){
	fj = ff[ibgcSPoint1(t,i,0)];
	x1 = fj[0]-xx[0]; y1=fj[1]-xx[1];
	fj = ff[ibgcSPoint1(t,i,1)];
	x2 = fj[0]-xx[0]; y2=fj[1]-xx[1];
	vv = x1 * y2 - x2 * y1;
	if(vv<-eps2) {c0=ccs(c0)[i]; goto beg;}
 }
 return c0;
}

/* Hilfsdaten f"ur Startsituation */
static int hcn2[6] =	{1,3,2,  1,2,4};
static int hcs2[6] =	{0,2,0,  0,0,1};
static int determ=1,dslast=0,nslast=0;

static int ibgg2New(ibGrid0 *g0)
{int n,b,s,s0,sh,ms,i,i0,ih,j,d,si,sl,del;
 int n0,nn,no,nt,nh,nl,c0,c1,cc,ct,ch,co,cl;
 int *cs0,*cs1,*csi,*csj,*csh,*cn0,*cni,*cnj,*cnh,*ss,*se;
 ibgFloat *cx,*cx1,*cx2,*cx3;
 ibgFloat x,y,z,xm,ym,zm,dx,dy,dz,dd,x1,x2,x3,y1,y2,y3,z1,z2,z3,vv;
 ibgFloat l1,l2,l3,s1,s2,s3,h1,h2,h3,n1,n2,n3,det,a1,a2,a3,mx,my,mz;
 ibgCellType t,t0,t1,th,tt;
 ibgSegmentType tg; ibgPoint nd; int bb;
 tcv v,cvert;
#define DSMAX 500
 int dsc[DSMAX],dscl[DSMAX][IBGCSMAX],dsl,dsi,ds0,dsh;
#define NSMAX 600
 int nsc[NSMAX],nscs[NSMAX],nsco[NSMAX],nsl,nsi;

 ibgg0= *g0; ibgg = *(g0->gg);
 g0BNShift();
 g0QFree();

/* Speicherbereitstellung: */
 ibgg.lastPoint += 4; if(g0NRealloc()==0) return ibgError;
 g0CAlloc(ibgg.lastPoint);

/* Aufbau der Startsituation:
	4 unendlich ferne Knoten vom Mittelpunkt aus am Ende dazu.
	Sie bilden 2 Dreiecke um die x-Achse herum.
*/
 xm=0.5*(cgmin(0)+cgmax(0)); dx=(cgmax(0)-cgmin(0));
 ym=0.5*(cgmin(1)+cgmax(1)); dy=(cgmax(1)-cgmin(1));
 dd=dx*dx+dy*dy;	big=0.01*dd/ibGeo->Delta;
 nl=ibgg.lastPoint-4;
 for(i=nl+1;i<=ibgg.lastPoint;i++){
	cnx(i)[0]=xm;	cnx(i)[1]=ym;
	cnt(i)=ibgSRegion;	cnu(i)=outside;
 }
 cnx(nl+1)[0] -= big; cnx(nl+2)[0] += big;
 cnx(nl+3)[1] -= big; cnx(nl+4)[1] += big;
 cvert.x1[0]=xm+big; cvert.x1[1]=ym;
 cvert.x2[0]=xm-big; cvert.x2[1]=ym;
 for(i=1;i<3;i++){
	cct(i)=ibgc2Triangle;	ccu(i)=0;
  	ccv(i)=cvert;
	ibgg.CPointFirst[i] = 3*i-2;
	ibgg.CSideFirst[i] = 3*i-2;
 }
 ibgg.freeCPoint=ibgg.freeCSide=7; ibgg.lastCell=2;
 for(i=1;i<7;i++){
	ibgg.CPoint[i] = nl+hcn2[i-1];
	ibgg.CSide[i] = hcs2[i-1];
 }

 if((dslast>=DSMAX)||(dslast<=1))	dslast=DSMAX-1;
/* Zyklus "uber die Knoten: */
 cnc(0)=1; nl=ibgg.lastPoint-4;
 for(n0=1;n0<=nl;n0++){
/* Suche nach einer Zelle, die n0 enth"alt: */
n0zyk:	no=ibgg0.no[n0]; /* Der nahegelegene Knoten */
	c0=ibgg2find(cnx(n0),cnc(no));
	x=cnx(n0)[0]; y=cnx(n0)[1];
/* Anlegen folgender Listen:
	dsc[dsl]: Zellen, die durch den Knoten n0 zerst"ort werden;
	nsc[nsl]: neue Zellen;
*/
	nsl=0; dsl=1; dsi=1; ccu(c0)= -dsl; dsc[dsl++]=c0;
	while(dsi<dsl){
		c0  = dsc[dsi];
		cs0 = ccs(c0); ms=ibgcSides((t0=cct(c0)));
	       	cn0 = ccn(c0);
		for(s=0;s<ms;s++){
		    cc= *(cs0++);
		    if(cc==0) {del=0; goto deltest;}
		    if(ccu(cc)<0) {dscl[dsi][s]= -1; continue;}
		    v = ccv(cc); t = cct(cc);
		    if((vv= (v.x1[0]-x)*(v.x2[0]-x)
                       +    (v.x1[1]-y)*(v.x2[1]-y)) < -eps2)	del=1;
		    else					del=0;
		    if(dsl>=dslast)				del=0;
deltest:	    if(del){
	/* n0 in Umkreis -> Zelle cc wird zerst"ort */
		      	ccu(cc) = -dsl; dsc[dsl++] = cc;
			dscl[dsi][s] = -1;}
		    else{
	/* Anstelle von c0 entsteht neue Nachbarzelle c1 von cc */
			g0CGet(&c1,ibgc2Triangle);
			if(c1<0) {ibgfatal; return ibgError;}
			cni=ccn(c1); csi=ccs(c1);
		       	cn0 = ccn(c0);
			cni[0]=n0; cnc(n0)=c1; cni++; cnj=cni;
			csi[0]=cc; csi++;
			for(i=0;i<2;i++){
		       	    cni[i]  = nt = cn0[ibgcSPoint1(t0,s,i)];
		    	    cnc(nt) = c1; csi[i] = -1;
			}
	/* Berechnen und Eintragen der Daten des Umkreises: */
			cx1=cnx(*cnj++); cx2=cnx(*cnj++);
			x1=cx1[0]-x; y1=cx1[1]-y;
			x2=cx2[0]-x; y2=cx2[1]-y;
			l1=x1*x1 + y1*y1;
			l2=x2*x2 + y2*y2;
			det=x1*y2 - x2*y1;
			if((det<eps2)){ /* Entartung */
				ibgassert(determ);
				g0CDel(c1);
/* cc has to be deleted, but it's neighbours don't know this if they are
deleted before (v-test is ok for cc). test deleted neighbours for this effect*/
				csh=ccs(cc);
				for(sh=0;sh<ibgcSides(t);sh++){
				     	if(!(dsh= -ccu(ch=csh[sh]))) continue;
					if(ch==c0) continue;
					if(dsh>dsi) continue;
					for(ih=0;ih<ibgcSides(cct(ch));ih++){
						if(ccs(ch)[ih]==cc){
							cl=(dscl[ih])[dsh];
							ibgassert(cl>0);
							g0CDel(cl);
							(dscl[ih])[dsh] = -1;
						}
				     	}
				}
				del=1; goto deltest;
			}
			mx=(l1*y2 - l2*y1)/det;
			my=(l2*x1 - l1*x2)/det;
			ccv(c1).x1[0]=x;
			ccv(c1).x1[1]=y;
			ccv(c1).x2[0]=x+mx;
			ccv(c1).x2[1]=y+my;
			cvert = ccv(c1);
	/* Eintragen von c1 in die nsc- und dsc-Listen: */
			nsc[nsl]=c1; nscs[nsl]=s; nsco[nsl]=c0;
			nsl++; if(nsl>=NSMAX) ibgfatal;
			dscl[dsi][s]=c1;
		    }
		}
		dsi++;
	}
/* Herstellung der Verbindungen der neuen Zellen untereinander: */
	for(nsi=0;nsi<nsl;nsi++){	/* Zyklus "uber neue Zellen */
nsizyk:		c1 = nsc[nsi]; cs1= ccs(c1);	/* neue Zelle */
		c0 =nsco[nsi]; s0 = nscs[nsi];	/* alte Zelle */
		ds0= -ccu(c0);
		t0 =cct(c0); t1=cct(c1); cn0=ccn(c0); cs0=ccs(c0);
	/* Eintragen des Zeigers von cc auf c1: */
		if((cc=cs1[0])>0){
    			se=(ss=ccs(cc))+ibgcSides(t);
			for(;ss<se;ss++) if(ss[0]==c0)
				{ss[0]=c1; break;}
		}
		ccu(c1)=0;
		for(i0=0;i0<2;i0++){
 			if(cs1[ibgcSSide(t1,0,i0)]>0) continue;
			nh=cn0[ibgcSPoint1(t0,s0,i0)];
			ct=dscl[ds0][ibgcSSide(t0,s0,i0)];
			if(ct<=0){
	/* "Ubergang zur Nachbarzelle */
			    ch=c0;th=t0;sh=s0;ih=i0;csh=cs0;
			    do{ co =ch;
				ch =csh[ibgcSSide(th,sh,ih)];
				dsh= -ccu(ch);	ibgassert(dsh>0);
			    	th =cct(ch);
			    	cnh=ccn(ch);
			    	csh=ccs(ch);
				for(sh=0;;sh++)	{
					if(csh[sh]==co) break;
					ibgassert(sh<4);}
				for(ih=0;;ih++) {
					if(cnh[ibgcSPoint1(th,sh,ih)]==nh)
						break;
					ibgassert(ih<2);}
				ct=dscl[dsh][ibgcSSide(th,sh,ih)];
			    }while(ct<=0);
			}
			cs1[ibgcSSide(t1,0,i0)]=ct;
			cnj = ccn(ct); csj = ccs(ct); tt=cct(ct);
			for(j=0;j<4;j++) if(cnj[ibgcSPoint1(tt,0,j)]==nh){
			    		    csj[ibgcSSide(tt,0,j)]= c1; break;
			}
		}
	}
 /* Eintragen der zerst"orten Zellen in die free-Listen */
	for(dsi=1;dsi<dsl;dsi++){
		cc=dsc[dsi]; g0CDel(cc);
	}
 /* Versuche, Nachbarzellen zusammenzukleben:
	for(nsi=1;nsi<nsl;nsi++){
		cc=nsc[nsi];
		switch(t=cct(cc)){
 		    case ibgc2Triangle:
			for(i=0;i<4;i++){
				cn=
	}	*/
 }
/* Wegschmei"sen der alten Felder: */
 free(ibgg0.no);
 ibgassert(ibgg.gridDimension==2);
 *ibgg0.gg = ibgg;
 *g0 = ibgg0;
 return ibgSuccess;
}


static ibgFloat ibggVAbs = eps3, ibggVRel = 0.05;
static int delaunay_on=0;
static int unc,mec,mbc,mrc;
static int g3half=1;
static int mst,*st,*st1,nosegment0,nosegment1,nosegment2;

void nwmax();
static int er1,er2;

ibGrid0 *ibgg0CoarseGrid(unsigned maxPoint, /* for initial memory allocation */
	      	unsigned nx, ibgFloat *x, unsigned fx,
   		unsigned ny, ibgFloat *y, unsigned fy,
      		unsigned nz, ibgFloat *z, unsigned fz)
{
 unsigned i,d,qq,pp0,ix,jx,kx,lx,ppx,iy,jy,ky,ly,ppy,iz,jz,kz,lz,ppz;
 int ii,i0,i1,i2,k0,k1,k2,kk,jj,o,q,r;
 ibGrid *gg; ibGrid0 *g0;
 ibgPoint *dd;
 ibgFloat xx,dx,yy,dy,zz,dz,f;
 ibgassert(nx> 0); ibgassert(ny> 0); ibgassert(nz> 0);
 ibgassert(fx>=0); ibgassert(fy>=0); ibgassert(fz>=0);
 g0  = (ibGrid0*) calloc(1,sizeof(ibGrid0));
 gg  = (ibGrid*)  calloc(1,sizeof(ibGrid));
 ibgg  = *gg;
 ibgg0 = *g0;
 ibgg0.gg = gg;
 ibgg.spaceDimension = ibgDIM;
 if(nz>1){
	if(nx==1) {ibgfatal; return ibgg0Free(g0);}
	if(ny==1) {ibgfatal; return ibgg0Free(g0);}
	ibgg.gridDimension = 3;
	ibgg0.maxd = 3;
	ibgg0.maxr = MAXR3;
	ibgg0.maxo = MAXO3;
 	ibgg0.maxq = MAXQ3;
	ibgg0.eref = g3eref;
	ibgg0.elot = g3elot;
	ibgg0.elo2 = g3elo2;
	ibgg0.nref = g3nref;
	for(r=0;r<ibgg0.maxr;r++){for(o=0;o<=ibgg0.maxo;o++){
					ibgg0.ror[r][o]= (int) g3ror[r][o];
					ibgg0.roq[r][o]= (int) g3roq[r][o];
			         }for(o=0;o<ibgg0.maxr;o++){
	  				ibgg0.rrq[r][o]= (int) g3rrq[r][o];
	  				ibgg0.rrr[r][o]= (int) g3rrr[r][o];
	}}
	for(q=0;q<ibgg0.maxq;q++)	ibgg0.qother[q]= (int) g3qother[q];
	ibgg0.firstq = (int) nnn;
	ibgg0.lastq  = (int) ppp;
 }else if(ny>1){
	if(nx==1) {ibgfatal; return ibgg0Free(g0);}
	ibgg.gridDimension = 2;
	ibgg0.maxd = 2;
	ibgg0.maxr = MAXR2;
	ibgg0.maxo = MAXO2;
	ibgg0.maxq = MAXQ2;
	ibgg0.eref = g2eref;
	ibgg0.elot = g2elot;
	ibgg0.elo2 = g2elot;
	ibgg0.nref = g2nref;
	ibggOmax = 0;
	for(r=0;r<ibgg0.maxr;r++)	for(o=0;o<ibgg0.maxo;o++)
					ibgg0.ror[r][o]= (int) g2ror[r][o];
#if MAXQ2 > 0
	for(r=0;r<ibgg0.maxr;r++)	for(o=0;o<ibgg0.maxo;o++)
					ibgg0.roq[r][o]= (int) g2roq[r][o];
	for(q=0;q<ibgg0.maxq;q++)	ibgg0.qother[q]= (int) g2qother[q];
#endif
	ibgg0.firstq = (int) nn2;
	ibgg0.lastq  = (int) pp2;
 }else if(nx>1){
      	ibgg.gridDimension = 1;
	ibgg0.maxd = 1;
	ibgg0.maxr = MAXR1;
       	ibgg0.maxo = 0;
	ibgg0.maxq = 0;
	ibgg0.eref = g1eref;
	ibgg0.nref = g1nref;
 }else{
      	ibgg.gridDimension = 0;
	ibgg0.maxd = 0;
	ibgg0.maxr = 0;
       	ibgg0.maxo = 0;
	ibgg0.maxq = 0;
 }
 fx++; fy++; fz++;
 lz=fz*(nz-1);	ppz=1;
 ly=fy*(ny-1);	ppy=lz+1;
 lx=fx*(nx-1);	ppx=ppy*(ly+1);
 i=1; 		pp0=ppx*(lx+1);

 if(pp0>maxPoint) maxPoint = (3*pp0+20)/2;
 g0NAlloc(pp0,maxPoint);
 if(ibgg0.maxq>0) g0QAlloc(lx*ly*lz+1,maxPoint);
/*
   to avoid exact match with sqrt2, which can lead to infinite cycles
*/
#define sqrt2m 1.4142
#define sqrt2p 1.4143
 if((nx>1)&&(ny>1)){
     f = (y[ny-1]-y[0])/(x[nx-1]-x[0]);
     if((f>sqrt2m) && (f<sqrt2p)){
	 y[ny-1] += 0.001;
     }
     f = 1/f;
     if((f>sqrt2m) && (f<sqrt2p)){
	 x[nx-1] += 0.001;
     }
 }

 ibgg.xmin[0]=x[0];
 ibgg.xmin[1]=y[0];
 ibgg.xmin[2]=z[0];
 ibgg.xmax[0]=x[nx-1];
 ibgg.xmax[1]=y[ny-1];
 ibgg.xmax[2]=z[nz-1];

 for(d=0;d<ibgg.gridDimension;d++){
	cndp(outside,d) = outside;
	cndn(outside,d) = outside;
 }
 iz=0; jz=0; kz=0; zz=z[iz]; if(nz>1) dz=(z[iz+1]-zz)/fz;
 iy=0; jy=0; ky=0; yy=y[iy]; if(ny>1) dy=(y[iy+1]-yy)/fy;
 ix=0; jx=0; kx=0; xx=x[ix]; if(nx>1) dx=(x[ix+1]-xx)/fx;
 while(1){
 	cnx(i)[0] = xx;	cnx(i)[1] = yy;	cnx(i)[2] = zz;
#ifdef GLEVEL
	cnl(i)[0] = 0;	cnl(i)[1] = 0;	cnl(i)[2] = 0;
#endif
	switch(ibgg.gridDimension){
	case 3:
		if(kz<lz) cndp(i,2)	= i+ppz;
		else      cndp(i,2)	= outside;
		if(kz> 0) cndn(i,2)	= i-ppz;
		else      cndn(i,2)	= outside;
	case 2:
		if(ky<ly) cndp(i,1)	= i+ppy;
		else      cndp(i,1)	= outside;
		if(ky> 0) cndn(i,1)	= i-ppy;
		else      cndn(i,1)	= outside;
	case 1:
		if(kx> 0) cndn(i,0)	= i-ppx;
		else      cndn(i,0)	= outside;
		if(kx<lx) cndp(i,0)	= i+ppx;
		else      cndp(i,0)	= outside;
  	case 0: break;
	}
	if(kz==0){
		if(ky==0){
			if(kx==0)ibgg0.no[i] = 0;
			else	 ibgg0.no[i] = i-(ly+1)*(lz+1);
		}else		 ibgg0.no[i] = i-(lz+1);
	}else			 ibgg0.no[i] = i-1;
	i++;
	if(iz==nz-1){
		iz=0; jz=0; kz=0; zz=z[iz];
		if(nz>1) dz=(z[iz+1]-zz)/fz;
		if(iy==ny-1){
			iy=0; jy=0; ky=0; yy=y[iy];
			if(ny>1) dy=(y[iy+1]-yy)/fy;
			if(ix==nx-1) break;
			else if(jx==fx-1) {
				jx=0; kx++; ix++; xx=x[ix];
				if(nx>ix+1) dx=(x[ix+1]-xx)/fx;
			}else{
				jx++; kx++; xx+=dx;
			}
		}else if(jy==fy-1) {
			jy=0; ky++; iy++; yy=y[iy];
			if(ny>iy+1) dy=(y[iy+1]-yy)/fy;
		}else{
			jy++; ky++; yy+=dy;
		}
	}else if(jz==fz-1){
 		jz=0; iz++; kz++; zz=z[iz];
		if(nz>iz+1) dz=(z[iz+1]-zz)/fz;
	}else{
		jz++; kz++; zz+=dz;
	}
 }
/* Erzeugung der Elementarquader */
 switch(ibgg.gridDimension){
 case 3:
	for(ii=1,jj=1;ii<=pp0;ii++){
	 	for(r=0;r<MAXR3;r++){
		    if(cndr(ii,r)==outside){
			for(o=0;o<MAXO3;o++) cnq(ii,g3roq[r][o]) = outside;
		    }
		}
		if(cnund(i0=cndr(ii,0))) continue;
		if(cnund(i1=cndr(ii,1))) continue;
		if(cnund(i2=cndr(ii,2))) continue;
		k0=cndr(i1,2); k1=cndr(i2,0); k2=cndr(i0,1);
		kk=cndr(k0,0);
		cqn(jj,nnn) = ii; cnq(ii,ppp) = jj;
		cqn(jj,pnn) = i0; cnq(i0,npp) = jj;
       		cqn(jj,npn) = i1; cnq(i1,pnp) = jj;
		cqn(jj,nnp) = i2; cnq(i2,ppn) = jj;
		cqn(jj,npp) = k0; cnq(k0,pnn) = jj;
		cqn(jj,pnp) = k1; cnq(k1,npn) = jj;
		cqn(jj,ppn) = k2; cnq(k2,nnp) = jj;
		cqn(jj,ppp) = kk; cnq(kk,nnn) = jj;
		jj++;
	 }
 default:break;
 }
 *gg = ibgg;
 *g0 = ibgg0;
 return g0;
}

ibGrid0 *ibgg0Free(ibGrid0 *gg)
{int i;
 if((gg->no)	!= ibgNULL) free(gg->no);
 if((gg->nb)	!= ibgNULL) free(gg->nb);
 if((gg->nstack)!= ibgNULL) free(gg->nstack);
 for(i=0;i<ibgDIM;i++){
	if((gg->ndp[i])	!= ibgNULL) free(gg->ndp[i]);
	if((gg->ndn[i])	!= ibgNULL) free(gg->ndn[i]);
	if((gg->qdp[i])	!= ibgNULL) free(gg->qdp[i]);
	if((gg->qdn[i])	!= ibgNULL) free(gg->qdn[i]);
 }
 for(i=0;i<ibgDIMQ;i++){
	if((gg->nq[i])	!= ibgNULL) free(gg->nq[i]);
	if((gg->qn[i])	!= ibgNULL) free(gg->qn[i]);
 }
 if((gg->qn0)	!= ibgNULL) free(gg->qn0);
 if((gg->bPoint)	!= ibgNULL) free(gg->bPoint);
 if((gg->bn)	!= ibgNULL) free(gg->bn);
 if((gg->bstack)!= ibgNULL) free(gg->bstack);
 ibGridFree(gg->gg);
 free(gg);
 return (ibGrid0*) ibgNULL;
}


int	ibgg0RegionOfPoint(ibGrid0 *g0, ibGeometry g)
{unsigned n,ni,nj,r,ist,lst,*nst,len; char *nok;
 ibgg0 = *g0;
 ibgg  = *ibgg0.gg;
 len = ibgg.lastPoint+1;
#undef calloc
 nok = (char *) calloc(len,sizeof(int));
 if(nok == ibgNULL) return ibgError;
 nst=(unsigned *) malloc((ibgg.lastPoint+1)*sizeof(unsigned));
 if(nst == ibgNULL) {free(nok); return ibgError;}

 ibgiRegionOfPoint(g,cnd(1),(ibgPoint*)ibgNULL);
 nok[1]=1; lst=1; ist=1; ni=1;
 do{
 	for(r=0;r<ibgg0.maxr;r++){
		nj=cndr(ni,r);
		if(cndef(nj)) if(!nok[nj]) {
			ibgiRegionOfPoint(g,cnd(nj),cnd(ni));
			nok[nj]=1; nst[lst++]=nj;
		}
	}
	ni=nst[ist++];
 }while(ist<=lst);

 ibgassert(lst==ibgg.lastPoint);
 free(nok); free(nst);
 return ibgSuccess;
}

void ibgg0SetRefPar(ibggRefPar *ref,
	     	void (*RefineRegion)(ibgPtObject This, ibgPoint *point,
			   	      	ibgFloat *length),
		ibgPtObject pRefineRegion,
		void (*RefineFace)(ibgPtObject This, ibgPoint *point,
				      	ibgFloat *normal, ibgFloat *tangential),
	     	ibgPtObject pRefineFace,
	     	void (*RefineLine)(ibgPtObject This, ibgPoint *point,
			   	      	ibgFloat *normal, ibgFloat *tangential),
	     	ibgPtObject pRefineLine,
	     	void (*RefineNode)(ibgPtObject This, ibgPoint *point,
			   	      	ibgFloat *length),
	     	ibgPtObject pRefineNode,
		ibgBoolean (*RefineEdge) (ibgPtObject This,
					ibgPoint *n1, ibgPoint *n2),
		ibgPtObject pRefineEdge)
{
 ref->RefineRegion	= RefineRegion;
 ref->pRefineRegion	= pRefineRegion;
 ref->RefineEdge	= RefineEdge;
 ref->pRefineEdge	= pRefineEdge;
 ref->RefineFace	= RefineFace;
 ref->pRefineFace	= pRefineFace;
 ref->RefineLine	= RefineLine;
 ref->pRefineLine	= pRefineLine;
}

/* refinement of the grid (main routine): */
int	ibgg0Refine	 (ibGrid0 *g0, ibGeometry geom, ibggRefPar *ref)
{int nn,qq;
 ibGeo	= geom;
 ibgg0 = *g0;
 ibgg  = *ibgg0.gg;
 ibggRef = *ref;
 for(nn=1;nn<ibgg.lastPoint;nn++) ibgg0.nref(nn);
 if(ibggOmax>0){
	qq=1;
	while(qq<ibgg0.lastQuad){
		g3qref(qq); qq++;
	  	while(nn<ibgg.lastPoint){
			ibgg0.nref(nn); nn++;
		}
	}
 }
 *ibgg0.gg = ibgg;
 *g0 = ibgg0;
 return ibgSuccess;
moverflow:
 *ibgg0.gg = ibgg;
 *g0 = ibgg0;
 ibgg0Free(g0);
 return ibgError;
}

int ibgg0Delaunay(ibGrid0 *g0)
{
 switch(ibgridGDIM(*(g0->gg))){
 case 0: return ibgg0New(g0);
 case 1: return ibgg1New(g0);
 case 2: return ibgg2New(g0);
 case 3: return ibgg3New(g0);
 }
}
/* boundary shift with regularization of the ibgg (main routine): */
int	ibgg0Boundary	 (ibGrid0 *g0, ibGeometry geom, ibggRefPar *ref,
	     	unsigned bmax)
{int b,bl,nn,ncl,n,qq,no,nh,o,d;
 ibgPoint *bx; int *bn,*nb;
 ibgg0 = *g0;
 ibgg  = *ibgg0.gg;
 ibggRef = *ref;
 ibGeo = geom;

/* Speicherreservierung */
 if(ibgg0.maxBound==0)	g0BAlloc(bmax);
 else if(ibgg0.maxBound<bmax){
	ibgg0.lastBound = bmax;
	g0BRealloc();
 }
 ibgg0.lastBound = 0;
 g0bstb = -1;

 ncl = ibgg.lastPoint;
 qq  = ibgg0.lastQuad+1;
 g0bfirst  = 0;
 g0nfirst  = 0;
/* first boundary-shift for old points: */
 while(g0nfirst<ncl){g0nfirst++;
	ibgg0.nstack[g0nfirst] = 0;
	g0bshift(g0nfirst,1);
 }
 do{						/* ibgFloat-boundary-stack */
	if(ibggOmax>0){	/* cuboid-stack */
		while(qq<=ibgg0.lastQuad){
			g3qref(qq); qq++;
		}
	}
	if(g0nfirst<ibgg.lastPoint){g0nfirst++;
		ibgg0.nstack[g0nfirst]=0;
	    	ibgg0.nref(g0nfirst);
		g0bshift(g0nfirst,1);
		continue;
	}
	if(g0nstb>0){
	     	n=g0nstb; g0nstb=ibgg0.nstack[n]; ibgg0.nstack[n]=0;
		g0bshift(n,0);
		continue;
	}
	if(g0bfirst<ibgg0.lastBound){g0bfirst++;
		if(ibgg0.bn[g0bfirst]>0){
			ibgg0.bstack[g0bfirst]=0;
		   	g0breg(g0bfirst);
		}
	 	continue;
	}
	if(g0bstb>0){
		b=g0bstb; g0bstb=ibgg0.bstack[b]; ibgg0.bstack[b]=0;
		if(ibgg0.bn[b]>0){
			g0breg(b);
		}else{
			ibgassert(ibgg0.bn[b] >= 0);
		}
		continue;
	}
	break;
 }while(1);		/* now all stacks are empty */
 *ibgg0.gg = ibgg;
 free(ibgg0.nstack);
 *g0 = ibgg0;
 return ibgSuccess;
moverflow:
 *ibgg0.gg = ibgg;
 *g0 = ibgg0;
 ibgg0Free(g0);
 return ibgError;
}
