/* last edit: Ilja Schmelzer -------------- 28-FEB-1995 11:07:13.73	*/
/************************************************************************/
/*                                                                      */
/*  <<< I B G >>> - Intersection - Based Grid generation package 	*/
/*                                                                      */
/*  Version 1.1 by Ilja Schmelzer   schmelzer@iaas-berlin.d400.de       */
/*                                                                      */
/*  to be distributed under IBG license conditions (see "readme.ibg")	*/
/*                                                                      */
/************************************************************************/
#include "ibg.h"
#include "ibgoutput.h"
#include "ibgapplication.h"
#include "ibgdefault.h"
#include "ibgg.h"
#include "ibgc.h"
#include "ibglib.h"
#include "ibgerror.h"
#include <stdio.h>
#include <string.h>

/* declaration of external parameters - ibgSMAX-arrays: */
int		ibgMaterial[ibgSMAX];
int		ibgBCondition[ibgSMAX];
int		ibgColRegion[ibgSMAX];
int		ibgColFace[ibgSMAX];
int		ibgColLine[ibgSMAX];
int		ibgColNode[ibgSMAX];
int		ibgColMaterial[ibgSMAX];
int		ibgColBCondition[ibgSMAX];
ibgBoolean	ibgOSSpecial[ibgSMAX];
char*		ibgNameRegion[ibgSMAX];
char*		ibgNameFace[ibgSMAX];
char*		ibgNameLine[ibgSMAX];
char*		ibgNameNode[ibgSMAX];
char*		ibgNameMaterial[ibgSMAX];
char*		ibgNameBCondition[ibgSMAX];

/* other arrays: */
ibgSegment		ibgOSList[ibgOSMaxLength];
ibgFloat	ibgWX[ibgDIM];
ibgFloat	ibgWAlpha,ibgWBeta;
ibgFloat	ibgWXMin[ibgDIM];
ibgFloat	ibgWXMax[ibgDIM];
ibgColor	ibgColorOf[ibgColorMAX];
char*		ibgColorName[ibgColorMAX];
ibgBoolean	ibgOutputType[IBGSTMAX];
char*		ibgSname[IBGSTMAX] = {"IsRegion","IsFace","IsLine","IsNode",
					""};

/* declaration and initialization of scalar external parameters: */
int 		(*ibgBCDefault)(ibgSegment u1, ibgSegment u2) = ibgBCDefaultDirichlet;
ibgBoolean 	ibgMOn;
ibgBoolean	ibgBCOn;
ibgSegmentType	ibgOSType;
ibgBoolean	ibgOSOutside;
int		ibgOSMode;
int		ibgOSListLength;
char*		ibgOSName;
int		ibgIdPoint;
int		ibgIdCell;
ibgBoolean	ibgWindow;
int		ibgColorFree;
int		ibgColorCurrent;
int		ibgColDoubleBoundary;
int		ibgColSingleBoundary;
int		ibgColOutsideBoundary;
ibgSegment		ibgOSSegment;
ibgBoolean	ibgResetGrid;
char*		ibgNameApplication;
char*		ibgNameFile;
int  		ibgFileMode;
ibGrid		*ibggCurrent;
int  	 	ibgFileModFree;
char*	 	ibgFileModName[ibgFileModMAX];
char*	 	ibgFileType[ibgFileModMAX];
int	 	(*(ibgFileOutput[ibgFileModMAX]))(ibGrid *g, FILE *file);
static 	int	ibgreadcolors();

void	ibgOutputInit()
{ibgSegment u; int color,fmod;
 ibgBCDefault	= ibgBCDefaultDirichlet;
 ibgMOn		= ibgFalse;
 ibgBCOn	= ibgFalse;
 ibgOSType	= ibgSRegion;
 ibgOSOutside	= ibgFalse;
 ibgOSMode	= ibgOSModSpecial;
 ibgOSListLength	= 0;
 ibgOSName	= ibgOSNoName;
 ibgIdPoint	= ibgIdModOff;
 ibgIdCell	= ibgIdModOff;
 ibgWindow	= ibgFalse;
 ibgColorFree	= 0;
 ibgColorCurrent	= -1;
 ibgColDoubleBoundary	= ibgColModMax;
 ibgColSingleBoundary	= ibgColModSelf;
 ibgColOutsideBoundary	= ibgColModSelf;
 ibgOSSegment		= ibgRSpace;
 ibgResetGrid 		= ibgTrue;
 if(ibgNameApplication == ibgNULL) ibgNameApplication	= "ibg";
 ibgNameFile		= "ibg";
 ibgFileMode		= ibgFileModNone;
 ibgFileModFree		= 0;
 ibggCurrent		= ibgNULL;

 for(u=0;u<ibgSMAX;u++){
	ibgMaterial[u]		= u;
	ibgBCondition[u]	= u;
	ibgNameRegion[u]	= ibgOSNoName;
	ibgNameFace[u]		= ibgOSNoName;
	ibgNameMaterial[u]	= ibgOSNoName;
	ibgNameBCondition[u]	= ibgOSNoName;
	ibgColRegion[u]		= ibgColDefaultRegion;
	ibgColFace[u]		= ibgColDefaultFace;
	ibgColLine[u]		= ibgColDefaultLine;
	ibgColNode[u]		= ibgColDefaultNode;
	ibgColMaterial[u]	= ibgColDefaultRegion;
	ibgColBCondition[u]	= ibgColDefaultFace;
	ibgOSSpecial[u]		= ibgTrue;
 }
 for(u=0;u<IBGSTMAX;u++)	ibgOutputType[u] = ibgFalse;
 ibgNameRegion[0]	= "Unknown";
 ibgNameFace[0]		= "Unknown";
 ibgNameMaterial[0]	= "Unknown";
 ibgNameBCondition[0]	= "Unknown";
 ibgNameMaterial[ibgMAir]		= "Air";
 ibgNameRegion[ibgRSpace]		= "Space";
 ibgNameBCondition[ibgBCDirichlet] 	= "Dirichlet";
 ibgNameBCondition[ibgBCNeumann]	= "Neumann";
 ibgNameBCondition[ibgBCMixed]		= "Mixed";
/* color declarations: */
 ibgreadcolors();
 color = ibgDeclareColor("ColBackground",	"#fdf5e6"); /* OldLace	*/
 ibgAssert(color == ibgColBackground);
 color = ibgDeclareColor("ColError",		"#ff0000"); /* red	*/
 ibgAssert(color == ibgColError);
 color = ibgDeclareColor("ColForeground",	"#000000"); /* black	*/
 ibgAssert(color == ibgColForeground);
 color = ibgDeclareColor("ColDefaultNode",	"#b22222"); /* firebrick  */
 ibgAssert(color == ibgColDefaultNode);
 color = ibgDeclareColor("ColDefaultLine",	"#ff6347"); /* tomato	*/
 ibgAssert(color == ibgColDefaultLine);
 color = ibgDeclareColor("ColDefaultFace",	"#e9967a"); /* DarkSalmon */
 ibgAssert(color == ibgColDefaultFace);
 ibgAssert(color == ibgColDefaultBCondition);
 ibgColBCondition[ibgBCDirichlet]    	=
 	 ibgDeclareColor("ColDirichlet", 	"#32cd32"); /* LimeGreen */
 ibgColBCondition[ibgBCNeumann]		=
	 ibgDeclareColor("ColNeumann",		"#4169e1"); /* RoyalBlue */
 ibgColBCondition[ibgBCMixed]		=
	 ibgDeclareColor("ColMixed",		"#20b2aa"); /* LightSeaGreen */
 color = ibgDeclareColor("ColDefaultRegion",	"#a0522d"); /* sienna	*/
 ibgAssert(color == ibgColDefaultRegion);
 ibgAssert(color == ibgColDefaultMaterial);
 ibgColMaterial[ibgMAir] 	       	=  ibgColRegion[ibgRSpace]	=
	 ibgDeclareColor("ColAir",  		"#c0c0c0");	/* gray	*/
/* output function list initialization: */
 fmod =	ibgDeclareFileMode("None",	"",	ibgNoFunction);
 ibgAssert(fmod == ibgFileModNone);
 fmod =	ibgDeclareFileMode("ibgDump",	"dmp",	ibGridSave);
 ibgAssert(fmod == ibgFileModSave);
 fmod =	ibgDeclareFileMode("ibGrid",	"grd",	ibGridWrite);
 ibgAssert(fmod == ibgFileModWrite);
 fmod =	ibgDeclareFileMode("Simplex",	"smp",	ibgToSimplex);
 ibgAssert(fmod == ibgFileModSimplex);
}

void	ibgOutput(ibGrid *g)
{char c,filename[500];
 int i,rc,notype=1;
 FILE *file;
 int d;
 ibggCurrent = g;
 if(ibgResetGrid){
	ibgResetGrid = ibgFalse;
	for(d=0;d<ibgDIM;d++){
		ibgWXMin[d] = ibgridXMin(*g,d);
		ibgWXMax[d] = ibgridXMax(*g,d);
		ibgWX[d]    = ibgridXMax(*g,d) - ibgridXMin(*g,d);
		if(ibgWX[d]<= 0.0) ibgWX[d] = 1.0;
	}
 	ibgWindow = ibgTrue;
	if(ibgridGDIM(*g) == 3){
	      	ibgOutputType[ibgSRegion]	= ibgFalse;
		ibgOutputType[ibgSFace]		= ibgTrue;
	}else{
	      	ibgOutputType[ibgSRegion]	= ibgTrue;
		ibgOutputType[ibgSFace]		= ibgFalse;
	}
 }
/* user control over output parameters: */
 ibgControl();
/* consistency tests: */
 if((ibgFileMode <= ibgFileModNone)||
    (ibgFileMode >= ibgFileModFree)) {
     ibgmessage(ibgMEIncFileMode);
     return;
 }
/* file name creation (with correct file type): */
 i=0;
 while(c=ibgNameFile[i]){
    	filename[i++] = c; if(c=='.') notype=0;
 }
 if(notype){int j=0;
    	filename[i++] = '.';
	while(filename[i++]=(ibgFileType[ibgFileMode])[j++]);
 }
 filename[i++] = '\0';
/* start of file writing: */
 file = fopen(filename,"w");
 if(file==NULL){
	/* unable to open file */
     ibgmessage(ibgMEOpenFile);
     ibgprintf("%s", filename);
     ibgmsgreset;
     return;
 }
 if(ibgFileOutput[ibgFileMode]){
     rc = ibgFileOutput[ibgFileMode](g,file);
     if(rc != ibgSuccess){
	 ibgmessage(ibgMEWriteFile);
     }else{
	 ibgmessage(ibgMIWriteFile);
     }
     ibgprintf("%s", filename);
 }else{
	/* output function not defined */
     ibgmessage(ibgMEIncFileMode);
 }
 ibgmsgreset;
 fclose(file);
 return;
}

ibgBoolean	ibgOSelected(ibgSegment u)
{char* nam;int i;
 switch(ibgOSMode){
 case ibgOSModAll:	return ibgTrue;
 case ibgOSModOne:
	if(u==ibgOSSegment)	return ibgTrue;	else	return ibgFalse;
 case ibgOSModSpecial:
	if((u<0)||(u>=ibgSMAX))	return ibgOSSpecial[0];
	else			return ibgOSSpecial[u];
 case ibgOSModName:
	if((u<0)||(u>=ibgSMAX))	u = 0;
	switch(ibgOSType){
	case ibgSRegion:	nam = ibgNameRegion[u];	break;
	case ibgSFace:		nam = ibgNameFace[u];	break;
	case ibgSLine:		nam = ibgNameLine[u];	break;
	case ibgSNode:	nam = ibgNameNode[u];	break;
	}
	if(strcmp(nam,ibgOSName)) return ibgFalse; else return ibgTrue;
 case ibgOSModNamed:
	if((u<0)||(u>=ibgSMAX))	u = 0;
	switch(ibgOSType){
	case ibgSRegion:	nam = ibgNameRegion[u];	break;
	case ibgSFace:		nam = ibgNameFace[u];	break;
	case ibgSLine:		nam = ibgNameLine[u];	break;
	case ibgSNode:	nam = ibgNameNode[u];	break;
	}
	if(nam == ibgOSNoName)	return ibgFalse; else return ibgTrue;
 case ibgOSModList:
	for(i=0;i<ibgOSListLength;i++){
		if(u == ibgOSList[i])	return ibgTrue;
	}
	return ibgFalse;
 }
}

ibgBoolean	ibgOSelectedRegion(ibgSegment u)
{if(ibgOutputType[ibgSRegion]){
	if(ibgOSType == ibgSRegion)
		return ibgOSelected(u);
	else
		return ibgFalse;
 }else{
	return ibgFalse;
 }
}

ibgBoolean	ibgOSelectedFace(ibgSegment u, ibgSegment left, ibgSegment right)
{if(ibgOutputType[ibgSFace]){
	switch(ibgOSType){
	case ibgSRegion:
		if(!ibgOSOutside){
			if(left==0)	return ibgFalse;
			if(right==0)	return ibgFalse;
		}
		if(ibgOSelected(left))	return ibgTrue;
		if(ibgOSelected(right))	return ibgTrue;
		return ibgFalse;
	case ibgSFace:
		if(u<ibgSMAX)
			return ibgOSelected(u);
	default:
       		return ibgFalse;
	}
 }else{
	return ibgFalse;
 }
}

ibgBoolean	ibgOSelectedLine(ibgSegment u)
{if(ibgOutputType[ibgSLine]){
	switch(ibgOSType){
	case  ibgSLine:
		return ibgOSelected(u);
	case  ibgSNode:
		return ibgFalse;
	default:
		return ibgTrue;
	}
 }else{
	return ibgFalse;
 }
}

ibgBoolean	ibgOSelectedNode(ibgSegment u)
{if(ibgOutputType[ibgSNode]){
	if(ibgOSType == ibgSNode)
		return ibgOSelected(u);
	else
		return ibgTrue;
 }else{
	return ibgFalse;
 }
}

ibgBoolean	ibgIdentPoint(int n, char* ident)
{int num; char* nam; ibgSegmentType typ;
 switch(ibgIdPoint){
 case ibgIdModOff:	return ibgFalse;
 case ibgIdModOn:
	num = n;
	goto numwrite;
 case ibgIdModTyp:
	num = ibgpType(ibgridPoint(*ibggCurrent,n));
	goto numwrite;
 case ibgIdModSegment:
	num = ibgpSegment(ibgridPoint(*ibggCurrent,n));
	goto numwrite;
 case ibgIdModMaterial:
	num = ibgpSegment(ibgridPoint(*ibggCurrent,n));
	if((num<0) || (num>=ibgSMAX))	num = 0;
	typ = ibgpType(ibgridPoint(*ibggCurrent,n));
	switch(typ){
	case ibgSRegion:
		num = ibgMaterial[num];	break;
	case ibgSFace:
		num = ibgBCondition[num]; break;
	}
	goto numwrite;
 case ibgIdModNType:
	num = ibgpType(ibgridPoint(*ibggCurrent,n));
	nam = ibgSname[num];
	goto txtwrite;
 case ibgIdModNSegment:
	num = ibgpSegment(ibgridPoint(*ibggCurrent,n));
	if((num<0) || (num>=ibgSMAX))	num = 0;
	typ = ibgpType(ibgridPoint(*ibggCurrent,n));
	switch(typ){
	case ibgSRegion:	nam = ibgNameRegion[num];	goto txtwrite;
	case ibgSFace:		nam = ibgNameFace[num];		goto txtwrite;
	case ibgSLine:		nam = ibgNameLine[num];		goto txtwrite;
	case ibgSNode:	nam = ibgNameNode[num];	goto txtwrite;
	}
 case ibgIdModNMaterial:
	num = ibgpSegment(ibgridPoint(*ibggCurrent,n));
	if((num<0) || (num>=ibgSMAX))	num = 0;
	typ = ibgpType(ibgridPoint(*ibggCurrent,n));
	switch(typ){
	case ibgSRegion:
		num = ibgMaterial[num];
		nam = ibgNameMaterial[num];	goto txtwrite;
	case ibgSFace:
		num = ibgBCondition[num];
		nam = ibgNameBCondition[num];	goto txtwrite;
	case ibgSLine:
		nam = ibgNameLine[num];		goto txtwrite;
	case ibgSNode:
		nam = ibgNameNode[num];	goto txtwrite;
	}
 }
numwrite:
 sprintf(ident,"%u",num);
 return ibgTrue;
txtwrite:
 sprintf(ident,"%s",nam);
 return ibgTrue;
}

ibgBoolean	ibgIdentCell(int c, char* ident)
{int num; char* nam; ibgCellType typ;
 switch(ibgIdCell){
 case ibgIdModOff:	return ibgFalse;
 case ibgIdModOn:
	num = c;
	goto numwrite;
 case ibgIdModTyp:
	num = ibgridCellType(*ibggCurrent,c);
	goto numwrite;
 case ibgIdModSegment:
	num = ibgridCellSegment(*ibggCurrent,c);
	goto numwrite;
 case ibgIdModMaterial:
	num = ibgridCellSegment(*ibggCurrent,c);
	if((num<0) || (num>=ibgSMAX))	num = 0;
	typ = ibgridCellType(*ibggCurrent,c);
	switch(typ){
	case ibgSRegion:
		num = ibgMaterial[num];	break;
	case ibgSFace:
		num = ibgBCondition[num]; break;
	}
	goto numwrite;
 case ibgIdModNType:
	num = ibgridCellType(*ibggCurrent,c);
	nam = ibgcname[num];
	goto txtwrite;
 case ibgIdModNSegment:
	num = ibgridCellSegment(*ibggCurrent,c);
	if((num<0) || (num>=ibgSMAX))	num = 0;
	typ = ibgridCellType(*ibggCurrent,c);
	switch(typ){
	case ibgSRegion:	nam = ibgNameRegion[num];	goto txtwrite;
	case ibgSFace:		nam = ibgNameFace[num];		goto txtwrite;
	case ibgSLine:		nam = ibgNameLine[num];		goto txtwrite;
	case ibgSNode:	nam = ibgNameNode[num];	goto txtwrite;
	}
 case ibgIdModNMaterial:
	num = ibgridCellSegment(*ibggCurrent,c);
	if((num<0) || (num>=ibgSMAX))	num = 0;
	typ = ibgridCellType(*ibggCurrent,c);
	switch(typ){
	case ibgSRegion:
		num = ibgMaterial[num];
		nam = ibgNameMaterial[num];	goto txtwrite;
	case ibgSFace:
		num = ibgBCondition[num];
		nam = ibgNameBCondition[num];	goto txtwrite;
	case ibgSLine:
		nam = ibgNameLine[num];		goto txtwrite;
	case ibgSNode:
		nam = ibgNameNode[num];	goto txtwrite;
	}
 }
numwrite:
 sprintf(ident,"%u",num);
 return ibgTrue;
txtwrite:
 sprintf(ident,"%s",nam);
 return ibgTrue;
}

int	ibgColorRegion(ibgSegment u)
{
 if((u>=ibgSMAX) || (u<0)){
	if(ibgMOn)	return ibgColDefaultMaterial;
	else		return ibgColDefaultRegion;
 }
 if(ibgMOn)	return ibgColMaterial[ibgMaterial[u]];
 else		return ibgColRegion[u];
}

int	ibgColorFace(ibgSegment u, ibgSegment left, ibgSegment right)
{int mod;ibgSegment u0,u1,u2;
 if(left){
	if(ibgOSelected(left)){
		if(right){
			if(ibgOSelected(right)){
			      	mod = ibgColDoubleBoundary;
				u1  = left; u2 = right;
			}else{
			      	mod = ibgColSingleBoundary;
				u1  = left; u2 = right;
			}
		}else{
			mod = ibgColOutsideBoundary;
			u1  = left; u2 = right;
		}
	}else{
		if(right){
			if(ibgOSelected(right)){
			      	mod = ibgColSingleBoundary;
				u1  = right; u2 = left;
			}else{
				if(u<ibgSMAX)	return ibgColFace[u];
				return ibgColError;
			}
		}else{
			if(u<ibgSMAX)	return ibgColFace[u];
			return ibgColError;
		}
 	}
 }else{
	if(right){
		if(ibgOSelected(right)){
		      	mod = ibgColOutsideBoundary;
			u1  = right; u2 = left;
		}else{
			if(u<ibgSMAX)	return ibgColFace[u];
			return ibgColError;
		}
	}else{
		if(u<ibgSMAX)	return ibgColFace[u];
		return ibgColError;
	}
 }
 switch(mod){
	case ibgColModZero:	return ibgColDefaultFace;
	case ibgColModMax:	if(u1>u2) u0 = u1; else u0 = u2; break;
	case ibgColModMin:	if(u1<u2) u0 = u1; else u0 = u2; break;
	case ibgColModSelf:	u0 = u1; break;
	case ibgColModOther:	u0 = u2; break;
	case ibgColModBC:
		if((u>=ibgSMAX) || (u<0)){
		 	if(ibgBCOn)
				return ibgColBCondition[ibgBCDefault(u1,u2)];
			else
				return ibgColDefaultFace;
		}
	 	if(ibgBCOn)
			return ibgColBCondition[ibgBCondition[u]];
		else
			return ibgColFace[u];
 }
 return ibgColorRegion(u0);
}

int	ibgColorLine(ibgSegment u)
{if((u>=ibgSMAX) || (u<0))	return ibgColDefaultLine;
 return ibgColLine[u];
}

int    	ibgColorNode(ibgSegment u)
{if((u>=ibgSMAX) || (u<0))	return ibgColDefaultNode;
 return ibgColNode[u];
}

int	ibgBCDefaultDirichlet(ibgSegment u1, ibgSegment u2)
{return ibgBCDirichlet;}
int	ibgBCDefaultNeumann(ibgSegment u1, ibgSegment u2)
{return ibgBCNeumann;}
int	ibgBCDefaultMixed(ibgSegment u1, ibgSegment u2)
{return ibgBCMixed;}
int	ibgBCDefaultMax(ibgSegment u1, ibgSegment u2)
{if(u1>u2) return u1; else return u2;}
int	ibgBCDefaultMin(ibgSegment u1, ibgSegment u2)
{if(u1<u2) return u1; else return u2;}


ibgBoolean	ibgInWindow(ibgPoint *n)
{ibgFloat x,y,z;
 if(((x=ibgpX(*n)[0]) < ibgWXMin[0]) || (x > ibgWXMax[0])) return ibgFalse;
 if(((y=ibgpX(*n)[1]) < ibgWXMin[1]) || (y > ibgWXMax[1])) return ibgFalse;
 if(((z=ibgpX(*n)[2]) < ibgWXMin[2]) || (z > ibgWXMax[2])) return ibgFalse;
 return ibgTrue;
}

ibgBoolean	ibgWindowCoord(ibgPoint *n, ibgFloat *x)
{
 x[0] = (ibgpX(*n)[0] - ibgWXMin[0]) / ibgWX[0];
 if((x[0] < -0.01) || (x[0] > 1.01))	return ibgFalse;
 x[1] = (ibgpX(*n)[1] - ibgWXMin[1]) / ibgWX[1];
 if((x[1] < -0.01) || (x[1] > 1.01))	return ibgFalse;
 x[2] = (ibgpX(*n)[2] - ibgWXMin[2]) / ibgWX[2];
 if((x[2] < -0.01) || (x[2] > 1.01))	return ibgFalse;
 x[0] *= 0.9; x[0] += 0.05;
 x[1] *= 0.9; x[1] += 0.05;
 x[2] *= 0.9; x[2] += 0.05;
 return ibgTrue;
}

#define R4MaxColors 1000
static ibgColor R4Colors[R4MaxColors];
static char*	R4ColorNames[R4MaxColors];
#define R4MaxChars 10000
static char	R4NameString[R4MaxChars];
static int   	R4colors;

ibgBoolean	ibgOSetColor(int id, char* rgb)
{int r,g,b,rc,i;
 if(*rgb == '#'){
	switch(strlen(rgb)){
	case 4:
		rc = sscanf(rgb,"#%1x%1x%1x",&r,&g,&b);
		r *= 0x1000;
		g *= 0x1000;
		b *= 0x1000;
		break;
	case 7:
		rc = sscanf(rgb,"#%2x%2x%2x",&r,&g,&b);
		r *= 0x100;
		g *= 0x100;
		b *= 0x100;
		break;
	case 10:
		rc = sscanf(rgb,"#%3x%3x%3x",&r,&g,&b);
		r *= 0x10;
		g *= 0x10;
		b *= 0x10;
		break;
	case 13:
		rc = sscanf(rgb,"#%4x%4x%4x",&r,&g,&b);
	      	break;
	default:
		rc = 0;
	}
	if(rc < 3)	return ibgFalse;
	(ibgColorOf[id]).r = r;
	(ibgColorOf[id]).g = g;
	(ibgColorOf[id]).b = b;
	return ibgTrue;
 }else{
	for(i=0;i<R4colors;i++){
		if(strcmp(R4ColorNames[i],rgb)) continue;
		ibgColorOf[id] = R4Colors[i];
		return ibgTrue;
	}
	return ibgFalse;
 }
}


static char buffer[500];
static int buflen=500;

/* This function reads the X11 color definition file rgb.txt.
It contains lines of the following format:
255 255 255		white
*/

static int ibgreadcolors()
{FILE *file;
/* FILE *fout; allows to print out in format "#ffffff white" */
 int last=0,col=0,r,g,b,rc;
 char s2[50];
 file = fopen("rgb.txt","r+");
/* fout = fopen("ibgcolors.txt","w");*/
 if(file==NULL){
	R4colors = 0;
	return ibgFalse;
 }
 while(!feof(file)){
  	fgets(buffer,buflen,file);
	rc = sscanf(buffer,"%d %d %d %s %s",&r,&g,&b,&R4NameString[last],s2);
	if(rc != 4) continue;
	R4Colors[col].r = r*0x100;
	R4Colors[col].g = g*0x100;
	R4Colors[col].b = b*0x100;
	R4ColorNames[col]= &R4NameString[last];
/*	fprintf(fout,"#%2x%2x%2x %s\n",r,g,b,&R4NameString[last]);*/
	last += strlen(R4ColorNames[col])+1;
	col++;
	if(col >= R4MaxColors) 		break;
	if(last>= R4MaxChars+20)	break;
 }
 R4colors = col;
 fclose(file);
/* fclose(fout);*/
 return ibgTrue;
}

int	ibgDeclareColor(char* name, char* value)
{int last = ibgColorFree++;
 if(ibgColorFree >= ibgColorMAX)	return 0;
 ibgColorName[last]		= name;
 ibgColorName[ibgColorFree]	= "";
 ibgOSetColor(last,value);
 return last;
}

int	ibgDeclareFileMode(char* mode, char* filetype,
	   	   	int (*output)(ibGrid *g, FILE *file)
	)
{int last = ibgFileModFree++;
 if(ibgFileModFree >= ibgFileModMAX)	return 0;
 ibgFileModName[last]	= mode;
 ibgFileType[last]	= filetype;
 ibgFileOutput[last]	= output;
 ibgFileModName[ibgFileModFree]	= "";
 return last;
}
